#!/bin/bash

#**********************************************
# this script controls the execution of a lattice free energy calculation,
# discretizes the coarse-grained solute-solute potential, and creates the .sym file
#
# note two command line arguments need to be present, the name of the solute, and the name of the
# framework
#**********************************************

executable1=main_lattice_grid
executable2=discretize_solute_potential
executable3=lattice_MC
simulationpmt=simulation_parameters.pmt
isotherm=isotherm.pressures



#************************************ check that all files are present *********************

# check that executables are present
if [ ! -f $executable1 ]
then
echo "";echo " cannot find executable $executable1 ";echo "";exit
fi
if [ ! -f $executable2 ]
then
echo "";echo " cannot find executable $executable2 ";echo "";exit
fi
if [ ! -f $executable3 ]
then
echo "";echo " cannot find executable $executable3 ";echo "";exit
fi
# check that simulation parameter file is present
if [ ! -f $simulationpmt ]
then
echo "";echo " cannot find simulation parameters file $simulationpmt ";echo "";exit
fi

# command line arguments
if (( $# != 2 )) 
then
echo ""; echo "  must have two command line arguments present, the name of the solute and"
echo "  the name of the crystal framework.  So if the name of the .conf file is "
echo "  co2.conf and the name of the .pmt file is co2.pmt, and the name of the "
echo "  .fconf file is mof5.fconf, run the script using './run_lattice.sh co2 mof5' ";echo ""
exit
fi

# base file names
solute=$1;framework=$2

conffile=empty;fconffile=empty;pmtfile=empty;symfile=empty;cgpot=empty;

#find all file names
for file in *; 
do 
if [[ ( $file == *.conf ) && ( $file == ${solute}* ) ]]
then
conffile=$file
fi
if [[ ( $file == *.pmt ) && ( $file == ${solute}* ) ]]
then
pmtfile=$file
fi
if [[ ( $file == *.fconf ) && ( $file == ${framework}* ) ]]
then
fconffile=$file
fi
if [[  $file == *.sym  ]]
then
symfile=$file
fi
if [[  $file == *.xvg  ]]
then
cgpot=$file
fi
done

# make sure we have all the files

if [[ $conffile == empty ]] 
then
echo ""; echo " cannot find .conf file ${solute}*.conf " ; echo "";exit
fi
if [[ $pmtfile == empty ]] 
then
echo ""; echo " cannot find .pmt file ${solute}*.pmt " ; echo "";exit
fi
if [[ $fconffile == empty ]] 
then
echo ""; echo " cannot find .fconf file ${framework}*.fconf " ; echo "";exit
fi
if [[ $symfile == empty ]] 
then
echo ""; echo " cannot find .sym file " ; echo "";exit
fi
if [[ $cgpot == empty ]] 
then
echo ""; echo " cannot find .xvg coarse grained potential file " ; echo "";exit
fi

cglattice="$(echo $cgpot | awk -F'.' '{print $1}').txt"
echo $cglattice

#*************************** Discretize the coarse grained potential ***********************
# get grid size from simulation_parameters.pmt file , also get n_threads for later
n_threads=1  #default
ngrida=-1;ngridb=-1;ngridc=-1
while read line; 
do flag="$(echo $line | awk '{print $1}')"; 
if [[ $flag == cav_grid_a ]]
then
ngrida="$(echo $line | awk '{print $2}')";
elif [[ $flag == cav_grid_b ]]
then
ngridb="$(echo $line | awk '{print $2}')";
elif [[ $flag == cav_grid_c ]]
then
ngridc="$(echo $line | awk '{print $2}')";
elif [[ $flag == n_threads ]]
then
n_threads="$(echo $line | awk '{print $2}')";
fi
done < $simulationpmt

# make sure grid size was found
if [ $ngrida == -1 ] || [ $ngridb == -1 ] || [ $ngridc == -1 ] 
then
echo " couldn't find grid size in $simulationpmt file "; exit
fi

echo " Discretizing coarse grained potential ..."
echo ""
echo " Found a lattice grid size of"
echo " $ngrida  $ngridb  $ngridc"
echo " in $simulationpmt file"
echo " using this grid for cg potential"
echo ""

#run coarse grain lattice grid
./${executable2} $fconffile $cgpot $ngrida $ngridb $ngridc > $cglattice

echo "...Discretization complete" 

#************************* Create the free energy lattice grid **********************
ofile="${solute}_${framework}_lattice_grid"
TARGET="${ofile}.sh"

cat >> $TARGET << EOF
#!/bin/bash
#PBS -N $ofile -l nodes=1:ppn=${n_threads}
export OMP_STACKSIZE=10M
 echo "started \`date\`"
./${executable1} ${conffile} ${fconffile} ${pmtfile} ${simulationpmt} ${ofile} ${symfile}
 echo "finished \`date\`"
EOF

echo ""
echo " submitting and running the lattice free energy grid calculation..."
qsub -d `pwd` $TARGET

