!*********************************************
! this module implements the Qeq charge equilibration
! method for periodic systems.
!
! originial paper : Rappe,A.K.,Goddard,W.A.III, J.Phys.Chem. 1991,95,3358-3363
! w/ PBC          : Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323
! new parameters  : Kadantsev et al., J.Phys.Chem.Lett.2013,4,3056-3061
!
! Note that since the periodic Qeq method relies on an Ewald sum (PME) to compute the
! interaction matrix, all PME settings (grid size, spline order, etc) are relevant
! in determining the charges
!**********************************************

module Qeq

  integer, parameter :: N_element=96  ! currently, have Qeq parameters for 96 elements
  real*8,dimension(N_element,3) :: Qeq_element_parameters
  character(5),dimension(N_element) :: atomic_number_element

  ! this is the scaling factor defined in Rappe,A.K.,Goddard,W.A.III, J.Phys.Chem. 1991,95,3358-3363
  ! such that exp = lambda*(2n+1)/(2R), and in this reference lambda=.5
  real*8,parameter :: Qeq_lambda=0.5d0

  !************ this should always be set to "yes", unless testing.  If not set to yes, testing will be assumed,
  ! and code will stop after charges are equilibrated
  character(3),parameter :: periodic_Qeq="yes"

  !*****this allows for the parameters of Hydrogen to be charge dependent, as in the original references
  character(3),parameter :: dynamic_hydrogen_param="yes"

  !*****this signals whether to use the parameters in Kadantsev et al., J.Phys.Chem.Lett.2013,4,3056-3061
  character(3),parameter :: Woo_parameters="yes"

  !********* this parameter controls what type of electrostatic screening we use for the real space contribution to
  ! the Jij matrix.  Options are:
  ! 1) "Slater1" is non-standard.
  ! the Slater orbitals are not squared, and the appropriate normalization is used
  ! 
  ! 2) "Slater2" is the standard screening, as implemented in GULP and also Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323
  ! where the Slater orbitals are squared to create the electron density, and the appropriate normalization is used
  character(7) :: real_Jij_screentype="Slater2"

  !***** save this reciprocal space potential for one reference atom, and shift it for different atoms
  real*8,dimension(:,:,:),allocatable  :: recip_pot_ref
  real*8,dimension(3) :: xyz_ref

  !***** save this array as it will be used during iterations for hydrogen atoms
  real*8,dimension(:,:),allocatable:: Jrecip_store

  !***** this determines whether the principle quantum number for elements is given by their row
  ! in the periodic table (.false.) ,consistent with GULP implementation
  ! or depends on the d, f orbital filling (.true.)
  ! setting d_valence=.true. gives the same exponents used by 
  ! Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323
  ! with exceptions being atomic numbers 55-58,64,71,>89
  logical,parameter :: d_valence =.false.

  real*8, parameter :: scf_frac=.25  ! this is damping fraction for iterative convergence
  integer,parameter :: maxit=20
  real*8, parameter :: qeq_thresh = 1.0d-6

contains


  !*********************************************
  ! this subroutine optimizes charges using the Qeq method,
  ! given a periodic unit cell
  !
  ! Following Kadantsev et al., J.Phys.Chem.Lett.2013,4,3056-3061 , we write the linear equations
  !
  ! (J  1) (Q     )  =  (-kai)     where the interaction matrix, J, has a charge-charge long-range
  ! (1  0) (lambda)     ( 0  )     contribution calculated using Ewald, and a short range
  !                                contribution from the Slater type orbital smeared charges
  !                                which is calculated in read space
  !
  ! For now, no symmetry is considered
  ! Also PME is used to calculate the interaction matrix, which is probably inefficient
  ! without symmetry (Ewald might be faster for 2 atoms ), but this is done in anticipation
  ! of symmetry implementation, in which case there might be a lot of atoms of the same symmetry
  ! and then PME should be faster than Ewald
  !
  ! The data structures for this subroutine are analogous to 
  ! the data structures used throughout the rest of the code.
  ! 
  ! We repeat the data structure convention here
  ! The system is divided into solute atoms , and fixed framework atoms.  These are distinguished in 
  ! data structures by the location.  All solute "molecules" are indexed with values 1 =< index =< n_mole
  ! while framework "molecules" are indexed with values n_mole < index =< tot_n_mole .  All framework 
  ! atoms are considered as individual "molecules" .  In this subroutine, we are only concerned with 
  ! the framework atoms, and so we only consider these latter indices.  The array n_atom, gives the
  ! number of atoms for each molecule, for n_mole < index =< tot_n_mole, n_atom(index)=1.  The
  ! cartesian coordinates of the framework are thus stored in  xyz(index,1,:), for 
  ! n_mole < index =< tot_n_mole
  !*********************************************
  subroutine Periodic_Qeq_solver(chg, alist, n_mole, tot_n_mole, n_atom, xyz, box, dfti_desc, dfti_desc_inv )
    use mkl95_lapack
    use global_variables
    use MKL_DFTI
    real*8, dimension(:,:), intent(inout) :: chg
    character(*),dimension(:,:),intent(in) :: alist
    integer, intent(in) :: n_mole, tot_n_mole
    integer,dimension(:),intent(in) :: n_atom
    real*8,dimension(:,:,:),intent(in) :: xyz
    real*8,dimension(:,:),intent(in) :: box
    TYPE(DFTI_DESCRIPTOR), pointer,intent(in):: dfti_desc,dfti_desc_inv

    integer :: n_framework_atoms,i_mole,i_atom, j_atom, index, ni, atomic_number_i,iteration, flag_converged
    real*8            :: dq, sum
    real*8, dimension(3) :: Qeq_param_i
    real*8, dimension(:,:), allocatable :: Jmatrix , Gmatrix
    real*8, dimension(:), allocatable :: charge_local, xvec
    integer,dimension(:),allocatable :: ipiv


    ! initialize Qeq parameters
    call initialize_Qeq_parameters

    n_framework_atoms = tot_n_mole - n_mole

    allocate(Jmatrix(n_framework_atoms,n_framework_atoms), Jrecip_store(n_framework_atoms,n_framework_atoms), charge_local(n_framework_atoms+1), Gmatrix(n_framework_atoms+1,n_framework_atoms+1),xvec(n_framework_atoms+1),ipiv(n_framework_atoms+1) )


    !*********************** iteratively solve Qeq equations, this is necessary for hydrogen atoms
    flag_converged=0
    do iteration=1,maxit

       ! construct J matrix, this will be in atomic units (although we do some converting back and forth to construct it)
       call construct_Jmatrix_periodic(Jmatrix,alist, n_mole, tot_n_mole, n_atom, xyz, box, chg, dfti_desc,dfti_desc_inv)

       ! construct G matrix and xvec
       do i_mole=n_mole+1,tot_n_mole
          index= i_mole - n_mole
          call get_Qeq_parameters( alist, i_mole, Qeq_param_i, ni, atomic_number_i )
          ! electronegativity is in the 2nd index of Qeq_param_i array, should be in a.u.
          xvec(index)= -Qeq_param_i(2)
       enddo

       ! construct G matrix
       do i_atom=1,n_framework_atoms
          do j_atom=1,n_framework_atoms
             Gmatrix(i_atom,j_atom) = Jmatrix(i_atom,j_atom)
          enddo
       enddo


       ! now ones and zeros
       xvec(n_framework_atoms+1)=0d0
       Gmatrix(n_framework_atoms+1,:)=1d0
       Gmatrix(:,n_framework_atoms+1)=1d0
       Gmatrix(n_framework_atoms+1,n_framework_atoms+1)=0d0

       ! LU factorization of matrix
       call DGETRF_MKL95(Gmatrix,ipiv)
       ! inverse of LU matrix
       call DGETRI_MKL95(Gmatrix,ipiv)  

       ! optimized charges
       charge_local = matmul(Gmatrix,xvec)


       ! update chgs in main data structure, see if iterations have converged
       sum=0d0
       do i_atom=1, n_framework_atoms
          index = n_mole + i_atom
          dq = charge_local(i_atom) - chg(index,1)
          sum = sum + abs(dq)
          ! damp change for convergence
          chg(index,1) = charge_local(i_atom) - scf_frac * dq
       enddo

       sum=sum/dble(n_framework_atoms)
       ! see if converged
       if ( sum < qeq_thresh ) then
          flag_converged=1
          exit
       endif

    enddo

    ! write Qeq output
    call write_Qeq_output(flag_converged, iteration, n_framework_atoms, n_mole, alist, chg )


    !******** deallocate arrays
    deallocate(Jrecip_store)

  end subroutine Periodic_Qeq_solver




  !***********************************************
  ! this subroutine constructs the J interaction matrix
  ! as in Kadantsev et al., J.Phys.Chem.Lett.2013,4,3056-3061
  !
  ! in words, the Jij element of the interaction matrix, is (for i/=j)
  ! the interaction of the charge of atom i with the potential of the periodic
  ! system of charges of atom j.
  ! 
  ! for real space interactions, we consider smeared slater orbital interactions,
  ! while for reciprocal space we consider point charge-gaussian interactions as usual
  !***********************************************
  subroutine construct_Jmatrix_periodic(Jmatrix,alist, n_mole, tot_n_mole, n_atom, xyz, box, chg, dfti_desc,dfti_desc_inv)
    use global_variables
    use MKL_DFTI
    real*8, dimension(:,:), intent(inout) :: Jmatrix
    real*8, dimension(:,:),intent(in) :: chg
    character(*),dimension(:,:),intent(in) :: alist
    integer, intent(in) :: n_mole, tot_n_mole
    integer,dimension(:),intent(in) :: n_atom
    real*8,dimension(:,:,:),intent(in) :: xyz
    real*8,dimension(:,:),intent(in) :: box
    TYPE(DFTI_DESCRIPTOR), pointer,intent(in):: dfti_desc,dfti_desc_inv

    integer :: i_mole, i_atom, j_mole, j_atom, index1, index2, ni, nj, atomic_number_i, atomic_number_j
    real*8 :: Jji_recip, Jji_real
    real*8, dimension(3) :: Qeq_param_i , Qeq_param_j
    integer, save :: initialize=0


    !*********************** We only need to compute the Jij matrix for non-hydrogen atoms once.  If this is not the first call to this subroutine, then we only need to consider hydrogen atoms
    if ( initialize == 0 ) then

       ! loop over all framework atoms, therefore loop starts at n_mole+1
       do i_mole=n_mole+1,tot_n_mole
          index1= i_mole - n_mole

          do j_mole=i_mole, tot_n_mole
             index2 = j_mole - n_mole

             ! get the Qeq parameters for this atom type, also principle quantum numbers of valence shell
             call get_Qeq_parameters( alist, i_mole, Qeq_param_i, ni, atomic_number_i )
             call get_Qeq_parameters( alist, j_mole, Qeq_param_j, nj, atomic_number_j )

             ! if either of these atoms are Hydrogen, we could potentially allow them to be charge dependent
             Select Case(dynamic_hydrogen_param)
             Case("yes")
                Select Case(atomic_number_i)
                Case(1)
                   call set_hydrogen_dynamic_Qeq( Qeq_param_i , chg(i_mole,1) )
                End Select
                Select Case(atomic_number_j)
                Case(1)
                   call set_hydrogen_dynamic_Qeq( Qeq_param_j , chg(j_mole,1) )
                End Select
             End Select


             ! compute the reciprocal space contribution to the interaction matrix, note the inversion of the indices here, Jji is the interaction of atom j with the periodic distribution of atom i
             ! this will be output in atomic units
             call construct_pairwise_interaction_recip( Jji_recip , j_mole, i_mole, box, xyz , dfti_desc , dfti_desc_inv)
             call construct_pairwise_interaction_real( Jji_real , box, j_mole, Qeq_param_j, nj, i_mole, Qeq_param_i, ni , xyz)

             ! if testing, and don't want periodic contribution, zero this
             Select Case(periodic_Qeq)
                Case("no")
                   Jji_recip = 0d0
             End Select

             ! store the reciprocal space contribution
             Jrecip_store(index2,index1) = Jji_recip
             Jrecip_store(index1,index2) = Jji_recip

             Jmatrix(index2,index1) = Jji_recip + Jji_real
             Jmatrix(index1,index2) = Jji_recip + Jji_real

          enddo
       enddo


       !***************************** this is not the first iteration.  Here, we only need to update hydrogen 
    else
       do i_mole=n_mole+1,tot_n_mole
          index1= i_mole - n_mole
          do j_mole=i_mole, tot_n_mole
             index2 = j_mole - n_mole

             ! get the Qeq parameters for this atom type, also principle quantum numbers of valence shell
             call get_Qeq_parameters( alist, i_mole, Qeq_param_i, ni, atomic_number_i )
             call get_Qeq_parameters( alist, j_mole, Qeq_param_j, nj, atomic_number_j )

             !******** if either of these atoms are hydrogen
             if ( (atomic_number_i == 1 ) .or. (atomic_number_j == 1 ) ) then

                Select Case(dynamic_hydrogen_param)
                Case("yes")
                   Select Case(atomic_number_i)
                   Case(1)
                      call set_hydrogen_dynamic_Qeq( Qeq_param_i , chg(i_mole,1) )
                   End Select
                   Select Case(atomic_number_j)
                   Case(1)
                      call set_hydrogen_dynamic_Qeq( Qeq_param_j , chg(j_mole,1) )
                   End Select
                End Select

                call construct_pairwise_interaction_real( Jji_real , box, j_mole, Qeq_param_j, nj, i_mole, Qeq_param_i, ni , xyz)

                Jmatrix(index2,index1) = Jrecip_store(index2,index1) + Jji_real
                Jmatrix(index1,index2) = Jrecip_store(index1,index2) + Jji_real

             endif
          enddo
       enddo
    end if

    initialize=1

  end subroutine construct_Jmatrix_periodic




  !*********************************
  ! this subroutine constructs the Jij pairwise reciprocal space interaction, 
  ! due to the periodic images of atom j, use the pme algorithm to construct the potential
  ! this will output in atomic units
  !
  ! it is assumed that the first molecule index (j_mole) is the molecule which interactions with the 
  ! potential, while the second index is the atom whose periodic images generated the potential
  !
  ! it is also assumed that we are only dealing with framework atoms, so that the only meaningful
  ! indices of the xyz array are for the second index, i_atom=1
  !*********************************
  subroutine construct_pairwise_interaction_recip(Jji, j_mole, i_mole, box, xyz, dfti_desc , dfti_desc_inv)
    use global_variables
    use MKL_DFTI
    use pme_routines
    use routines
    real*8, intent(out)  :: Jji
    integer, intent(in) :: j_mole, i_mole
    real*8,dimension(:,:), intent(in) :: box
    real*8,dimension(:,:,:),intent(in) :: xyz
    TYPE(DFTI_DESCRIPTOR), pointer,intent(in):: dfti_desc,dfti_desc_inv


    integer,save :: initialize=0, ref_atom
    !******************** local data structures to input to the pme routines
    real*8,dimension(1,1,3)  :: xyz_local
    real*8,dimension(1,1)  :: charge_local
    integer,dimension(1)   :: n_atom_local
    integer :: tot_n_mole_local, n_mole_local
    !***********************************************************************

    real*8, dimension(-lgrg_order+1:lgrg_order,-lgrg_order+1:lgrg_order,-lgrg_order+1:lgrg_order) :: weight
    real*8, dimension(3) :: rtmp, r_ij, shift, xyz_shift
    real*8, parameter :: convert_bohr_ang = 1.889726
    real*8 :: ra, rb, rc, norm_dr
    integer :: ia0 , ib0, ic0, ia, ib, ic, igrida, igridb, igridc

    Jji =0d0


    !*************************** if this is first call , construct the reciprocal space potential for the periodic system based on the target atom
    Select Case(initialize)
    Case(0)
       ! get reciprocal space potential from periodic charge distribution of this atom, note we use a unit charge to construct the J matrix
       ! set up local data structures
       xyz_local(1,1,:) = xyz(i_mole,1,:)
       charge_local(1,1) = 1d0 ! unit charge
       n_atom_local=1;tot_n_mole_local=1;n_mole_local=0

       allocate ( recip_pot_ref(pme_grid,pme_grid,pme_grid) )
       recip_pot_ref=0d0
       call recip_pme_grid( box, tot_n_mole_local, n_mole_local, n_atom_local, xyz_local, charge_local , dfti_desc , dfti_desc_inv , recip_pot_ref, 1 )

       ref_atom = i_mole
       xyz_ref(:) = xyz(i_mole,1,:)
       xyz_shift(:) = xyz(j_mole,1,:)
    Case(1)

       ! here we have to shift the reference potential grid by the displacement from the reference atom to this atom, this is equivalent to shifting the position of the atom
       ! experiencing the electrostatic potential
       xyz_shift(:) = xyz(j_mole,1,:) - ( xyz(i_mole,1,:) -  xyz_ref(:) )
    End Select

    initialize=1



    ! evaluate the interaction of the charge with the reciprocal space potential.  Here, we use
    ! Lagrangian interpolation.  This code has been duplicated from the subroutine pme_usegrid
    ! we would rather not duplicate code, but there was no way around this without messing with the 
    ! pme_usegrid subroutine, which i didn't want to do
    rtmp(:) = matmul(xyz_shift,inv_box(:,:))
    ra = rtmp(1) * pme_grid
    rb = rtmp(2) * pme_grid
    rc = rtmp(3) * pme_grid
    !     shift into the box
    ra = ra - floor(ra/pme_grid)*pme_grid
    rb = rb - floor(rb/pme_grid)*pme_grid
    rc = rc - floor(rc/pme_grid)*pme_grid
    weight = 1.0d0
    ia0 = floor(ra)
    ib0 = floor(rb)
    ic0 = floor(rc)
    do ia = -lgrg_order+1, lgrg_order
       weight( ia,:,: ) = weight( ia,:,: ) * wfun( ia - (ra-ia0) )
    enddo
    do ib = -lgrg_order+1, lgrg_order
       weight( :,ib,: ) = weight( :,ib,: ) * wfun( ib - (rb-ib0) )
    enddo
    do ic = -lgrg_order+1, lgrg_order
       weight( :,:,ic ) = weight( :,:,ic ) * wfun( ic - (rc-ic0) )
    enddo
    !     interpolate the reciprocal potential grid
    do ia = -lgrg_order+1, lgrg_order
       igrida = ia0 + ia
       igrida = igrida - floor(dble(igrida)/dble(pme_grid))*pme_grid
       igrida = igrida + 1
       do ib = -lgrg_order+1, lgrg_order
          igridb = ib0 + ib
          igridb = igridb - floor(dble(igridb)/dble(pme_grid))*pme_grid
          igridb = igridb + 1
          do ic = -lgrg_order+1, lgrg_order
             igridc = ic0 + ic
             igridc = igridc - floor(dble(igridc)/dble(pme_grid))*pme_grid
             igridc = igridc + 1
             Jji = Jji + weight(ia,ib,ic)*recip_pot_ref(igrida,igridb,igridc) 

          enddo
       enddo
    enddo

    ! now convert to atomic units, right now Jji is in Ang^-1
    Jji = Jji / convert_bohr_ang


    ! now subtract off the self interaction, or closest pair interaction, which will be treated in real space
    ! this is different depending on whether i_mole = j_mole

    !***************************** i_mole /= j_mole **************************
    if ( i_mole /= j_mole ) then

       ! subtract off the interaction between closest atoms, which we treat in real space

       shift = pbc_shift( xyz(i_mole,1,:),xyz(j_mole,1,:), box   )
       r_ij = pbc_dr( xyz(i_mole,1,:) , xyz(j_mole,1,:), shift )
       norm_dr = sqrt( dot_product( r_ij, r_ij ) )
       ! remove reciprocal space term, this is just error function
       ! we have the complementary error function, so that's why there's a -1
       Jji = Jji + (erfc_g(norm_dr*alpha_sqrt)-1.D0) / norm_dr / convert_bohr_ang   ! Jji in a.u.


       !***************************** i_mole == j_mole **************************
    else
       ! remove reciprocal space self-self interaction, 
       ! NOTE that this is different from ewald_self energy by a factor of 2
       ! because this is a full correction for the interaction of a point
       ! charge with its image Gaussian.  The standard Ewald sum corrects for
       ! double counting, and therefore the contribution to the  Ewald self energy
       ! is half of this interaction.
       !
       ! alpha_sqrt is in inverse Angstroms, convert to Bohr
       Jji = Jji - 2d0 * alpha_sqrt / sqrt(pi)  / convert_bohr_ang   ! Jji in a.u.

    endif


  end subroutine construct_pairwise_interaction_recip



  !*************************************
  ! this subroutine calculates the real space contribution
  ! to the J matrix.  For non-identical atoms, this is a 
  ! coulomb integral over slater orbitals
  !*************************************
  subroutine construct_pairwise_interaction_real(Jji, box, j_mole, Qeq_param_j, nj,  i_mole, Qeq_param_i, ni, xyz)
    use global_variables
    use routines
    real*8, intent(out)  :: Jji
    real*8,dimension(:,:), intent(in) :: box
    integer, intent(in) :: j_mole, i_mole, nj, ni
    real*8, dimension(:), intent(in) :: Qeq_param_j, Qeq_param_i
    real*8,dimension(:,:,:),intent(in) :: xyz

    real*8, parameter :: convert_bohr_ang = 1.889726
    real*8, dimension(3) :: r_ij, shift
    real*8 ::  lambda_i, lambda_j, Jij_slater, norm_dr


    !***************************** i_mole /= j_mole **************************
    if ( i_mole /= j_mole ) then

       ! now we have to consider the interaction between closest atoms, for which we use smeared out slater-type functions

       ! slater exponents are the 3rd element of the
       ! Qeq_param_i, Qeq_param_j arrays, note that these exponents are input in atomic units, our geometries are in 
       ! angstrom, so convert exponents to inverse angstrom
       lambda_i = Qeq_param_i(3) * convert_bohr_ang
       lambda_j = Qeq_param_j(3) * convert_bohr_ang

       shift = pbc_shift( xyz(i_mole,1,:),xyz(j_mole,1,:), box   )
       r_ij = pbc_dr( xyz(i_mole,1,:) , xyz(j_mole,1,:), shift )
       norm_dr = sqrt( dot_product( r_ij, r_ij ) )

       ! there are different formulations for the real space Coulomb screening.  See comments at the top of the module 
       Select Case(real_Jij_screentype)
       Case("Slater1")
          call Slater_orbital_integral( Jij_slater, lambda_i , ni, lambda_j , nj, norm_dr)
       Case("Slater2")
          call Slater_repulsion_integral( Jij_slater, lambda_i , ni, lambda_j , nj, norm_dr)
       case default
          write(*,*) "option ", real_Jij_screentype, " for parameter real_Jij_screentype is not recognized"
          stop
       End Select



       ! Jij_slater is output in Ang^-1

       Jji = Jij_slater / convert_bohr_ang   ! Jji in a.u.

       !***************************** i_mole == j_mole **************************
    else
       ! this is the idempotential, this is in atomic units
       Jji = Qeq_param_i(1) 

    endif



  end subroutine construct_pairwise_interaction_real


  !*****************************************
  !  for two slater orbitals, u(a)=Na*r^(m-1)*exp(-alpha*r) , u(b)
  !  this evaluates the integral
  !  integral( u(a) * u(b) /r12 , dr1*dr2 )
  !  NOTICE no squares!
  !
  !  this integral is given in terms of the integral evaluated
  !  in the Slater_repulsion_integral subroutine
  !  if ni, nj are odd, then this is explicitly given in terms of 
  !  a Slater repulsion integral, with a variable transformation, and 
  !  appropriate normalization
  !
  !  if ni and(or) nj is even, then we have to take a derivative w.r.t. one or both
  !  of the exponents, to get a new variable transformation that ensures that the 
  !  transformed mi,mj principle quantum numbers are integers
  !
  !***************************************** 
  subroutine Slater_orbital_integral( Jij, alpha_i , ni, alpha_j , nj, norm_dr)
    use global_variables
    real*8,intent(out) :: Jij
    real*8,intent(in) :: alpha_i, alpha_j, norm_dr
    integer, intent(in) :: ni, nj

    real*8 :: beta_i, beta_j, JSij, dJSij, dJiSij, dJjSij, d2JSij
    integer :: mi, mj
    real*8 :: NormS2_i, NormS2_j, dNormS_i, dNormS_j, Norm_i, Norm_j

    ! exponent transformation
    beta_i=alpha_i/2d0
    beta_j=alpha_j/2d0

    ! factors Norm are normalization factors for the orbitals not squared
    Norm_i = alpha_i**(ni+2) / ( 4d0 * pi * factorial(ni+1) )
    Norm_j = alpha_j**(nj+2) / ( 4d0 * pi * factorial(nj+1) )

    ! ni , nj both odd
    if ( ( mod(ni,2) == 1 ) .and. ( mod(nj,2) == 1 ) ) then
       ! principle quantum number transformation
       mi=(ni+1)/2
       mj=(nj+1)/2
       call Slater_repulsion_integral( JSij , beta_i , mi , beta_j , mj , norm_dr)

       ! Normalization factors NormS2 are squared normalization factors for the square of the orbitals
       NormS2_i = NormS2( beta_i , mi )
       NormS2_j = NormS2( beta_j , mj )

       Jij = JSij * Norm_i * Norm_j / NormS2_i / NormS2_j

       ! ni even, nj odd, write this in terms of a derivative w.r.t. alpha_i
    else if ( ( mod(ni,2) == 0 ) .and. ( mod(nj,2) == 1 ) ) then
       mi=ni/2
       mj=(nj+1)/2

       call Slater_repulsion_integral( JSij , beta_i , mi , beta_j , mj , norm_dr)
       call dSlater_repulsion_integral( dJSij , beta_i , mi , beta_j , mj , norm_dr, 1)

       ! Normalization factors NormS2 are squared normalization factors for the square of the orbitals
       NormS2_i = NormS2( beta_i , mi )
       NormS2_j = NormS2( beta_j , mj )
       ! this is the derivative of the unsquared normalization factor for the square of the orbitals, w.r.t its exponent
       dNormS_i = dNormS( beta_i , mi )

       ! the 0.5d0 is dbeta/dalpha
       Jij = Norm_i * Norm_j / NormS2_i / NormS2_j * 0.5d0 * ( 2d0 * JSij * dNormS_i / dsqrt( NormS2_i )  - dJSij )

       ! ni odd, nj even, write this in terms of a derivative w.r.t. alpha_j
    else if ( ( mod(ni,2) == 1 ) .and. ( mod(nj,2) == 0 ) ) then
       mi=(ni+1)/2
       mj=nj/2

       call Slater_repulsion_integral( JSij , beta_i , mi , beta_j , mj , norm_dr)
       call dSlater_repulsion_integral( dJSij , beta_i , mi , beta_j , mj , norm_dr, 2)

       ! Normalization factors NormS2 are squared normalization factors for the square of the orbitals
       NormS2_i = NormS2( beta_i , mi )
       NormS2_j = NormS2( beta_j , mj )
       ! this is the derivative of the unsquared normalization factor for the square of the orbitals, w.r.t its exponent
       dNormS_j = dNormS( beta_j , mj )

       ! the 0.5d0 is dbeta/dalpha
       Jij = Norm_i * Norm_j / NormS2_i / NormS2_j * 0.5d0 * ( 2d0 * JSij * dNormS_j / dsqrt( NormS2_j )  - dJSij )

       ! both ni, nj even , need two derivatives
    else
       mi=ni/2
       mj=nj/2

       call Slater_repulsion_integral( JSij , beta_i , mi , beta_j , mj , norm_dr)
       call dSlater_repulsion_integral( dJiSij , beta_i , mi , beta_j , mj , norm_dr, 1)
       call dSlater_repulsion_integral( dJjSij , beta_i , mi , beta_j , mj , norm_dr, 2)
       call d2Slater_repulsion_integral( d2JSij , beta_i , mi , beta_j , mj , norm_dr)

       ! Normalization factors NormS2 are squared normalization factors for the square of the orbitals
       NormS2_i = NormS2( beta_i , mi )
       NormS2_j = NormS2( beta_j , mj )
       ! this is the derivative of the unsquared normalization factor for the square of the orbitals, w.r.t its exponent
       dNormS_i = dNormS( beta_i , mi )
       dNormS_j = dNormS( beta_j , mj )

       Jij = Norm_i * Norm_j / NormS2_i / NormS2_j * 0.5d0 * 0.5d0 * ( d2JSij - 2d0 * dJiSij * dNormS_j / dsqrt( NormS2_j ) - 2d0 * dJjSij * dNormS_i / dsqrt( NormS2_i ) + 4d0 * JSij * dNormS_i / dsqrt( NormS2_i ) * dNormS_j / dsqrt( NormS2_j ) )

    endif


  end subroutine Slater_orbital_integral


  !*****************************************
  !  for two slater orbitals, u(a)=Na*r^(m-1)*exp(-alpha*r) , u(b)
  !  this evaluates the integral
  !  integral( |u(a)|^2*|u(b)|^2/r12 , dr1*dr2 )
  !
  !  This integral is evaluated using the formula from Rosen, N. Phys Rev. 38, 1931, 255
  !
  !  there are two different formulas for integrals, namely one for alpha_i = alpha_j and one
  !  for alpha_i /= alpha_j
  !
  !  Note that these formula are not symmetric in the slater orbitals.  It is assumed
  !  that alpha > beta , m is the principle quantum number corresponding to the alpha orbital
  !  and n is the principle quantum number corresponding to the beta orbital
  !
  !  ****************** other Qeq implementations ***************
  !  This is the integral used for the Jij term in 
  !  Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323
  !  as well as in the GULP implementation
  !
  !
  !*****************************************
  subroutine Slater_repulsion_integral( Jij, alpha_i , ni, alpha_j , nj, norm_dr)
    real*8,intent(out) :: Jij
    real*8,intent(in) :: alpha_i, alpha_j, norm_dr
    integer, intent(in) :: ni, nj

    real*8 :: alpha, beta
    integer :: m , n
    real*8 :: term1, term2, arg1, arg2, arg3
    real*8 :: sum1,sum2,factor
    integer :: i,j,index, arg4
    real*8,parameter :: small=1D-4

    ! our formula for Bn assumes that ni, nj are less than equal to 7
    ! this could not be the case if this subroutine is called from Slater_orbital_integral
    if ( ( ni > 7 ) .or. (  nj > 7 ) ) then
       write(*,*) "Slater_repulsion_integral subroutine is called with exponents"
       write(*,*) " > 7, this might be the case if real_Jij_screentype='Slater1' "
       write(*,*) " and 5th row or higher elements are used "
       write(*,*) " please change this setting "
       stop
   end if

    ! order slater orbitals by exponents
    if ( alpha_i > alpha_j ) then
       alpha = alpha_i
       beta = alpha_j
       m = ni
       n = nj
    else
       alpha = alpha_j
       beta = alpha_i
       m = nj
       n = ni
    endif


    arg1 = 2d0*alpha*norm_dr

    ! term1 is the same for both formulas
    term1 = 1d0/norm_dr + arg1**(2*m)/(factorial(2*m)*norm_dr) * ( (arg1-2d0*m)*An_plus( (2*m-1), arg1) - exp(-arg1))

    ! if alpha /= beta
    if ( abs( alpha - beta ) > small ) then
       sum1=0d0
       do i=0,2*n
          factor=dble(2*n-i)/factorial(i) *(beta*norm_dr)**i
          sum2=0d0
          index=2*m-1+i
          do j=0, index
             arg2=norm_dr*(alpha-beta)
             arg3=norm_dr*(alpha+beta)
             sum2=sum2+Dpmn(j, 2*m-1,i) * Bn(j,arg2) * An_plus(index-j,arg3)
          enddo
          sum1=sum1+factor*sum2
       enddo

       term2= alpha*(alpha*norm_dr)**(2*m)/2d0/factorial(2*m)/dble(n) * sum1

    else
       ! if alpha == beta
       sum1=0d0
       do i=0,2*n-1
          factor=dble(2*n-i)/factorial(i) *(beta*norm_dr)**i
          sum2=0d0
          ! note integer arithmetic here
          index= (2*m+i-1)/2
          do j=0, index
             arg4 = 2*m+i-1-2*j
             sum2=sum2 + Dpmn(2*j, 2*m-1,i)/dble(2*j+1)* An_plus( arg4 , arg1 )
          enddo
          sum1=sum1+factor*sum2
       enddo

       term2= alpha*(alpha*norm_dr)**(2*m)/factorial(2*m)/dble(n) * sum1
    end if

    Jij = term1 - term2

  end subroutine Slater_repulsion_integral




 !***************************************
 ! this is the derivative of the subroutine Slater_repulsion_integral
 ! with respect to either of the exponents, depending on the value of the input
 ! flag index
 !***************************************
  subroutine dSlater_repulsion_integral( dJij, alpha_i , ni, alpha_j , nj, norm_dr , index_derivative )
    real*8,intent(out) :: dJij
    real*8,intent(in) :: alpha_i, alpha_j, norm_dr
    integer, intent(in) :: ni, nj, index_derivative

    real*8 :: alpha, beta
    integer :: m , n , dlarge
    real*8 :: term1, dterm1, dterm2, arg1, arg2, arg3
    real*8 :: sum1,sum2,dsum1,dsum2,factor, dfactor
    integer :: i,j,index, arg4
    real*8,parameter :: small=1D-4

    ! figure out whether we are taking the derivative with respect to the smaller (dlarge=0) or larger exponent (dlarge=1)

    ! order slater orbitals by exponents
    if ( alpha_i > alpha_j ) then
       alpha = alpha_i
       beta = alpha_j
       m = ni
       n = nj

       if ( index_derivative == 1 ) then
          dlarge=1
       else
          dlarge=0
       endif
    else
       alpha = alpha_j
       beta = alpha_i
       m = nj
       n = ni

       if ( index_derivative == 1 ) then
          dlarge=0
       else
          dlarge=1
       endif
    endif


    !******************* for derivative, use corresponding formulas for Slater_repulsion_integral depending on magnitude of exponents

    ! first do dterm1/dalpha , where term1 is defined as in the Slater_repulsion_integral subroutine
    arg1 = 2d0*alpha*norm_dr

    dterm1 =  dble(2*m) * arg1**(2*m-1)/(factorial(2*m)*norm_dr) * ( (arg1-2d0*m)*An_plus( (2*m-1), arg1) - exp(-arg1))     
    dterm1 = dterm1 + arg1**(2*m)/(factorial(2*m)*norm_dr) * (  An_plus( (2*m-1), arg1) + (arg1-2d0*m)* dAn_plus( (2*m-1), arg1) + exp(-arg1) )
    dterm1 = dterm1 * 2d0 * norm_dr

    ! if alpha /= beta
    if ( abs( alpha - beta ) > small ) then

       !********************** this is the derivative of the Slater repulsion integral for the large exponent (alpha)
       Select Case(dlarge)
       Case(1)

       sum1=0d0
       ! dsum1 is derivative of sum1
       dsum1=0d0
       do i=0,2*n
          factor=dble(2*n-i)/factorial(i) *(beta*norm_dr)**i
          sum2=0d0
          ! dsum2 is derivative of sum2
          dsum2=0d0
          index=2*m-1+i
          do j=0, index
             arg2=norm_dr*(alpha-beta)
             arg3=norm_dr*(alpha+beta)
             sum2=sum2+Dpmn(j, 2*m-1,i) * Bn(j,arg2) * An_plus(index-j,arg3)

             dsum2=dsum2+norm_dr*Dpmn(j, 2*m-1,i)*( dBn(j,arg2) * An_plus(index-j,arg3) + Bn(j,arg2) * dAn_plus(index-j,arg3) )
          enddo
          sum1=sum1+factor*sum2
          dsum1=dsum1+factor*dsum2
       enddo

       dterm2= alpha*(alpha*norm_dr)**(2*m)/2d0/factorial(2*m)/dble(n) * dsum1 + dble(2*m+1)*(alpha*norm_dr)**(2*m)/2d0/factorial(2*m)/dble(n)  * sum1

       dJij = dterm1 - dterm2

       Case(0)
       !********************** this is the derivative of the Slater repulsion integral for the small exponent (beta)

       sum1=0d0
       do i=0,2*n
          factor=dble(2*n-i)/factorial(i) *(beta*norm_dr)**i
          dfactor=factor * dble(i) / beta
          sum2=0d0
          ! dsum2 is derivative of sum2
          dsum2=0d0
          index=2*m-1+i
          do j=0, index
             arg2=norm_dr*(alpha-beta)
             arg3=norm_dr*(alpha+beta)
             sum2=sum2+Dpmn(j, 2*m-1,i) * Bn(j,arg2) * An_plus(index-j,arg3)

             dsum2=dsum2+norm_dr*Dpmn(j, 2*m-1,i)*( - dBn(j,arg2) * An_plus(index-j,arg3) + Bn(j,arg2) * dAn_plus(index-j,arg3) )
          enddo
          sum1=sum1 + factor*dsum2 + dfactor * sum2

       enddo

       dterm2= alpha*(alpha*norm_dr)**(2*m)/2d0/factorial(2*m)/dble(n) * sum1

       dJij = - dterm2

       End Select


    else
       ! if alpha == beta
       !************************************* NOTE **************************
       ! this is a little tricky.  Note here we are calculating limit(da-->0) [J(alpha+da,beta+da)-J(alpha,beta)]/da  at alpha=beta , 
       ! since this formula implicitly assumes alpha=beta
       ! this is equal to 2 * dJ(alpha,beta)/dalpha at alpha=beta
       dsum1=0d0
       sum1=0d0
       do i=0,2*n-1
          factor=dble(2*n-i)/factorial(i) *(beta*norm_dr)**i
          dfactor=factor * dble(i) / beta

          sum2=0d0
          dsum2=0d0
          ! note integer arithmetic here
          index= (2*m+i-1)/2
          do j=0, index
             arg4 = 2*m+i-1-2*j
             sum2=sum2 + Dpmn(2*j, 2*m-1,i)/dble(2*j+1)* An_plus( arg4 , arg1 )
             dsum2=dsum2 + Dpmn(2*j, 2*m-1,i)/dble(2*j+1)* dAn_plus( arg4 , arg1 ) * 2d0 * norm_dr
          enddo
          sum1=sum1+factor*sum2
          dsum1=dsum1+dfactor*sum2 + factor*dsum2
       enddo

       dterm2 = dble(2*m+1)*(alpha*norm_dr)**(2*m)/factorial(2*m)/dble(n) * sum1 + alpha*(alpha*norm_dr)**(2*m)/factorial(2*m)/dble(n) * dsum1

       dJij = dterm1 - dterm2

       ! divide by 2 here, per above comment
       dJij = dJij / 2d0

    endif


  end subroutine dSlater_repulsion_integral




 !***************************************
 ! this is the 2nd derivative of the subroutine Slater_repulsion_integral
 ! namely d2Jij/dalpha*dbeta
 !***************************************
  subroutine d2Slater_repulsion_integral( d2Jij, alpha_i , ni, alpha_j , nj, norm_dr  )
    real*8,intent(out) :: d2Jij
    real*8,intent(in) :: alpha_i, alpha_j, norm_dr
    integer, intent(in) :: ni, nj

    real*8 :: alpha, beta
    integer :: m , n 
    real*8 :: d2term1, d2term2, arg1, arg2, arg3
    real*8 :: sum1,sum2,dsum1,dsum2,dasum2,dbsum2,d2sum1,d2sum2,factor, dfactor, d2factor, DAP, DAP1, DAP2
    integer :: i,j,index, arg4
    real*8,parameter :: small=1D-4


    ! order slater orbitals by exponents
    if ( alpha_i > alpha_j ) then
       alpha = alpha_i
       beta = alpha_j
       m = ni
       n = nj
    else
       alpha = alpha_j
       beta = alpha_i
       m = nj
       n = ni
    endif


    !******************* for derivative, use corresponding formulas for Slater_repulsion_integral depending on magnitude of exponents

    ! if alpha /= beta
    if ( abs( alpha - beta ) > small ) then


       ! dsum1 is derivative of sum1 w.r.t beta
       dsum1=0d0
       ! d2sum1 is d2sum1/dalpha*dbeta
       d2sum1=0d0
       do i=0,2*n
          factor=dble(2*n-i)/factorial(i) *(beta*norm_dr)**i
          dfactor=factor * dble(i) / beta
          sum2=0d0
          ! this is dsum2/dalpha
          dasum2=0d0
          ! this is dsum2/dbeta
          dbsum2=0d0
          ! this is d2sum2/dalpha*dbeta
          d2sum2=0d0
          index=2*m-1+i
          do j=0, index
             arg2=norm_dr*(alpha-beta)
             arg3=norm_dr*(alpha+beta)
             sum2=sum2+Dpmn(j, 2*m-1,i) * Bn(j,arg2) * An_plus(index-j,arg3)
             dbsum2=dbsum2+norm_dr*Dpmn(j, 2*m-1,i)*( - dBn(j,arg2) * An_plus(index-j,arg3) + Bn(j,arg2) * dAn_plus(index-j,arg3) )
             dasum2=dasum2+norm_dr*Dpmn(j, 2*m-1,i)*( dBn(j,arg2) * An_plus(index-j,arg3) + Bn(j,arg2) * dAn_plus(index-j,arg3) )
             d2sum2= d2sum2 + norm_dr**2 * Dpmn(j, 2*m-1,i)* ( -d2Bn(j,arg2)* An_plus(index-j,arg3) + Bn(j,arg2) * d2An_plus(index-j,arg3) )
          enddo
          dsum1=dsum1 + factor*dbsum2 + dfactor * sum2
          d2sum1=d2sum1 + factor * d2sum2 + dfactor * dasum2
       enddo

       ! this is dsum/dbeta contribution
       d2term2=  dble(2*m+1)*(alpha*norm_dr)**(2*m)/2d0/factorial(2*m)/dble(n)  * dsum1
       ! this is d2sum/dalpha*dbeta contribution
       d2term2 = d2term2 + alpha*(alpha*norm_dr)**(2*m)/2d0/factorial(2*m)/dble(n) * d2sum1

       d2Jij = - d2term2 


    else
       ! if alpha == beta
       !************************************* NOTE **************************
       ! Similar comment as in dSlater_repulsion_integral
       ! see that comment for explanation of the 2 in the next line
       ! We are calculating limit(da-->0) [2dJ(alpha+da,beta+da)-2dJ(alpha,beta)]/da  at alpha=beta , 
       ! where dJ is shorthand for dJ/dalpha
       ! since this formula implicitly assumes alpha=beta
       ! this is equal to 4 * d2J(alpha,beta)/dalpha2 at alpha=beta

       ! first do d2term1/dalpha**2 , where term1 is defined as in the Slater_repulsion_integral subroutine
       arg1 = 2d0*alpha*norm_dr

       d2term1 =  dble(2*m) * dble(2*m-1) * arg1**(2*m-2)/(factorial(2*m)*norm_dr) * ( (arg1-2d0*m)*An_plus( (2*m-1), arg1) - exp(-arg1))     
       d2term1 = d2term1 + 2d0 * dble(2*m) * arg1**(2*m-1)/(factorial(2*m)*norm_dr) * (  An_plus( (2*m-1), arg1) + (arg1-2d0*m)* dAn_plus( (2*m-1), arg1) + exp(-arg1) )
       d2term1 = d2term1 + arg1**(2*m)/(factorial(2*m)*norm_dr) * ( 2d0 * dAn_plus( (2*m-1), arg1) + (arg1-2d0*m)*d2An_plus( (2*m-1), arg1) - exp(-arg1) )
       d2term1 = d2term1 * ( 2d0 * norm_dr )**2

       ! now term2
       d2sum1=0d0
       dsum1=0d0
       sum1=0d0
       do i=0,2*n-1
          factor=dble(2*n-i)/factorial(i) *(beta*norm_dr)**i
          dfactor=factor * dble(i) / beta
          d2factor = dfactor * dble(i-1) / beta

          sum2=0d0
          dsum2=0d0
          d2sum2=0d0
          ! note integer arithmetic here
          index= (2*m+i-1)/2
          do j=0, index
             arg4 = 2*m+i-1-2*j
             DAP = Dpmn(2*j, 2*m-1,i)/dble(2*j+1)* An_plus( arg4 , arg1 )
             DAP1 = Dpmn(2*j, 2*m-1,i)/dble(2*j+1)* dAn_plus( arg4 , arg1 )
             DAP2 = Dpmn(2*j, 2*m-1,i)/dble(2*j+1)* d2An_plus( arg4 , arg1 )
             sum2=sum2 + DAP
             dsum2=dsum2 + DAP1 * 2d0 * norm_dr
             d2sum2= d2sum2 + DAP2 * ( 2d0 * norm_dr ) **2
          enddo
          sum1=sum1+factor*sum2
          dsum1=dsum1+dfactor*sum2 + factor*dsum2
          d2sum1=d2sum1+d2factor*sum2 + 2d0 * dfactor*dsum2 + factor*d2sum2
       enddo

       factor = alpha*(alpha*norm_dr)**(2*m)/factorial(2*m)/dble(n)
       dfactor = dble(2*m+1) * factor / alpha
       d2factor = dble(2*m) * dfactor / alpha

       d2term2 = d2factor * sum1 +  2d0 * dfactor * dsum1 + factor * d2sum1

       d2Jij = d2term1 - d2term2

       ! divide by 4 here, per above comment
       d2Jij = d2Jij / 4d0

    endif


  end subroutine d2Slater_repulsion_integral



  !****************************************
  ! This is the square of the normalization factor
  ! for a Slater orbital
  ! i.e. NormS2 * integral( (r**(n-1)*exp(-alpha*r) , dx,dy,dz ) = 1
  !****************************************
  function NormS2( alpha, n )
   use global_variables
    real*8 :: NormS2
    real*8,intent(in) :: alpha
    integer,intent(in) :: n


   NormS2 = ( 2d0 * alpha )**(2*n+1) / ( 4d0 * pi * factorial(2*n) )

  end function NormS2


  !****************************************
  ! This is the derivative of NormS = sqrt(NormS2) w.r.t alpha
  !****************************************
  function dNormS( alpha, n )
   use global_variables
    real*8 :: dNormS
    real*8,intent(in) :: alpha
    integer,intent(in) :: n

    dNormS = 2d0**(dble(n)+0.5d0) * (dble(n)+0.5d0)*alpha**(dble(n)-0.5d0) / dsqrt( 4d0 * pi * factorial(2*n) )

  end function dNormS



  !****************************************
  ! this function is defined in Rosen, N. Phys Rev. 38, 1931, 255
  ! defined as integral ( x^n*exp(-alpha*x), dx,1,infinity )
  ! which is equal to An_minus(n,alpha) - An_plus(n,alpha)
  !
  ! Note the above formula is highly oscillator for alpha less than or close to one,
  ! and becomes numerically unstable for large n,
  ! The maximum value for n for which we need to calculate Bn , is for two 7s orbitals,
  ! in which case n_max = 2*7 + 2*7 -1 = 27
  !
  ! for this value of n_max, we have found through  numerical experimentation that the
  ! above formula is stable for alpha > 2 , while it is numerically  unstable for some
  ! values 1 < alpha < 2
  !
  ! therefore for values of alpha < 2
  ! we use the formula given by
  ! Zener, Guillemin, Phys. Rev. 1929, 34, 999
  !
  ! for alpha=2, the maximum values of the last term in the sum, 
  ! ~ alpha^(2*max_sum) / (2*max_sum)!
  ! has a value of 1E-36
  !***************************************** 
  function Bn(n,alpha)
    real*8 :: Bn
    integer,intent(in) :: n
    real*8, intent(in) :: alpha

    real*8,parameter :: threshold=2d0
    ! the last term in the Taylor expansion has (2*max_sum)! in the denominator
    integer,parameter :: max_sum = 20  
    integer :: i, index
    real*8 :: sum

    ! alpha should be positive
    if ( alpha < 0d0 ) stop "alpha should be positive in function Bn"

    if ( alpha >= threshold ) then
       Bn = An_minus(n,alpha) - An_plus(n,alpha)
    else
       sum=0d0
       ! Zener formula
       do i=0, max_sum
          if ( mod(n,2) == 0 ) then
             ! n is even
             index = 2*i
          else
             ! n is odd
             index = 2*i+1
          endif
          sum=sum + alpha**index/(factorial(index)*dble(index+n+1))
       enddo
       Bn = 2d0 * sum
       ! if odd, multiply by negative 1
       if ( mod(n,2) /= 0 ) then
          Bn = Bn * -1d0
       endif
    end if

  end function Bn



!**********************************
! this is the derivative of the function
! Bn w.r.t. alpha.  See the comments
! for the function Bn
!**********************************
function dBn(n,alpha)
    real*8 :: dBn
    integer,intent(in) :: n
    real*8, intent(in) :: alpha

    real*8,parameter :: threshold=2d0
    ! the last term in the Taylor expansion has (2*max_sum)! in the denominator
    integer,parameter :: max_sum = 20  
    integer :: i, index
    real*8 :: sum

    ! alpha should be positive
    if ( alpha < 0d0 ) stop "alpha should be positive in function Bn"

    if ( alpha >= threshold ) then
       dBn = dAn_minus(n,alpha) - dAn_plus(n,alpha)
    else
       sum=0d0
       ! Zener formula
       do i=0, max_sum
          if ( mod(n,2) == 0 ) then
             ! n is even
             index = 2*i
          else
             ! n is odd
             index = 2*i+1
          endif
          sum=sum + dble(index)*alpha**(index-1)/(factorial(index)*dble(index+n+1))
       enddo
       dBn = 2d0 * sum
       ! if odd, multiply by negative 1
       if ( mod(n,2) /= 0 ) then
         dBn = dBn * -1d0
       endif
    end if

  end function dBn


!**********************************
! this is 2nd the derivative of the function
! Bn w.r.t. alpha.  See the comments
! for the function Bn
!**********************************
function d2Bn(n,alpha)
    real*8 :: d2Bn
    integer,intent(in) :: n
    real*8, intent(in) :: alpha

    real*8,parameter :: threshold=2d0
    ! the last term in the Taylor expansion has (2*max_sum)! in the denominator
    integer,parameter :: max_sum = 20  
    integer :: i, index
    real*8 :: sum


    ! alpha should be positive
    if ( alpha < 0d0 ) stop "alpha should be positive in function Bn"

    if ( alpha >= threshold ) then
       d2Bn = d2An_minus(n,alpha) - d2An_plus(n,alpha)
    else
       sum=0d0
       ! Zener formula
       do i=0, max_sum
          if ( mod(n,2) == 0 ) then
             ! n is even
             index = 2*i
          else
             ! n is odd
             index = 2*i+1
          endif
          sum=sum + dble(index)*dble(index-1)*alpha**(index-2)/(factorial(index)*dble(index+n+1))
       enddo
       d2Bn = 2d0 * sum
       ! if odd, multiply by negative 1
       if ( mod(n,2) /= 0 ) then
         d2Bn = d2Bn * -1d0
       endif
    end if

  end function d2Bn



  !******************************************
  ! this function is defined in Rosen, N. Phys Rev. 38, 1931, 255
  ! defined as integral ( x^n*exp(-alpha*x), dx,1,infinity )
  !******************************************
  function An_plus(n,alpha)
    real*8 :: An_plus
    integer, intent(in) :: n
    real*8, intent(in) :: alpha
    integer :: i
    real*8 :: factor, sum
    factor = factorial(n)*exp(-alpha)/alpha**(n+1)
    sum=1d0
    ! note we implicitly include the i=0 term by setting sum=1 before hand
    do i=1,n
       sum=sum+alpha**i/factorial(i)
    enddo
    An_plus = factor * sum
  end function An_plus


  !******************************************
  ! this is the derivative of An_plus w.r.t. alpha
  !******************************************
  function dAn_plus(n,alpha)
    real*8 :: dAn_plus
    integer, intent(in) :: n
    real*8, intent(in) :: alpha
    integer :: i
    real*8 :: factor1, factor2,  sum1 , sum2

    factor1 = factorial(n)*exp(-alpha)/alpha**(n+1)
    factor2 = -1d0 * factor1 - factorial(n+1)*exp(-alpha)/alpha**(n+2)

    sum1=1d0
    sum2=0d0
    ! note we implicitly include the i=0 term by setting sum=1 before hand
    do i=1,n
       sum1=sum1+alpha**i/factorial(i)
       sum2=sum2+dble(i)*alpha**(i-1)/factorial(i)
    enddo

    dAn_plus = factor1 * sum2 + factor2 * sum1

  end function dAn_plus


  !******************************************
  ! this is the 2nd derivative of An_plus w.r.t. alpha
  !******************************************
  function d2An_plus(n,alpha)
    real*8 :: d2An_plus
    integer, intent(in) :: n
    real*8, intent(in) :: alpha
    integer :: i
    real*8 :: factor1, factor2, factor3, sum1 , sum2 , sum3, term1, term2, term3

    factor1 = factorial(n)*exp(-alpha)/alpha**(n+1)
    factor2 = dble(n+1) * factor1 / alpha 
    factor3 = dble(n+2) * factor2 / alpha

    sum1=1d0
    sum2=0d0
    sum3=0d0
    ! note we implicitly include the i=0 term by setting sum1=1 before hand
    do i=1,n
       term1 = alpha**i/factorial(i)
       term2 = dble(i) * term1 / alpha
       term3 = dble(i-1) * term2 / alpha
       sum1=sum1+term1
       sum2=sum2+term2
       sum3=sum3+term3
    enddo

    d2An_plus = (factor1+2d0*factor2+factor3)*sum1-2d0*(factor1+factor2)*sum2+factor1*sum3

  end function d2An_plus



  !******************************************
  ! this function is defined in Rosen, N. Phys Rev. 38, 1931, 255
  ! defined as integral ( x^n*exp(-alpha*x), dx,-1,infinity )
  !******************************************
  function An_minus(n,alpha)
    real*8 :: An_minus
    integer, intent(in) :: n
    real*8, intent(in) :: alpha
    integer :: i
    real*8 :: factor, sum
    factor = factorial(n)*exp(alpha)/alpha**(n+1)
    sum=1d0
    ! note we implicitly include the i=0 term by setting sum=1 before hand
    do i=1,n
       sum=sum+(-alpha)**i/factorial(i)
    enddo
    An_minus = factor * sum
  end function An_minus


  !******************************************
  ! this is the derivative of An_minus w.r.t. alpha
  !******************************************
  function dAn_minus(n,alpha)
    real*8 :: dAn_minus
    integer, intent(in) :: n
    real*8, intent(in) :: alpha
    integer :: i
    real*8 :: factor1, factor2,  sum1 , sum2

    factor1 = factorial(n)*exp(alpha)/alpha**(n+1)
    factor2 = factor1 - factorial(n+1)*exp(alpha)/alpha**(n+2)

    sum1=1d0
    sum2=0d0
    ! note we implicitly include the i=0 term by setting sum=1 before hand
    do i=1,n
       sum1=sum1+(-alpha)**i/factorial(i)
       sum2=sum2-dble(i)*(-alpha)**(i-1)/factorial(i)
    enddo

    dAn_minus = factor1 * sum2 + factor2 * sum1

  end function dAn_minus


  !******************************************
  ! this is the 2nd derivative of An_minus w.r.t. alpha
  !******************************************
  function d2An_minus(n,alpha)
    real*8 :: d2An_minus
    integer, intent(in) :: n
    real*8, intent(in) :: alpha
    integer :: i
    real*8 :: factor1, factor2, factor3, sum1 , sum2 , sum3, term1, term2, term3

    factor1 = factorial(n)*exp(alpha)/alpha**(n+1)
    factor2 = dble(n+1) * factor1 / alpha 
    factor3 = dble(n+2) * factor2 / alpha

    sum1=1d0
    sum2=0d0
    sum3=0d0
    ! note we implicitly include the i=0 term by setting sum1=1 before hand
    do i=1,n
       term1 = (-alpha)**i/factorial(i)
       term2 = - dble(i) * term1 / (-alpha)
       term3 = - dble(i-1) * term2 / (-alpha)
       sum1=sum1+term1
       sum2=sum2+term2
       sum3=sum3+term3
    enddo

    d2An_minus = (factor1-2d0*factor2+factor3)*sum1+2d0*(factor1-factor2)*sum2+factor1*sum3

  end function d2An_minus



  !*********************************************
  ! this function is defined in Rosen, N. Phys Rev. 38, 1931, 255
  ! it is defined by the equation
  ! (x+y)^m(x-y)^n= sum(Dpmn * x^(m+n-p) * y^p, p=0, m+n)
  !*********************************************
  function Dpmn( p , m , n )
    real*8 :: Dpmn
    integer, intent(in) :: p , m, n

    integer :: k , min_k , max_k

    Dpmn=0d0

    ! we need to find the integers we are summing over. k must satisfy the following
    ! k <= n , k <= p , p - m  <= k

    max_k = min( n, p )
    min_k = max( 0 , p - m )

    ! p should never be > n + m
    if ( p > ( m + n ) ) stop "p cannot be > m + n in function Dpmn"

    do k=min_k,max_k
       Dpmn = Dpmn + (-1d0)**k * Binomial_coefficient( m , p - k ) * Binomial_coefficient( n , k )
    enddo

  end function Dpmn



  !**************************************
  ! Binomial coefficient, n choose k
  !**************************************
  function Binomial_coefficient( n , k )
    real*8 :: Binomial_coefficient
    integer, intent(in) :: n, k

    Binomial_coefficient = factorial(n) / factorial(n-k) / factorial(k)

  end function Binomial_coefficient



  function factorial(n)
    real*8 :: factorial
    integer, intent(in) :: n

    integer :: i

    factorial=1d0

    if ( n < 0 ) stop "cannot compute factorial of negative number"

    if ( n > 1 ) then
       do i=2,n
          factorial=factorial * dble(i)
       enddo
    endif

  end function factorial




  !*********************************
  ! this subroutine gets the Qeq parameters from the elemental
  ! global arrays based on the atomtype, and converts the 3rd element
  ! of the parameter array from the covalent radius to the slater exponent, given by
  ! exp = lambda*(2n+1)/(2R)   (Rappe,A.K.,Goddard,W.A.III, J.Phys.Chem. 1991,95,3358-3363)
  ! ( lambda=.5 in original reference )
  ! 
  ! we assume that all atom labels begin with the element name
  ! and then possibly add additional integers as labels
  ! therefore we take the first 2 characters of the string as the
  ! element name, and drop the 2nd character if this is an integer
  !*********************************
  subroutine get_Qeq_parameters( alist, i_mole,  Qeq_param_i, ni, atomic_number  )
    use routines
    character(*),dimension(:,:),intent(in) :: alist
    integer, intent(in) :: i_mole
    real*8,dimension(:), intent(out) :: Qeq_param_i
    integer, intent(out) :: ni, atomic_number

    character(1),dimension(10) :: numbers
    integer :: i_element , index, flag
    character(len(alist(1,1))) :: string, element
    character(1) :: temp

    integer :: i

    numbers= (/"0","1","2","3","4","5","6","7","8","9"/)
    ! figure out what element this is
    string=alist(i_mole,1)
    call trim_head( string )
    temp = string(2:2)
    ! see if this is an  integers
    flag=0
    do index=1, size(numbers)
       if ( temp .eq. numbers(index) ) flag=1
    enddo
    if (flag == 1 ) then
       element=string(1:1)
    else
       element=string(1:2)
    endif

    ! now get the atomic_number
    flag=0
    do index=1,size(atomic_number_element)
       if ( element .eq. atomic_number_element(index) ) then
          flag=1
          atomic_number = index
          exit
       endif
    enddo
    if ( flag == 0 ) then
       write(*,*) "couldn't find atomic number for element ",element
       stop
    end if

    Qeq_param_i(:) = Qeq_element_parameters(atomic_number,:)


    ! now principle quantum number for element
    ! note that some implementations might have a different definition here
    ! see subroutine initialize_Qeq_parameters for comment
    if ( ( 1 <= atomic_number ) .and. ( atomic_number <= 2 ) ) then
       ni=1
    elseif( ( 3 <= atomic_number ) .and. ( atomic_number <= 10 ) ) then
       ni=2
    elseif( ( 11 <= atomic_number ) .and. ( atomic_number <= 18 ) ) then
       ni=3
    elseif( ( 19 <= atomic_number ) .and. ( atomic_number <= 36 ) ) then
       ni=4
    elseif( ( 37 <= atomic_number ) .and. ( atomic_number <= 54 ) ) then
       ni=5
    elseif( ( 55 <= atomic_number ) .and. ( atomic_number <= 86 ) ) then
       ni=6
    elseif( ( 87 <= atomic_number ) .and. ( atomic_number <= 109 ) ) then
       ni=7
    else
       write(*,*) "bad atomic number"
       stop
    endif

    if ( d_valence ) then
       ! here we are using the principle quantum numbers of the d-orbitals (f-orbitals)
       if ( ( 21 <= atomic_number ) .and. ( atomic_number <= 30 ) ) then
          ni=3
       elseif ( ( 39 <= atomic_number ) .and. ( atomic_number <= 48 ) ) then
          ni=4
       elseif ( ( 57 <= atomic_number ) .and. ( atomic_number <= 71 ) ) then
          ni=4
       elseif ( ( 72 <= atomic_number ) .and. ( atomic_number <= 80 ) ) then
          ni=5       
       elseif ( ( 89 <= atomic_number ) .and. ( atomic_number <= 103 ) ) then
          ni=5 
       elseif ( ( 104 <= atomic_number ) .and. ( atomic_number <= 109 ) ) then
          ni=6
       endif
    endif

    ! now convert radius to exponent
    Qeq_param_i(3) = Qeq_lambda * dble(2*ni+1)/(2d0 * Qeq_param_i(3))


  end subroutine get_Qeq_parameters


  !*********************************************
  ! this subroutine adjusts the Qeq parameters of hydrogen
  ! based on its charge
  ! 
  ! See Rappe,A.K.,Goddard,W.A.III, J.Phys.Chem. 1991,95,3358-3363
  !*********************************************
  subroutine set_hydrogen_dynamic_Qeq( Qeq_param , chg )
    real*8,dimension(3), intent(inout) :: Qeq_param
    real*8 , intent(in) :: chg

    real*8,dimension(3) :: Qeq_param_0

    ! this dynamic procedure is well-behaved for abs(chg) < 1
    if ( ( chg >= 1d0 ) .or. ( chg <= -1d0 ) ) then
       stop "hydrogen charge is either greater than 1 or less than -1"
    endif

    ! the parameters that are input are the zeroth order parameters
    Qeq_param_0 = Qeq_param


    ! exponent
    Qeq_param(3) = Qeq_param_0(3) + chg
    ! idempotential
    Qeq_param(1) = ( 1d0 + chg/ Qeq_param_0(3) ) * Qeq_param_0(1)


  end subroutine set_hydrogen_dynamic_Qeq



  !************************** parameters ********************
  ! these were taken from  Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323
  ! and GULP 4.0 source code  - (J.D. Gale, JCS Faraday Trans., 93, 629 (1997) )
  ! all parameters are in a.u. order is idempotential(J), electronegativity(X), covalent radius
  ! the Qeq_element_parameters array is ordered by atomic number
  !
  ! the exponents are derived from exponent = (2*n+1)/(4*R)
  ! where n is the principle quantum number for the valence electrons
  ! and R is the UFF covalent radius
  !
  ! however, there may be different definitions of "n" among programs
  ! so BE CAREFUL!
  !
  ! for example, the exponents in Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323
  ! are different from those in GULP, because the definition of n is different
  !
  ! to derive the Haldoupis parameters, a 3d transition metal should use "n=3", whereas
  ! in Gulp, the smae metal would have "n=4"
  !
  ! In Gulp, "n" is given by the row in the periodic table for ALL elements
  !**********************************************************
  subroutine initialize_Qeq_parameters

    Qeq_element_parameters( 1,:)=(/  0.51045,   0.16640,   0.70109/)
    Qeq_element_parameters( 2,:)=(/  1.09660,   0.35500,   2.45665/)
    Qeq_element_parameters( 3,:)=(/  0.17537,   0.11047,   2.94231/)
    Qeq_element_parameters( 4,:)=(/  0.32655,   0.17923,   2.34326/)
    Qeq_element_parameters( 5,:)=(/  0.34912,   0.18779,   1.55336/)
    Qeq_element_parameters( 6,:)=(/  0.37212,   0.19635,   1.43430/)
    Qeq_element_parameters( 7,:)=(/  0.43217,   0.25353,   1.35116/)
    Qeq_element_parameters( 8,:)=(/  0.49112,   0.32123,   1.26423/)
    Qeq_element_parameters( 9,:)=(/  0.54933,   0.39961,   1.33415/)
    Qeq_element_parameters(10,:)=(/  0.77541,   0.40571,   3.34104/)
    Qeq_element_parameters(11,:)=(/  0.16875,   0.10448,   3.94008/)
    Qeq_element_parameters(12,:)=(/  0.27143,   0.14520,   2.83459/)
    Qeq_element_parameters(13,:)=(/  0.26386,   0.14920,   2.26956/)
    Qeq_element_parameters(14,:)=(/  0.25629,   0.15317,   2.22232/)
    Qeq_element_parameters(15,:)=(/  0.29399,   0.20076,   2.08248/)
    Qeq_element_parameters(16,:)=(/  0.32971,   0.25460,   1.97855/)
    Qeq_element_parameters(17,:)=(/  0.36352,   0.31472,   1.87839/)
    Qeq_element_parameters(18,:)=(/  0.46708,   0.34783,   3.98355/)
    Qeq_element_parameters(19,:)=(/  0.14112,   0.08897,   4.88684/)
    Qeq_element_parameters(20,:)=(/  0.21168,   0.11874,   3.77946/)
    Qeq_element_parameters(21,:)=(/  0.22638,   0.12476,   3.30703/)
    Qeq_element_parameters(22,:)=(/  0.24843,   0.12752,   3.03679/)
    Qeq_element_parameters(23,:)=(/  0.25063,   0.13413,   2.77790/)
    Qeq_element_parameters(24,:)=(/  0.28407,   0.12550,   2.64940/)
    Qeq_element_parameters(25,:)=(/  0.30171,   0.12219,   2.89695/)
    Qeq_element_parameters(26,:)=(/  0.30428,   0.13818,   2.63239/)
    Qeq_element_parameters(27,:)=(/  0.30686,   0.15086,   2.65696/)
    Qeq_element_parameters(28,:)=(/  0.30906,   0.16409,   2.64184/)
    Qeq_element_parameters(29,:)=(/  0.31016,   0.15435,   2.70987/)
    Qeq_element_parameters(30,:)=(/  0.31494,   0.18764,   2.64562/)
    Qeq_element_parameters(31,:)=(/  0.23226,   0.13380,   2.28846/)
    Qeq_element_parameters(32,:)=(/  0.25269,   0.14887,   2.24689/)
    Qeq_element_parameters(33,:)=(/  0.27996,   0.19066,   2.27523/)
    Qeq_element_parameters(34,:)=(/  0.30362,   0.23622,   2.31303/)
    Qeq_element_parameters(35,:)=(/  0.32523,   0.28628,   2.15618/)
    Qeq_element_parameters(36,:)=(/  0.42004,   0.31255,   4.28968/)
    Qeq_element_parameters(37,:)=(/  0.13568,   0.08566,   5.23455/)
    Qeq_element_parameters(38,:)=(/  0.17934,   0.11113,   4.56370/)
    Qeq_element_parameters(39,:)=(/  0.20653,   0.14075,   3.77568/)
    Qeq_element_parameters(40,:)=(/  0.26092,   0.12495,   3.32214/)
    Qeq_element_parameters(41,:)=(/  0.24843,   0.13046,   3.02924/)
    Qeq_element_parameters(42,:)=(/  0.27599,   0.12734,   2.89129/)
    Qeq_element_parameters(43,:)=(/  0.29326,   0.12091,   2.83459/)
    Qeq_element_parameters(44,:)=(/  0.29510,   0.13138,   2.83459/)
    Qeq_element_parameters(45,:)=(/  0.29436,   0.14608,   2.85160/)
    Qeq_element_parameters(46,:)=(/  0.29399,   0.15876,   2.91774/)
    Qeq_element_parameters(47,:)=(/  0.23034,   0.16302,   3.06514/)
    Qeq_element_parameters(48,:)=(/  0.29083,   0.18500,   3.02357/)
    Qeq_element_parameters(49,:)=(/  0.21285,   0.12884,   2.65318/)
    Qeq_element_parameters(50,:)=(/  0.22961,   0.14652,   2.55869/)
    Qeq_element_parameters(51,:)=(/  0.24563,   0.18003,   2.65318/)
    Qeq_element_parameters(52,:)=(/  0.25916,   0.21373,   2.60783/)
    Qeq_element_parameters(53,:)=(/  0.27650,   0.25070,   2.51901/)
    Qeq_element_parameters(54,:)=(/  0.36566,   0.27911,   4.64684/)
    Qeq_element_parameters(55,:)=(/  0.12576,   0.08022,   5.63895/)
    Qeq_element_parameters(56,:)=(/  0.17610,   0.10341,   4.61472/)
    Qeq_element_parameters(57,:)=(/  0.20150,   0.10422,   3.91363/)
    Qeq_element_parameters(58,:)=(/  0.19786,   0.10194,   3.63773/)
    Qeq_element_parameters(59,:)=(/  0.18845,   0.10503,   3.79269/)
    Qeq_element_parameters(60,:)=(/  0.19260,   0.10543,   3.79269/)
    Qeq_element_parameters(61,:)=(/  0.19646,   0.10587,   3.77946/)
    Qeq_element_parameters(62,:)=(/  0.19988,   0.10701,   3.73788/)
    Qeq_element_parameters(63,:)=(/  0.20488,   0.10580,   4.20843/)
    Qeq_element_parameters(64,:)=(/  0.21862,   0.11639,   3.71899/)
    Qeq_element_parameters(65,:)=(/  0.20830,   0.11091,   3.69253/)
    Qeq_element_parameters(66,:)=(/  0.21105,   0.11231,   3.65474/)
    Qeq_element_parameters(67,:)=(/  0.21248,   0.11492,   3.63773/)
    Qeq_element_parameters(68,:)=(/  0.21421,   0.11712,   3.61883/)
    Qeq_element_parameters(69,:)=(/  0.21557,   0.11947,   3.77946/)
    Qeq_element_parameters(70,:)=(/  0.21792,   0.12087,   4.07804/)
    Qeq_element_parameters(71,:)=(/  0.18103,   0.10889,   3.58293/)
    Qeq_element_parameters(72,:)=(/  0.24990,   0.13597,   3.32403/)
    Qeq_element_parameters(73,:)=(/  0.20947,   0.18742,   3.03302/)
    Qeq_element_parameters(74,:)=(/  0.24328,   0.17015,   2.90640/)
    Qeq_element_parameters(75,:)=(/  0.28811,   0.14553,   3.02357/)
    Qeq_element_parameters(76,:)=(/  0.26680,   0.18889,   3.21254/)
    Qeq_element_parameters(77,:)=(/  0.29399,   0.18375,   3.52623/)
    Qeq_element_parameters(78,:)=(/  0.32560,   0.17603,   2.94231/)
    Qeq_element_parameters(79,:)=(/  0.19007,   0.17985,   3.05758/)
    Qeq_element_parameters(80,:)=(/  0.30575,   0.23042,   3.02357/)
    Qeq_element_parameters(81,:)=(/  0.21315,   0.11760,   2.89129/)
    Qeq_element_parameters(82,:)=(/  0.25945,   0.14332,   2.72877/)
    Qeq_element_parameters(83,:)=(/  0.27488,   0.17235,   2.86105/)
    Qeq_element_parameters(84,:)=(/  0.30943,   0.15471,   2.79680/)
    Qeq_element_parameters(85,:)=(/  0.34912,   0.17456,   2.77790/)
    Qeq_element_parameters(86,:)=(/  0.39469,   0.19734,   4.15740/)
    Qeq_element_parameters(87,:)=(/  0.14700,   0.07350,   4.34638/)
    Qeq_element_parameters(88,:)=(/  0.17890,   0.10448,   4.15740/)
    Qeq_element_parameters(89,:)=(/  0.20837,   0.10418,   3.98355/)
    Qeq_element_parameters(90,:)=(/  0.21351,   0.11668,   3.81347/)
    Qeq_element_parameters(91,:)=(/  0.21351,   0.10970,   3.40151/)
    Qeq_element_parameters(92,:)=(/  0.20969,   0.12278,   3.23711/)
    Qeq_element_parameters(93,:)=(/  0.19970,   0.13042,   3.40151/)
    Qeq_element_parameters(94,:)=(/  0.20719,   0.11918,   3.47710/)
    Qeq_element_parameters(95,:)=(/  0.22075,   0.10988,   3.66985/)
    Qeq_element_parameters(96,:)=(/  0.22075,   0.10407,   3.59049/)


    ! these are the exact parameters from Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323
    ! given in a.u. order is idempotential(J), electronegativity(X), exponent
!!$    Qeq_element_parameters( 1,:)=(/  0.51045,   0.16640,   1.0698 /)
!!$    Qeq_element_parameters( 2,:)=(/  1.09660,   0.35500,   0.3053 /)
!!$    Qeq_element_parameters( 3,:)=(/  0.17537,   0.11047,   0.4248 /)
!!$    Qeq_element_parameters( 4,:)=(/  0.32655,   0.17923,   0.5334 /)
!!$    Qeq_element_parameters( 5,:)=(/  0.34912,   0.18779,   0.8047 /)
!!$    Qeq_element_parameters( 6,:)=(/  0.37212,   0.19635,   0.8715 /)
!!$    Qeq_element_parameters( 7,:)=(/  0.43217,   0.25353,   0.9251 /)
!!$    Qeq_element_parameters( 8,:)=(/  0.49112,   0.32123,   0.9887 /)
!!$    Qeq_element_parameters( 9,:)=(/  0.54933,   0.39961,   0.9369 /)
!!$    Qeq_element_parameters(10,:)=(/  0.77541,   0.40571,   0.3741 /)
!!$    Qeq_element_parameters(11,:)=(/  0.16875,   0.10448,   0.4441 /)
!!$    Qeq_element_parameters(12,:)=(/  0.27143,   0.14520,   0.6174 /)
!!$    Qeq_element_parameters(13,:)=(/  0.26386,   0.14920,   0.7711 /)
!!$    Qeq_element_parameters(14,:)=(/  0.25629,   0.15317,   0.7875 /)
!!$    Qeq_element_parameters(15,:)=(/  0.29399,   0.20076,   0.8403 /)
!!$    Qeq_element_parameters(16,:)=(/  0.32971,   0.25460,   0.8845 /)
!!$    Qeq_element_parameters(17,:)=(/  0.36352,   0.31472,   0.9316 /)
!!$    Qeq_element_parameters(18,:)=(/  0.46708,   0.34783,   0.4393 /)
!!$    Qeq_element_parameters(19,:)=(/  0.14112,   0.08897,   0.4604 /)
!!$    Qeq_element_parameters(20,:)=(/  0.21168,   0.11874,   0.5953 /)
!!$    Qeq_element_parameters(21,:)=(/  0.22638,   0.12476,   0.5292 /)
!!$    Qeq_element_parameters(22,:)=(/  0.24843,   0.12752,   0.5763 /)
!!$    Qeq_element_parameters(23,:)=(/  0.25063,   0.13413,   0.6300 /)
!!$    Qeq_element_parameters(24,:)=(/  0.28407,   0.12550,   0.6605 /)
!!$    Qeq_element_parameters(25,:)=(/  0.30171,   0.12219,   0.6041 /)
!!$    Qeq_element_parameters(26,:)=(/  0.30428,   0.13818,   0.6648 /)
!!$    Qeq_element_parameters(27,:)=(/  0.30686,   0.15086,   0.6586 /)
!!$    Qeq_element_parameters(28,:)=(/  0.30906,   0.16409,   0.6624 /)
!!$    Qeq_element_parameters(29,:)=(/  0.31016,   0.15435,   0.6458 /)
!!$    Qeq_element_parameters(30,:)=(/  0.31494,   0.18764,   0.6615 /)
!!$    Qeq_element_parameters(31,:)=(/  0.23226,   0.13380,   0.9832 /)
!!$    Qeq_element_parameters(32,:)=(/  0.25269,   0.14887,   1.0014 /)
!!$    Qeq_element_parameters(33,:)=(/  0.27996,   0.19066,   0.9889 /)
!!$    Qeq_element_parameters(34,:)=(/  0.30362,   0.23622,   0.9727 /)
!!$    Qeq_element_parameters(35,:)=(/  0.32523,   0.28628,   1.0435 /)
!!$    Qeq_element_parameters(36,:)=(/  0.42004,   0.31255,   0.5245 /)
!!$    Qeq_element_parameters(37,:)=(/  0.13568,   0.08566,   0.5253 /)
!!$    Qeq_element_parameters(38,:)=(/  0.17934,   0.11113,   0.6026 /)
!!$    Qeq_element_parameters(39,:)=(/  0.20653,   0.14075,   0.5959 /)
!!$    Qeq_element_parameters(40,:)=(/  0.26092,   0.12495,   0.6773 /)
!!$    Qeq_element_parameters(41,:)=(/  0.24843,   0.13046,   0.7428 /)
!!$    Qeq_element_parameters(42,:)=(/  0.27599,   0.12734,   0.7782 /)
!!$    Qeq_element_parameters(43,:)=(/  0.29326,   0.12091,   0.7938 /)
!!$    Qeq_element_parameters(44,:)=(/  0.29510,   0.13138,   0.7938 /)
!!$    Qeq_element_parameters(45,:)=(/  0.29436,   0.14608,   0.7890 /)
!!$    Qeq_element_parameters(46,:)=(/  0.29399,   0.15876,   0.7711 /)
!!$    Qeq_element_parameters(47,:)=(/  0.23034,   0.16302,   0.7341 /)
!!$    Qeq_element_parameters(48,:)=(/  0.29083,   0.18500,   0.7441 /)
!!$    Qeq_element_parameters(49,:)=(/  0.21285,   0.12884,   1.0365 /)
!!$    Qeq_element_parameters(50,:)=(/  0.22961,   0.14652,   1.0748 /)
!!$    Qeq_element_parameters(51,:)=(/  0.24563,   0.18003,   1.0365 /)
!!$    Qeq_element_parameters(52,:)=(/  0.25916,   0.21373,   1.0545 /)
!!$    Qeq_element_parameters(53,:)=(/  0.27650,   0.25070,   1.0917 /)
!!$    Qeq_element_parameters(54,:)=(/  0.36566,   0.27911,   0.5918 /)
!!$    Qeq_element_parameters(55,:)=(/  0.12576,   0.08022,   0.4877 /)
!!$    Qeq_element_parameters(56,:)=(/  0.17610,   0.10341,   0.5959 /)
!!$    Qeq_element_parameters(57,:)=(/  0.20150,   0.10422,   0.7027 /)
!!$    Qeq_element_parameters(58,:)=(/  0.19786,   0.10194,   0.7560 /)
!!$    Qeq_element_parameters(59,:)=(/  0.18845,   0.10503,   0.5932 /)
!!$    Qeq_element_parameters(60,:)=(/  0.19260,   0.10543,   0.5932 /)
!!$    Qeq_element_parameters(61,:)=(/  0.19646,   0.10587,   0.5953 /)
!!$    Qeq_element_parameters(62,:)=(/  0.19988,   0.10701,   0.6019 /)
!!$    Qeq_element_parameters(63,:)=(/  0.20488,   0.10580,   0.5346 /)
!!$    Qeq_element_parameters(64,:)=(/  0.21862,   0.11639,   0.7394 /)
!!$    Qeq_element_parameters(65,:)=(/  0.20830,   0.11091,   0.6093 /)
!!$    Qeq_element_parameters(66,:)=(/  0.21105,   0.11231,   0.6156 /)
!!$    Qeq_element_parameters(67,:)=(/  0.21248,   0.11492,   0.6185 /)
!!$    Qeq_element_parameters(68,:)=(/  0.21421,   0.11712,   0.6217 /)
!!$    Qeq_element_parameters(69,:)=(/  0.21557,   0.11947,   0.5953 /)
!!$    Qeq_element_parameters(70,:)=(/  0.21792,   0.12087,   0.5517 /)
!!$    Qeq_element_parameters(71,:)=(/  0.18103,   0.10889,   0.7675 /)
!!$    Qeq_element_parameters(72,:)=(/  0.24990,   0.13597,   0.8273 /)
!!$    Qeq_element_parameters(73,:)=(/  0.20947,   0.18742,   0.9067 /)
!!$    Qeq_element_parameters(74,:)=(/  0.24328,   0.17015,   0.9462 /)
!!$    Qeq_element_parameters(75,:)=(/  0.28811,   0.14553,   0.9095 /)
!!$    Qeq_element_parameters(76,:)=(/  0.26680,   0.18889,   0.8560 /)
!!$    Qeq_element_parameters(77,:)=(/  0.29399,   0.18375,   0.7799 /)
!!$    Qeq_element_parameters(78,:)=(/  0.32560,   0.17603,   0.9346 /)
!!$    Qeq_element_parameters(79,:)=(/  0.19007,   0.17985,   0.8994 /)
!!$    Qeq_element_parameters(80,:)=(/  0.30575,   0.23042,   0.9095 /)
!!$    Qeq_element_parameters(81,:)=(/  0.21315,   0.11760,   1.1241 /)
!!$    Qeq_element_parameters(82,:)=(/  0.25945,   0.14332,   1.1910 /)
!!$    Qeq_element_parameters(83,:)=(/  0.27488,   0.17235,   1.1359 /)
!!$    Qeq_element_parameters(84,:)=(/  0.30943,   0.15471,   1.1620 /)
!!$    Qeq_element_parameters(85,:)=(/  0.34912,   0.17456,   1.1699 /)
!!$    Qeq_element_parameters(86,:)=(/  0.39469,   0.19734,   0.7817 /)
!!$    Qeq_element_parameters(87,:)=(/  0.14700,   0.07350,   0.8628 /)
!!$    Qeq_element_parameters(88,:)=(/  0.17890,   0.10448,   0.9020 /)
!!$    Qeq_element_parameters(89,:)=(/  0.20837,   0.10418,   0.8158 /)
!!$    Qeq_element_parameters(90,:)=(/  0.21351,   0.11668,   0.8522 /)
!!$    Qeq_element_parameters(91,:)=(/  0.21351,   0.10970,   0.9554 /)
!!$    Qeq_element_parameters(92,:)=(/  0.20969,   0.12278,   1.0040 /)
!!$    Qeq_element_parameters(93,:)=(/  0.19970,   0.13042,   0.9554 /)
!!$    Qeq_element_parameters(94,:)=(/  0.20719,   0.11918,   0.7909 /)
!!$    Qeq_element_parameters(95,:)=(/  0.22075,   0.10988,   0.7493 /)
!!$    Qeq_element_parameters(96,:)=(/  0.22075,   0.10407,   0.9052 /)

    !************ if using parameters from Kadantsev et al., J.Phys.Chem.Lett.2013,4,3056-3061
    ! defined for H,V,Cu,Zn,C,N,O,F,Cl,Br,I
    Select Case(Woo_parameters)
    Case("yes")
       Qeq_element_parameters(1,1)=0.51045; Qeq_element_parameters(1,2)=0.16640;  ! H is unchanged
       Qeq_element_parameters(6,1)=0.43048; Qeq_element_parameters(6,2)=0.19959;
       Qeq_element_parameters(7,1)=0.48671; Qeq_element_parameters(7,2)=0.24578;
       Qeq_element_parameters(8,1)=0.62974; Qeq_element_parameters(8,2)=0.32023;
       Qeq_element_parameters(9,1)=0.81811; Qeq_element_parameters(9,2)=0.23578;
       Qeq_element_parameters(17,1)=0.53456; Qeq_element_parameters(17,2)=0.21392;
       Qeq_element_parameters(23,1)=0.30994; Qeq_element_parameters(23,2)=0.15041;
       Qeq_element_parameters(29,1)=0.25489; Qeq_element_parameters(29,2)=0.19951;
       Qeq_element_parameters(30,1)=0.32802; Qeq_element_parameters(30,2)=0.13601;
       Qeq_element_parameters(35,1)=0.64385; Qeq_element_parameters(35,2)=0.20918;
       Qeq_element_parameters(53,1)=0.42041; Qeq_element_parameters(53,2)=0.19959;

    End Select


    atomic_number_element( 1)="H    "
    atomic_number_element( 2)="He   "
    atomic_number_element( 3)="Li   "
    atomic_number_element( 4)="Be   "
    atomic_number_element( 5)="B    "
    atomic_number_element( 6)="C    "
    atomic_number_element( 7)="N    "
    atomic_number_element( 8)="O    "
    atomic_number_element( 9)="F    "
    atomic_number_element(10)="Ne   "
    atomic_number_element(11)="Na   "
    atomic_number_element(12)="Mg   "
    atomic_number_element(13)="Al   "
    atomic_number_element(14)="Si   "
    atomic_number_element(15)="P    "
    atomic_number_element(16)="S    "
    atomic_number_element(17)="Cl   "
    atomic_number_element(18)="Ar   "
    atomic_number_element(19)="K    "
    atomic_number_element(20)="Ca   "
    atomic_number_element(21)="Sc   "
    atomic_number_element(22)="Ti   "
    atomic_number_element(23)="V    "
    atomic_number_element(24)="Cr   "
    atomic_number_element(25)="Mn   "
    atomic_number_element(26)="Fe   "
    atomic_number_element(27)="Co   "
    atomic_number_element(28)="Ni   "
    atomic_number_element(29)="Cu   "
    atomic_number_element(30)="Zn   "
    atomic_number_element(31)="Ga   "
    atomic_number_element(32)="Ge   "
    atomic_number_element(33)="As   "
    atomic_number_element(34)="Se   "
    atomic_number_element(35)="Br   "
    atomic_number_element(36)="Kr   "
    atomic_number_element(37)="Rb   "
    atomic_number_element(38)="Sr   "
    atomic_number_element(39)="Y    "
    atomic_number_element(40)="Zr   "
    atomic_number_element(41)="Nb   "
    atomic_number_element(42)="Mo   "
    atomic_number_element(43)="Tc   "
    atomic_number_element(44)="Ru   "
    atomic_number_element(45)="Rh   "
    atomic_number_element(46)="Pd   "
    atomic_number_element(47)="Ag   "
    atomic_number_element(48)="Cd   "
    atomic_number_element(49)="In   "
    atomic_number_element(50)="Sn   "
    atomic_number_element(51)="Sb   "
    atomic_number_element(52)="Te   "
    atomic_number_element(53)="I    "
    atomic_number_element(54)="Xe   "
    atomic_number_element(55)="Cs   "
    atomic_number_element(56)="Ba   "
    atomic_number_element(57)="La   "
    atomic_number_element(58)="Ce   "
    atomic_number_element(59)="Pr   "
    atomic_number_element(60)="Nd   "
    atomic_number_element(61)="Pm   "
    atomic_number_element(62)="Sm   "
    atomic_number_element(63)="Eu   "
    atomic_number_element(64)="Gd   "
    atomic_number_element(65)="Tb   "
    atomic_number_element(66)="Dy   "
    atomic_number_element(67)="Ho   "
    atomic_number_element(68)="Er   "
    atomic_number_element(69)="Tm   "
    atomic_number_element(70)="Yb   "
    atomic_number_element(71)="Lu   "
    atomic_number_element(72)="Hf   "
    atomic_number_element(73)="Ta   "
    atomic_number_element(74)="W    "
    atomic_number_element(75)="Re   "
    atomic_number_element(76)="Os   "
    atomic_number_element(77)="Ir   "
    atomic_number_element(78)="Pt   "
    atomic_number_element(79)="Au   "
    atomic_number_element(80)="Hg   "
    atomic_number_element(81)="Tl   "
    atomic_number_element(82)="Pb   "
    atomic_number_element(83)="Bi   "
    atomic_number_element(84)="Po   "
    atomic_number_element(85)="At   "
    atomic_number_element(86)="Rn   "
    atomic_number_element(87)="Fr   "
    atomic_number_element(88)="Ra   "
    atomic_number_element(89)="Ac   "
    atomic_number_element(90)="Th   "
    atomic_number_element(91)="Pa   "
    atomic_number_element(92)="U    "
    atomic_number_element(93)="Np   "
    atomic_number_element(94)="Pu   "
    atomic_number_element(95)="Am   "
    atomic_number_element(96)="Cm   "


  end subroutine initialize_Qeq_parameters




  ! write Qeq output
  subroutine  write_Qeq_output(flag_converged,iteration, n_framework_atoms, n_mole, alist, chg )
    integer,intent(in) :: flag_converged, iteration,n_framework_atoms, n_mole
    character(*),dimension(:,:), intent(in) :: alist
    real*8,dimension(:,:),intent(in) :: chg

    integer :: i_atom, index

    write(*,*) "****************** Qeq charge optimization *****************"
    write(*,*) " periodic Qeq charge optimization has been conducted using  "
    write(*,*) " the following parameters "
    write(*,*) ""
    write(*,*) " Qeq parameters are from the UFF force field"
    if ( Woo_parameters .eq. "yes" ) then
       write(*,*) " parameters for H,V,Cu,Zn,C,N,O,F,Cl,Br,I are taken from "
       write(*,*) " Kadantsev et al., J.Phys.Chem.Lett.2013,4,3056-3061 "
    end if

    if ( dynamic_hydrogen_param .eq. "yes" ) then
       write(*,*) " Qeq is solved iteratively as hydrogen idempotential and "
       write(*,*) " exponent are charge dependent"
       write(*,*) ""
    end if

    if ( d_valence ) then
       write(*,*) " The principle quantum number used to "
       write(*,*) " define the slater orbitals is d(f)-shell dependent "
       write(*,*) " (i.e. for Zn, n=3 )"
       write(*,*) " this is consistent with (most of) the exponents given in "
       write(*,*) " Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323"
       write(*,*) " but is inconsistent with the implementation in "
       write(*,*) " GULP 4.0 source code  - (J.D. Gale, JCS Faraday Trans., 93, 629 (1997)"
       write(*,*) " as GULP always sets n equal to the Periodic table row of the element"
       write(*,*) " (i.e. Zn, n=4) "
       write(*,*) ""
    else
       write(*,*) " It is always assumed that the principle quantum number used to "
       write(*,*) " define the slater orbitals is equal to the row of the element "
       write(*,*) " in the periodic table (i.e. for Zn, n=4 )"
       write(*,*) " this is consistent with the implementation in "
       write(*,*) " GULP 4.0 source code  - (J.D. Gale, JCS Faraday Trans., 93, 629 (1997) )"
       write(*,*) " but is inconsistent with the exponents given in "
       write(*,*) " Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323"
       write(*,*) " As this latter reference uses d shell for n, (Zn, n=3) "
       write(*,*) ""
    endif

    Select Case(real_Jij_screentype)
    Case("Slater1")
       write(*,*) " *****   Are you sure you want the setting real_Jij_screentype='Slater1'? *****"
       write(*,*) " Note this is non-standard, as it is inconsistent with the implementation"
       write(*,*) " in GULP as well as Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323"
       write(*,*) ""
    Case("Slater2")
       write(*,*) " The real space contribution to the Jij matrix is calculated as in"
       write(*,*) " the standard implementation in GULP,"
       write(*,*) " (as well as Haldoupis,E.,Nair,S.,Sholl,D.S.,JACS,2012,134,4313-4323)"
       write(*,*) " where the Slater orbitals are squared in the Coulomb integral"
       write(*,*) ""
    End Select


    Select Case(flag_converged)
    Case(0)
       write(*,*) " iterative Qeq solver did not converge to precision"
       write(*,*) qeq_thresh, "electrons (average per atom)"
       write(*,*) " within ", maxit, " iterations"
    Case(1)
       write(*,*) "iterative Qeq solver converged to precision"
       write(*,*) qeq_thresh, "electrons (average per atom)"
       write(*,*) "within ", iteration, " iterations"
    End Select

    write(*,*) "****************Final Qeq charges****************"
    do i_atom=1, n_framework_atoms
       index = n_mole + i_atom
       write(*,*) alist(index,1) ,index, chg(index,1)
    enddo

    ! if periodic_Qeq was set to "no", then we assume this is for testing.  Stop the code
    Select Case(periodic_Qeq)
       Case("no")
          write(*,*) ""
          write(*,*) "parameter periodic_Qeq is set to 'no' in Qeq_charge_equilibration."
          write(*,*) "Therefore this is a test run, and the code will now stop"
          stop
    End Select

  end subroutine write_Qeq_output



end module Qeq
