module boltzman_factor
!*****************************************************************************************
! This module is used to define function bz_fac, which calculates the Boltzmann factor
! with a given set of position parameters, defined in posit_p. The module is constructed
! for the need of Vegas sampling method in GNU Scientific Library
! Important variables :
! real*8, dimension(5) :: posit_p
! First three components represent translational degrees of freedom, are reandom numbers
! ranging from -0.5 ~ 0.5. Last two components specifies orientaions, they are defined as
! phi and psi, respectively. 
! currently only works for linear molecules
!*****************************************************************************************
use global_variables
use pme_routines
use linklist
use fgsl
use, intrinsic :: iso_c_binding
implicit none

  integer :: tot_n_mole_bz, n_mole_bz
  integer, dimension(MAX_N_MOLE) :: n_atom_bz 
  real*8 :: cubelet_vol_bz
  real*8, dimension(3) :: center_bz, u_axis_bz
  real*8, dimension(3,3) :: box_bz, cubelet_vec_bz
  real*8, dimension(MAX_N_MOLE,MAX_N_ATOM,3) :: xyz_bz, xyz_config_bz
  real*8,dimension(MAX_N_MOLE,3) :: new_r_com_bz, r_com_bz
  real*8, dimension(MAX_N_MOLE,MAX_N_ATOM) :: atom_chg_bz, mass_bz
  integer, dimension(:,:,:,:), allocatable :: headlist_bz
  integer, dimension(MAX_N_MOLE,MAX_N_ATOM,2) :: nslist_bz
  real*8,dimension(:,:,:), allocatable :: recip_pot_bz
  ! Note this array should be allocated out side of the module
  real*8, dimension(:,:,:,:), allocatable :: lrdisp_pot_bz

contains

  real*8 function bz_fac( posit_p, n, params ) bind(c)
  implicit none
    real*8, dimension(*), intent(in) :: posit_p
    integer(c_size_t), intent(in) :: n
    type(c_ptr), intent(in) :: params
    ! local variables
    integer :: is_overlap
    integer :: i_mole, i_atom, j_mole, j_atom
    real*8 :: phi, theta, energy, dE_elec, dE_bh
    real*8, dimension(3) :: disp, unit
    real*8,parameter :: small = 1D-4, max_rep=100.

    theta = posit_p(4)
    phi = posit_p(5)
    i_mole = 1

    ! make sure this is a linear molecule
    if ( molecule_shape(i_mole) .ne. 1 ) then
       stop "can only use bz_fac for linear molecules"
    endif

    ! Note, the given posit(1-3) must range from -0.5 ~ 0.5
    new_r_com_bz(i_mole,:) = center_bz + posit_p(1)*cubelet_vec_bz(1,:) &
                         & + posit_p(2)*cubelet_vec_bz(2,:) + posit_p(3)*cubelet_vec_bz(3,:)
    unit(1) = dsin(theta)*dcos(phi)
    unit(2) = dsin(theta)*dsin(phi)
    unit(3) = dcos(theta)
    do i_atom = 1, n_atom_bz(i_mole)
      disp(:) = xyz_bz(i_mole,i_atom,:) - r_com_bz(i_mole,:)
      if ( dot_product(disp, u_axis_bz) > 0.0d0 ) then
        xyz_config_bz(i_mole, i_atom, :) = dsqrt(dot_product(disp,disp))*unit(:) + new_r_com_bz(i_mole,:)
      else
        xyz_config_bz(i_mole, i_atom, :) = -dsqrt(dot_product(disp,disp))*unit(:) + new_r_com_bz(i_mole,:)
      endif
    enddo
    is_overlap = overlap_frame( xyz_config_bz, tot_n_mole_bz, n_mole_bz, n_atom_bz, i_mole, box_bz &
                             &, new_r_com_bz, headlist_bz, nslist_bz )
    if ( is_overlap == 1 ) then
      energy = max_rep
    else
      select case(electrostatic_type)
      case("none")
        dE_elec = 0.0d0
      case default
        dE_elec = pme_usegrid( box_bz, i_mole, tot_n_mole_bz, n_mole_bz, n_atom_bz, new_r_com_bz, xyz_config_bz, atom_chg_bz &
                            &, recip_pot_bz, headlist_bz, nslist_bz )
      end select
      dE_bh = disp_usegrid( box_bz, i_mole, tot_n_mole_bz, n_mole_bz, n_atom_bz, new_r_com_bz, xyz_config_bz, lrdisp_pot_bz &
                         &, headlist_bz, nslist_bz, mass_bz)
      energy = dE_elec + dE_bh
    endif

    ! with normalization
    bz_fac = dexp( -energy*1000d0/8.314d0/temp ) * dsin(theta) /2.0d0/pi

    return
  end function bz_fac

end module boltzman_factor
