!*********************************************************************
!  This module is filling with miscellaneous subroutines that are used
!  extensively throughout the program, but seemingly have no better
!  place to reside.
!*********************************************************************

module routines
  implicit none

contains

  !***********************************************************************
  ! This subroutine read input files ( *.conf file )
  ! It also assigns charges to each atoms, based on parameters in glob_v.f90 file
  !***********************************************************************
  subroutine read_conf( ifn, n_mole, n_atom, xyz, alist, box )
    use global_variables
    implicit none
    character(MAX_FN), intent(in) :: ifn
    character(MAX_ANAME), intent(out), dimension(MAX_N_MOLE,MAX_N_ATOM) :: alist
    integer, intent(out) :: n_mole
    integer, intent(out), dimension(MAX_N_MOLE) :: n_atom
    real*8, intent(out), dimension(MAX_N_MOLE,MAX_N_ATOM,3) :: xyz
    real*8, intent(out), dimension(3,3) :: box
    integer :: ifile, i, n_line, i_mole, i_atom, i_mole_prev, junk, inputstatus, ind, flag, i_type, i_param, n_mole_atoms
    real*8, dimension(3) :: r_tmp
    character(MAX_MNAME) :: mname
    character(MAX_ANAME) :: aname
    character(50)::line
    real*8 :: tmp1, tmp2, tmp3
    integer :: ios

    ifile=99
    open( ifile, file=ifn, status='old' )
    read( ifile, '(I)' ) n_line
    n_mole = 0
    i_atom = 0
    if (n_line .gt. 0 ) then
       do i = 1, n_line
          read( ifile, '(I5,2A5,I5,3F14.6)' ), i_mole, mname, aname, junk, r_tmp(1), r_tmp(2), r_tmp(3)
          call trim_head( aname )
          if ( i_mole /= n_mole ) then
             if ( n_mole > 0 ) then
                n_atom(n_mole) = i_atom
                ! make sure we dont have too many atoms for this molecule
                if ( i_atom > MAX_N_ATOM ) then
                   stop "please increase the value of MAX_N_ATOM in glob_v.f90 to accomodate the size of the solute molecules"
                endif
             end if
             n_mole = i_mole
             i_atom = 0
          end if
          i_atom = i_atom + 1
          xyz( i_mole, i_atom, : ) = r_tmp(:)
          alist( i_mole, i_atom ) = aname
       end do
       n_atom(n_mole) = i_atom
       box = 0.0
    endif

    read(ifile,*) tmp1, tmp2, tmp3
    read( ifile, *, iostat=ios ) box(2,1), box(2,2), box(2,3)
    if ( ios < 0 ) then ! end of the file
       box = 0.0d0
       box(1,1) = tmp1
       box(2,2) = tmp2
       box(3,3) = tmp3
    else
       box(1,1) = tmp1
       box(1,2) = tmp2
       box(1,3) = tmp3
       read( ifile, * ) box(3,1), box(3,2), box(3,3)
    endif

    close( ifile )

    ! should only be 1 molecule here
    if ( n_mole .ne. 1 ) then
       stop "Must have one and only one molecule in the *.conf file!"
    endif

  end subroutine read_conf

  !********************************************************
  ! This function find the center of mass of a molecule
  !********************************************************
  function pos_com( xyz, i_mole, n_atom, mass )
    implicit none
    real*8, intent(in), dimension(:,:,:) :: xyz
    real*8, intent(in), dimension(:,:) :: mass
    integer, intent(in) :: i_mole
    integer, intent(in), dimension(:) :: n_atom
    integer :: i_atom
    real*8, dimension(3) :: pos_com
    real*8 :: m_tot
    pos_com(:) = 0D0
    m_tot = 0D0

    do i_atom = 1, n_atom(i_mole)
       pos_com(:) = pos_com(:) + xyz(i_mole,i_atom,:) * mass(i_mole,i_atom)
       m_tot = m_tot + mass(i_mole,i_atom)
    end do
    pos_com(:) = pos_com(:) / m_tot

  end function pos_com

  !*********************************************************
  ! This function update all COM positions
  !*********************************************************
  subroutine update_r_com( xyz, n_mole, n_atom, mass, r_com )
    implicit none
    real*8, intent(in), dimension(:,:,:) :: xyz
    integer, intent(in) :: n_mole
    integer, intent(in), dimension(:) :: n_atom
    real*8, intent(in), dimension(:,:) :: mass
    real*8, intent(out), dimension(:,:) :: r_com
    integer :: i_mole
    do i_mole = 1, n_mole
       r_com( i_mole, : ) = pos_com( xyz, i_mole, n_atom, mass )
    end do
  end subroutine update_r_com



  !***************************************************************************
  ! This function calculate shift vector for PBC
  ! Modified by Kuang on 2012-11-08, to support general box shapes
  !***************************************************************************
  function pbc_shift( r_com_i, r_com_j, box )
  use global_variables
    real*8, dimension(3) :: pbc_shift
    real*8, intent(in), dimension(3) :: r_com_i, r_com_j
    real*8, intent(in), dimension(3,3) :: box
    real*8, dimension(3) :: dr_com, shift, ds_com
    integer :: i
    dr_com(:) = r_com_j(:) - r_com_i(:)
    ds_com(:) = matmul(dr_com, inv_box) ! convert to direct coordinates
    do i = 1, 3
       ds_com(i) = floor( ds_com(i) + 0.5d0 )
    end do
    pbc_shift(:) = matmul(ds_com(:),box(:,:))
  end function pbc_shift

  !******************************************************************
  ! This function calculate dr between two atoms, 
  ! @ith shift vector calcuated by pbc_shift
  !******************************************************************
  function pbc_dr( ri, rj, shift )
    real*8, dimension(3) :: pbc_dr
    real*8, intent(in), dimension(3) :: shift, ri, rj
    pbc_dr(:) = rj(:) - ri(:) - shift(:)
  end function pbc_dr


  !**************************************************************************
  ! this function searches the error function grid, and picks out best value
  ! this assumes all arguments are positive
  !**************************************************************************
  function erfc_g(x)
    use global_variables
    real*8::erfc_g
    real*8,intent(in)::x

    Select case (grid_erfc)
    case("yes")
       if(x > erfc_max) then
          erfc_g=0D0
       else if ( x > 0d0 ) then
          erfc_g=erfc_table(ceiling(x/erfc_max*dble(erfc_grid)))
       else
          erfc_g=1.0d0
       endif
    case("no")
       erfc_g = erfc(x)
    case default
       stop "grid_erfc option not recognized"
    end select

  end function erfc_g

  !**************************************************************************
  ! this function searches the B spline grids, and interpolates B_spline of a
  ! real number from the grids, the accuracy gained from using this subroutine is 
  ! probably not worth the extra time, but this subroutine is mainly for
  ! use when calculating forces numerically using pme energy routine
  !**************************************************************************
  function interp_B_spline(x,n)
    use global_variables
    real*8::interp_B_spline
    real*8,intent(in)::x
    integer,intent(in)::n
    real*8::grid_val
    integer::n_int


    n_int=floor(x*dble(spline_grid)/dble(n))
    grid_val=x*dble(spline_grid)/dble(n)
    if(n_int.eq.0) then
       if(n .eq. 6 ) then
          interp_B_spline= (grid_val-dble(n_int))* B6_spline(n_int+1)
       elseif(n .eq. 5) then
          interp_B_spline= (grid_val-dble(n_int))* B5_spline(n_int+1)
       elseif(n .eq. 4) then
          interp_B_spline= (grid_val-dble(n_int))* B4_spline(n_int+1)
       elseif(n .eq. 3) then
          interp_B_spline= (grid_val-dble(n_int))* B3_spline(n_int+1)
       endif
    else
       if(n .eq. 6 ) then
          interp_B_spline= (1. -(grid_val-dble(n_int)))*B6_spline(n_int) + (grid_val-dble(n_int))* B6_spline(n_int+1)
       elseif(n.eq.5) then
          interp_B_spline= (1. -(grid_val-dble(n_int)))*B5_spline(n_int) + (grid_val-dble(n_int))* B5_spline(n_int+1)
       elseif(n.eq.4) then
          interp_B_spline= (1. -(grid_val-dble(n_int)))*B4_spline(n_int) + (grid_val-dble(n_int))* B4_spline(n_int+1)
       elseif(n.eq.3) then
          interp_B_spline= (1. -(grid_val-dble(n_int)))*B3_spline(n_int) + (grid_val-dble(n_int))* B3_spline(n_int+1)
       endif
    endif


  end function interp_B_spline


  !***************************************************************************
  ! This subroutine generate mass array
  ! we use the index function here because atom types might be numbered
  !***************************************************************************
  subroutine gen_mass(alist,n_atom,mass)
    character(*),dimension(:),intent(in)::alist
    integer,intent(in)::n_atom
    real*8,dimension(:),intent(inout)::mass
    integer::j

    do j=1,n_atom
       if((index(alist(j),'h').eq.1).or.(index(alist(j),'H').eq.1)) then
          if((index(alist(j),'he').eq.1).or.(index(alist(j),'He').eq.1)) then
             mass(j)=4.00
          else
             mass(j)=1.01
          endif
       else if((index(alist(j),'l').eq.1).or.(index(alist(j),'Li').eq.1)) then
          mass(j)=6.94
       else if((index(alist(j),'b').eq.1).or.(index(alist(j),'B').eq.1)) then
          if((index(alist(j),'br').eq.1).or.(index(alist(j),'Br').eq.1)) then
             mass(j)=79.90
          else if((index(alist(j),'be').eq.1).or.(index(alist(j),'Be').eq.1)) then
             mass(j)=9.01
          else
             mass(j)=10.81
          endif
       else if((index(alist(j),'c').eq.1).or.(index(alist(j),'C').eq.1)) then
          if((index(alist(j),'cl').eq.1).or.(index(alist(j),'Cl').eq.1)) then
             mass(j)=35.45
          else if((index(alist(j),'cu').eq.1).or.(index(alist(j),'Cu').eq.1)) then
             mass(j)=63.55
          else if((index(alist(j),'co').eq.1).or.(index(alist(j),'Co').eq.1)) then
             mass(j)=58.93
          else if((index(alist(j),'cr').eq.1).or.(index(alist(j),'Cr').eq.1)) then
             mass(j)=52.00
          else if((index(alist(j),'ca').eq.1).or.(index(alist(j),'Ca').eq.1)) then
             mass(j)=40.08
          else
             mass(j)=12.01
          endif
       else if((index(alist(j),'m').eq.1).or.(index(alist(j),'M').eq.1)) then
          if((index(alist(j),'mg').eq.1).or.(index(alist(j),'Mg').eq.1)) then
             mass(j)=24.31
          else if((index(alist(j),'mn').eq.1).or.(index(alist(j),'Mn').eq.1)) then
             mass(j)=54.94
          endif
       else if((index(alist(j),'n').eq.1).or.(index(alist(j),'N').eq.1)) then
          if((index(alist(j),'ne').eq.1).or.(index(alist(j),'Ne').eq.1)) then
             mass(j)=20.18
          else if((index(alist(j),'na').eq.1).or.(index(alist(j),'Na').eq.1)) then
             mass(j)=22.99
          else if((index(alist(j),'ni').eq.1).or.(index(alist(j),'Ni').eq.1)) then
             mass(j)=58.69
          else
             mass(j)=14.01
          endif
       else if((index(alist(j),'o').eq.1).or.(index(alist(j),'O').eq.1)) then
          mass(j)=16.00
       else if((index(alist(j),'f').eq.1).or.(index(alist(j),'F').eq.1)) then
          if((index(alist(j),'fe').eq.1).or.(index(alist(j),'Fe').eq.1)) then
             mass(j)=55.85
          else
             mass(j)=19.00
          endif
       else if((index(alist(j),'s').eq.1).or.(index(alist(j),'S').eq.1)) then
          if((index(alist(j),'si').eq.1).or.(index(alist(j),'Si').eq.1)) then
             mass(j)=28.09
          else if((index(alist(j),'sc').eq.1).or.(index(alist(j),'Sc').eq.1)) then
             mass(j)=44.96
          else
             mass(j)=32.07
          endif
       else if((index(alist(j),'p').eq.1).or.(index(alist(j),'P').eq.1)) then
          mass(j)=30.97
       else if((index(alist(j),'a').eq.1).or.(index(alist(j),'A').eq.1)) then
          mass(j)=39.95
       else if((index(alist(j),'i').eq.1).or.(index(alist(j),'I').eq.1)) then
          mass(j)=126.9
       else if((index(alist(j),'z').eq.1).or.(index(alist(j),'Z').eq.1)) then
          mass(j)=65.4
       else if((index(alist(j),'v').eq.1).or.(index(alist(j),'V').eq.1)) then
          mass(j)=50.94
       else if((index(alist(j),'t').eq.1).or.(index(alist(j),'Ti').eq.1)) then
          mass(j)=47.87
       endif
    enddo


  end subroutine gen_mass

  !************************************************************************
  ! this subroutine finds the reduced mass of two interacting sites
  ! only used in Feymann-Hibbs quantum correction 
  ! Note, only works for H2 center of mass dummy sites.
  !************************************************************************
  real*8 function reduced_mass( atom_id1, atom_id2, n_atom )
  use global_variables
  implicit none
    integer, intent(in) :: atom_id1, atom_id2
    integer, intent(in), dimension(:) :: n_atom
    integer :: i_atom, atom_id
    real*8 :: m1, m2, m
    
    m1 = atype_mass(atom_id1)
    m2 = atype_mass(atom_id2)

    if ( abs(m1)<1d-6 .or. abs(m2)<1d-6 ) then ! zero mass, means it is a dummy site
      ! Note, oly dummy sites in solute moelcule is allowed, so i_mole always equals to 1
      m = 0d0
      do i_atom = 1, n_atom(1)
        atom_id = atom_index(1,i_atom)
        m = m + atype_mass(atom_id)
      enddo
      if ( abs(m1)<1d-6 ) then
        m1 = m
      else
        m2 = m
      endif
    endif

    reduced_mass = m1*m2/(m1+m2)
    return

  end function reduced_mass

  !************************************************************************
  ! this subroutine seeds the fortran random number generator based on the system clock
  ! based on the gnu documentation
  !************************************************************************
 subroutine initialize_random_seed
   use global_variables
   integer :: i, n, clock
   integer, dimension(:), allocatable :: seed

   call random_seed(size = n)
   allocate(seed(n))
   call system_clock(count=clock)
   seed = clock + 37 * (/ (i-1,i=1,n) /)
   call random_seed(put = seed)

 end subroutine initialize_random_seed



!************************************
! this subroutine does a random rotation on a polyatomic molecule
! no quaternion information is given with this rotation
!***********************************
subroutine random_rotation(xyz,r_com,n_atom, i_mole)
  real*8,dimension(:,:,:),intent(inout) :: xyz
  real*8,dimension(:,:),intent(in) :: r_com
  integer,dimension(:),intent(in) :: n_atom
  integer, intent(in ) :: i_mole

 integer:: i_atom,i,j
    real*8 :: a(3,3),a_trans(3,3), r_vec(3)
    real*8 :: R1,R2,x0,y1,y2
    real*8 :: q0,q1,q2,q3
    real*8,dimension(4) :: quaternion
  real*8, parameter :: pi=3.14159265

    ! Shoemake algorithm for choosing quaternions for random rotation matrix

    call random_number(x0)
    call random_number(y1)
    call random_number(y2)
    y1=y1*2d0*pi
    y2=y2*2d0*pi

    R1=sqrt(1d0-x0)
    R2=sqrt(x0)

    q0=dcos(y2)*R2
    q1=dsin(y1)*R1
    q2=dcos(y1)*R1
    q3=dsin(y2)*R2

    quaternion(1)=q0 ;
    quaternion(2)=q1 ;
    quaternion(3)=q2 ;
    quaternion(4)=q3 ;

    call quaternion_transformation_matrix(a,quaternion )


    do i=1,3
       do j=1,3
          a_trans(i,j) = a(j,i)
       enddo
    enddo

    do i_atom=1, n_atom(i_mole)
       r_vec(:) = xyz(i_mole,i_atom,:) - r_com(i_mole,:)
       r_vec(:) = matmul( a_trans ,r_vec )
       xyz(i_mole,i_atom,:) = r_vec(:) + r_com(i_mole,:)
    enddo


  end subroutine random_rotation


  !************************************
  ! this subroutine creates a transformation matrix from the space fixed coordinate system
  ! to the body centered coordinate system given by the quaternions, q0,q1,q2,q3
  !***********************************
  subroutine quaternion_transformation_matrix(a,quaternion)
    real*8,dimension(3,3),intent(out) :: a
    real*8,dimension(:), intent(in) :: quaternion

    real*8 :: q0,q1,q2,q3

    q0 = quaternion(1) ;  q1 = quaternion(2) ; q2 = quaternion(3) ; q3 = quaternion(4) ;

    a(1,1) = q0**2 + q1**2 - q2**2 - q3**2
    a(2,1) = 2d0 * (q1*q2 - q0*q3)
    a(3,1) = 2d0 * (q1*q3 + q0*q2)
    a(1,2) = 2d0 * (q1*q2 + q0*q3)
    a(2,2) = q0**2 - q1**2 + q2**2 - q3**2
    a(3,2) = 2d0 * (q2*q3 - q0*q1)
    a(1,3) = 2d0 * (q1*q3 - q0*q2)
    a(2,3) = 2d0 * (q2*q3 + q0*q1)
    a(3,3) = q0**2 - q1**2 - q2**2 + q3**2

  end subroutine quaternion_transformation_matrix



  !*************************************************************************
  ! This subroutine generates a random unit vector
  ! from 'computer simulation of liquids', Allen,M.P.,Tildesley,D.J.
  !*************************************************************************
  subroutine random_unit_vector(unit)
    real*8,dimension(3),intent(out)::unit
    real*8::ransq,ran1,ran2,ranh
    real*8,dimension(3)::randnum3
    ransq=2.
    do while (ransq.ge.1)
       call random_number(randnum3)
       ran1=1.-2.*randnum3(1)
       ran2=1.-2.*randnum3(2)
       ransq=ran1*ran1+ran2*ran2
    enddo
    ranh=2.*sqrt(1.-ransq)
    unit(1)=ran1*ranh
    unit(2)=ran2*ranh
    unit(3)=(1.-2.*ransq)

  end subroutine random_unit_vector


  !*************************************************************************
  ! This subroutine moves all spaces at the beginning of a string to the end
  !*************************************************************************
  subroutine trim_head( aname )
    implicit none
    character(*), intent(inout) :: aname
    integer :: i, n, n_space, flag
    n = len( aname )
    n_space = 0
    flag = 0
    do i = 1, n
       if ( flag == 0 .and. aname(i:i) == ' ' ) then
          n_space = n_space + 1
       else if ( flag == 0 ) then
          flag = 1
          aname(i-n_space:i-n_space) = aname(i:i)
       else 
          aname(i-n_space:i-n_space) = aname(i:i)
       end if
    end do
    do i = n-n_space+1, n
       aname(i:i) = ' ' 
    end do
  end subroutine trim_head


  !*************************************************************************
  ! This subroutine moves all spaces at the beginning of a string to the end
  !*************************************************************************
  subroutine trim_end( aname )
    implicit none
    character(*), intent(inout) :: aname
    integer :: i, n, n_space, flag
    n = len( aname )
    n_space=0
    flag = 0
    do i = n, 1, -1
       if ( flag == 0 .and. aname(i:i) == ' ' ) then
          n_space = n_space + 1
       else if ( flag == 0 ) then
          flag = 1
          aname(i+n_space:i+n_space) = aname(i:i)
       else
          aname(i+n_space:i+n_space) = aname(i:i)
       end if
    end do
    do i = 1, n_space
       aname(i:i) = ' '
    end do
  end subroutine trim_end



  !********************************************************
  ! this function is a Tang-Toennies damping function for R6,R8,R10 dispersion interaction
  ! even though it is the same functional form as that for electrostatics, (just a different summation
  ! limit), it is simpler to put this in its own subroutine as the electrostatic screening function 
  ! routine is a bit messy thanks to the drude oscillators
  !********************************************************
  function C6_C10_damp(atom_id1,atom_id2,norm_dr,n)
    use global_variables
    real*8            :: C6_C10_damp
    integer,intent(in):: atom_id1,atom_id2,n
    real*8,intent(in) :: norm_dr

    integer,save :: initialize=0
    real*8,dimension(14),save :: factorial
    integer :: i
    real*8 :: lambda, lambda_i,sum


    ! define factorial if this is first call
    if ( initialize .eq. 0 ) then
       factorial(1)=1d0
       do i=2,size(factorial)
          factorial(i) = dble(i) * factorial(i-1)
       enddo
       initialize=1
    endif

    ! don't damp solute-solute interaction

    if ( (atype_solute(atom_id1) .eq. 1 ) .and. (atype_solute(atom_id2) .eq. 1 ) ) then
       C6_C10_damp = 1d0

    else

       ! buckingham exponent is the second parameter in atype_lj_parameter array, use this for screening
          lambda = atype_lj_parameter(atom_id1,atom_id2,2) * norm_dr

       ! sum = 1 from 0 order term which should be included in damping
       sum=1d0
       lambda_i=1d0
       do i=1,n
          lambda_i=lambda_i*lambda
          sum=sum + lambda_i/factorial(i)
       enddo

       C6_C10_damp = 1d0 - sum * exp(- lambda )

    endif

  end function C6_C10_damp

  !********************************************************
  ! this function is the gradient of the Tang-Toennies damping function for R6,R8,R10 dispersion interaction
  !********************************************************
  function dC6_C10_damp(atom_id1,atom_id2,r_vec,norm_dr,n)
    use global_variables
    real*8,dimension(3) :: dC6_C10_damp
    integer,intent(in):: atom_id1,atom_id2,n
    real*8,dimension(3),intent(in) :: r_vec
    real*8,intent(in) :: norm_dr

    integer,save :: initialize=0
    real*8,dimension(14),save :: factorial
    integer :: i
    real*8 :: exponent, lambda, lambda_i,sum,sum2,Ex_fac
    real*8,dimension(3) :: term1,term2

    ! define factorial if this is first call
    if ( initialize .eq. 0 ) then
       factorial(1)=1d0
       do i=2,size(factorial)
          factorial(i) = dble(i) * factorial(i-1)
       enddo
       initialize=1
    endif

    ! decide whether damping for solute-solute interaction

    if ( (atype_solute(atom_id1) .eq. 1 ) .and. (atype_solute(atom_id2) .eq. 1 ) .and. (damp_solute_solute_dispersion .eq. "no" ) ) then
       dC6_C10_damp = 0d0

    else
       ! buckingham exponent is the second parameter in atype_lj_parameter array, use this for screening
          exponent = atype_lj_parameter(atom_id1,atom_id2,2)

       ! two terms here, mostly cancel to give derivative = exp(- exponent *r) * ( exponent^(n+1)*r^n / n! )

       lambda = exponent * norm_dr

       dC6_C10_damp = exp(-lambda  ) * lambda ** ( n + 1 ) / norm_dr / factorial(n) * r_vec / norm_dr

    endif

  end function dC6_C10_damp

  real*8 function dC6_C10_damp_scalar(atom_id1,atom_id2,norm_dr,n)
  use global_variables
  implicit none
    integer, intent(in):: atom_id1,atom_id2,n
    real*8, intent(in) :: norm_dr
    real*8, dimensioN(3) :: r_vec, gradient

    r_vec = (/1.0d0,0d0,0d0/)
    r_vec = r_vec * norm_dr

    gradient = dC6_C10_damp(atom_id1,atom_id2,r_vec,norm_dr,n)
    dC6_C10_damp_scalar = gradient(1)
    return
  end function dC6_C10_damp_scalar

  !********************************************************
  ! Second derivative of the damping function
  ! This subroutine returns a scalar
  !********************************************************
  real*8 function ddC6_C10_damp(atom_id1,atom_id2,norm_dr,n)
  use global_variables
  implicit none
    integer, intent(in) :: atom_id1, atom_id2, n
    real*8, intent(in) :: norm_dr
    real*8,dimension(14),save :: factorial
    real*8 :: lambda, var
    integer, save :: initialize=0
    integer :: i

    ! define factorial if this is first call
    if ( initialize == 0 ) then
       factorial(1) = 1d0
       do i = 2, size(factorial)
          factorial(i) = dble(i) * factorial(i-1)
       enddo
       initialize = 1
    endif 

    ! decide whether damping for solute-solute interaction
    if ( (atype_solute(atom_id1).eq.1) .and. (atype_solute(atom_id2).eq.1) .and. (damp_solute_solute_dispersion.eq."no" ) ) then
       ddC6_C10_damp = 0d0
    else
       ! buckingham exponent is the second parameter in atype_lj_parameter array, use this for screening
          lambda = atype_lj_parameter(atom_id1,atom_id2,2)

       var = lambda * norm_dr
       ddC6_C10_damp = lambda**2 * exp(-var) * var**n / factorial(n) * ( n/var - 1d0 )
    endif

    return
  end function ddC6_C10_damp



  !********************************************************
  ! this subroutine determines whether the input atoms i_atom, j_atom on
  ! i_mole are either drude oscillators, or have drude oscillators attached to them
  !
  ! this subroutine returns corresponding drude oscillator indexes, signs of charges,
  ! and flags.  
  !
  ! This subroutine is called by intra_pme_energy, intra_pme_force, intra_elec_energy
  ! and intra_elec_force subroutines, so see those subroutines for how i_drude, sign_chg_i, flag_drudei,
  ! etc. are used.
  !********************************************************
  subroutine drude_atom_indices(i_mole,n_atom,i_atom,j_atom,i_drude,j_drude,sign_chg_i,sign_chg_j,flag_drudei,flag_drudej, flag_same_atom)
    use global_variables
    integer,intent(in)::i_mole,i_atom,j_atom
    integer, intent(in), dimension(:) :: n_atom
    integer, intent(out):: i_drude,j_drude,sign_chg_i,sign_chg_j,flag_drudei,flag_drudej,flag_same_atom

    integer :: i_test, j_test

    flag_drudei=0
    flag_drudej=0
    flag_same_atom=0

    ! give i_drude, j_drude, sign_chg_i,sign_chg_j values so that code doesn't get mad at undefined
    ! values
    ! if these variables are not set in the if statements below, then the flags will be set such
    ! that they are not used
    i_drude=0;j_drude=0;sign_chg_i=0;sign_chg_j=0;

  
    ! determine whether i_atom,j_atom are atoms or shells, this is needed because only shell charge is used for intra-molecular 
    ! screened interactions, and this charge needs to be taken from the corresponding drude_oscillator

    ! make sure that j_atom is not the drude_oscillator attached to i_atom or vice-versa
    ! use flag_same_atom to check this

    ! use drude_atom_map, iff drude_atom_map(i_mole,i_atom) = i_atom, then this is not a drude oscillator

    if ( drude_atom_map(i_mole,i_atom ) .eq. i_atom ) then
       ! this is not a drude oscillator, see if this atom has a shell on it
       sign_chg_i=-1
       ! shell indices are always larger than atom indices, so loop from i_atom + 1
       if ( i_atom < n_atom(i_mole) ) then
       do i_test=i_atom+1, n_atom(i_mole)
          if ( drude_atom_map(i_mole,i_test ) .eq. i_atom ) then
             ! found drude oscillator
             i_drude = i_test
             flag_drudei = 1
          endif
       enddo
       endif
    else
       ! this is a drude oscillator
       i_drude = i_atom
       flag_drudei = 1
       sign_chg_i=1
       ! make sure this drude oscillator is not attached to j_atom
       if ( drude_atom_map(i_mole,i_atom ) .eq. j_atom ) then
          flag_same_atom=1
       endif
    endif


    if ( drude_atom_map(i_mole,j_atom ) .eq. j_atom ) then
       ! this is not a drude oscillator, see if this atom has a shell on it
       sign_chg_j=-1
       ! shell indices are always larger than atom indices, so loop from j_atom + 1
       if ( j_atom < n_atom(i_mole) ) then
       do j_test=j_atom+1, n_atom(i_mole)
          if ( drude_atom_map(i_mole,j_test ) .eq. j_atom ) then
             ! found drude oscillator
             j_drude = j_test
             flag_drudej = 1
          endif
       enddo
       endif
    else
       ! this is a drude oscillator
       j_drude = j_atom
       flag_drudej = 1
       sign_chg_j=1
       ! make sure this drude oscillator is not attached to i_atom
       if ( drude_atom_map(i_mole,j_atom ) .eq. i_atom ) then
          flag_same_atom=1
       endif
    endif


  end subroutine drude_atom_indices


  !****************************************************
  !
  !  IMPORTANT: This subroutine should always be called, even if this is not
  !             a Drude oscillator simulation, as the array drude_atom_map will be used
  !             in electrostatic screening functions for all atoms
  ! 
  !  Framework atoms are stored in locations from n_mole to tot_n_mole, and we do not
  !  use drude oscillators on the framework, but still initialize those parts of arrays
  !
  !  This subroutine initializes the drude oscillator data
  !  The tot_chg array is modified to reflect the drude oscillator charges
  !  drude_atom_map, stored as a global variable, is initialized
  !  which maps which atom each drude oscillator is bound to
  !  Also, n_atom_drude is also constructed, which stores how many total atoms and drudes
  !  are on each molecule (for looping over), compared to n_atom , which is just atoms
  !****************************************************
  subroutine initialize_drude_oscillators(tot_n_mole, n_mole, n_atom, n_atom_drude, tot_chg , chg_drude)
    use global_variables
    integer,intent(in) :: tot_n_mole, n_mole
    integer,dimension(:),intent(in) :: n_atom
    integer,dimension(:),intent(inout) :: n_atom_drude
    real*8,dimension(:,:),intent(inout) :: tot_chg
    real*8,dimension(:,:),intent(in) :: chg_drude

    real*8,parameter::small=1D-6

    integer :: i_mole, i_atom, index_drude

    ! first take care of framework atoms, (just loop over all atoms here), and 
    ! fill in trivial map for atoms
    do i_mole=1, tot_n_mole
       do i_atom=1,n_atom(i_mole)
          ! trivial map from atom to atom
          drude_atom_map(i_mole,i_atom) = i_atom
       enddo
    enddo

    ! now take care of any drude oscillators, note we use n_mole here, so that we're not
    ! looping over framework atoms, since these don't have oscillators

    do i_mole=1,n_mole
       ! this is the starting index for Drude oscillators for this molecule
       index_drude = n_atom(i_mole) + 1
       do i_atom=1,n_atom(i_mole)

          if ( abs( chg_drude(i_mole,i_atom) ) > small ) then

	! lets not confuse anyone here, make sure drude_simulation is set to 1
	Select Case(drude_simulation)
	 Case(0)
	  write(*,*) ""
	  write(*,*) "There is a non zero polarizability on atom ", i_atom
	  write(*,*) "of molecule ", i_mole, " but drude_simulation is set"
	  write(*,*) " to ", drude_simulation
	  write(*,*) "if you want to run a Drude oscillator simulation, please"
	  write(*,*) "set drude_simulation to 1"
	  write(*,*) ""
	  stop
	End Select

             ! there is a drude oscillator on this atom, index this drude oscillator
             ! in the next empty spot in drude_atom_map

             drude_atom_map(i_mole,index_drude) = i_atom

             ! fix tot_chg array for both atom, and drude
             tot_chg(i_mole,i_atom) = tot_chg(i_mole,i_atom) - chg_drude(i_mole,i_atom)
             tot_chg(i_mole,index_drude)=chg_drude(i_mole,i_atom)

             index_drude = index_drude + 1

                ! make sure arrays are big enough to support these drude oscillators
                   if ( index_drude > MAX_N_ATOM ) then   
                      write(*,*) " "
                      write(*,*) " for Drude oscillator simulations, must have MAX_N_ATOM set to at least the"
                      write(*,*) " maximum number of atoms and drude oscillators in the largest molecule."
                      write(*,*) " Currently, this is not the case."
                      write(*,*) " Please increase this value in the global variables file"
                   stop 
                endif

          endif

       enddo
       ! total number of atoms and drude oscillators on this molecule
       n_atom_drude(i_mole) = index_drude - 1

    enddo


  end subroutine initialize_drude_oscillators
    

 !********************************************************
  ! this function is a Tang-Toennies screening function for coulombic interactions
  ! when it is turned on, and "1.0" when it is off.  Currently, it is meant to use exponents
  ! from buckingham potential as parameters.  For drude oscillators, it takes parameters from
  ! corresponding atoms
  ! 
  !  we use the map drude_atom_map, to link Drude oscillators
  ! to their corresponding atoms.  We also use this map trivially for atoms, and so it 
  ! should be constructed regardless of whether this is a drude oscillator simulation
  !*******************************************************
  function screen(i_mole,j_mole,i_atom,j_atom,n_atom,norm_dr,chg)
    use global_variables
    real*8           :: screen
    integer,intent(in):: i_mole, j_mole, i_atom, j_atom
    integer,dimension(:),intent(in) :: n_atom
    real*8,intent(in) :: norm_dr
    real*8,dimension(:,:),intent(in) :: chg


    integer :: p_atom1, p_atom2, atom_id1,atom_id2
    real*8  :: lambda

    if(screen_type .eq. 0 ) then

       screen = 1D0

    else if (screen_type .eq. 1 ) then

       ! these atoms may or may not be drude oscillators, either way, we can use the drude_atom_map array
       ! which will index both drude oscillators as well as regular atoms

       ! parent atom for drude oscillator ( or just atom )
       p_atom1 = drude_atom_map(i_mole,i_atom)
       p_atom2 = drude_atom_map(j_mole,j_atom)

       atom_id1 = atom_index(i_mole,p_atom1)
       atom_id2 = atom_index(j_mole,p_atom2)


       ! buckingham exponent is the second parameter in atype_lj_parameter array, use this for screening

       lambda = atype_lj_parameter(atom_id1,atom_id2,2)
       
       screen=1D0-(1D0 + lambda * norm_dr)*exp(- lambda * norm_dr)

    else
       stop "screen type not implemented"

    endif

  end function screen
 

  !********************************************************
  ! this function is the gradient of the above screening function
  ! see comments to function above
  !*******************************************************
  function d_screen(i_mole,j_mole,i_atom,j_atom,n_atom,r_ij,r_mag,chg)
    use global_variables
    real*8,dimension(3) :: d_screen
    integer,intent(in):: i_mole, j_mole, i_atom, j_atom
    integer,dimension(:),intent(in) :: n_atom
    real*8,dimension(3),intent(in) :: r_ij
    real*8, intent(in)     :: r_mag
    real*8,dimension(:,:),intent(in) :: chg

    integer :: p_atom1, p_atom2,atom_id1,atom_id2
    real*8  :: lambda,Exponential

    if(screen_type .eq. 0 ) then

       d_screen = 0D0

    else if (screen_type .eq. 1 ) then

       ! these atoms may or may not be drude oscillators, either way, we can use the drude_atom_map array
       ! which will index both drude oscillators as well as regular atoms

       ! parent atom for drude oscillator ( or just atom )
       p_atom1 = drude_atom_map(i_mole,i_atom)
       p_atom2 = drude_atom_map(j_mole,j_atom)


       atom_id1 = atom_index(i_mole,p_atom1)
       atom_id2 = atom_index(j_mole,p_atom2)

       ! buckingham exponent is the second parameter in atype_lj_parameter array, use this for screening
       lambda = atype_lj_parameter(atom_id1,atom_id2,2)

       Exponential = exp(-lambda*r_mag)

       d_screen = (r_ij/r_mag) * (lambda * (1D0 + lambda * r_mag) * Exponential - lambda*Exponential )


    endif

  end function d_screen


  !*******************************************************
  ! this function is a screening function for intermolecular induced dipole 
  ! (drude oscillator) interactions
  ! for our N2 model, the center site has no drude oscillator, however the 
  ! code treats the whole molecule as either all atoms having drude oscillators
  ! or none having drude oscillators.  Therefore we have to consider the case with
  ! a zero charge drude oscillator, so don't divide by zero
  !*******************************************************
  function thole_screen(pol1,pol2,r_ij,thole)
    real*8::thole_screen
    real*8,intent(in)::r_ij,pol1,pol2,thole
    real*8::a
    real*8,parameter :: drude_chg_thresh=1d-8

    a=thole

    ! don't divide by zero
    if ( ( pol1 > drude_chg_thresh ) .and. ( pol2 > drude_chg_thresh ) ) then
    thole_screen=1.D0-(1.D0+(a*r_ij)/(2.D0*(pol1*pol2)**(1.D0/6.D0)))*exp(-a*r_ij/(pol1*pol2)**(1.D0/6.D0))
    else
    ! this interaction is with a zero charge, so will be zero, just set screen to 1
    thole_screen=1d0
    endif

  end function thole_screen


  function  d_thole_screen(pol1,pol2,xyz,thole)
    real*8,dimension(3)::d_thole_screen
    real*8,intent(in)::pol1,pol2,thole
    real*8,dimension(3),intent(in)::xyz
    real*8::r_ij,fac,Ex,a
    real*8,parameter :: drude_chg_thresh=1d-8

    a=thole

    ! don't divide by zero
    if ( ( pol1 > drude_chg_thresh ) .and. ( pol2 > drude_chg_thresh ) ) then

    r_ij=sqrt(dot_product(xyz,xyz))
    fac=a/(pol1*pol2)**(1.D0/6.D0)
    Ex=exp(-fac*r_ij)

    d_thole_screen=(xyz/r_ij)*(fac*(1.D0+fac*r_ij/2.D0)*Ex-(fac/2.D0)*Ex)

    else
    d_thole_screen=0d0
    endif

  end function d_thole_screen


  subroutine crossproduct( a,b,ans )
    implicit none
    real*8, intent(in) :: a(3),b(3)
    real*8, intent(out) :: ans(3)
    ans(1) = a(2)*b(3)-a(3)*b(2)
    ans(2) = -a(1)*b(3)+a(3)*b(1)
    ans(3) = a(1)*b(2)-a(2)*b(1)
  end subroutine crossproduct

  real function volume( a, b, c )
    implicit none
    real*8, intent(in) :: a(3), b(3), c(3)
    volume = a(1) * (b(2)*c(3)-b(3)*c(2)) - a(2) * (b(1)*c(3)-b(3)-c(1)) + a(3) * (b(1)*c(2)-b(2)*c(1))
    volume = abs(volume)
  end function volume

!*************************************************************************************
! This subroutine check whether the guest molecule overlap with the framework or not
! Cell link list is used to accelerate the calculation
! Modified in 2012-09-14, general box shape compatible
!*************************************************************************************
  integer function overlap_frame( xyz, tot_n_mole, n_mole, n_atom, i_mole, box, r_com, headlist, nslist )
  use global_variables
  implicit none
    real*8, dimension(:,:), intent(in) :: box, r_com
    real*8, dimension(:,:,:), intent(in) :: xyz
    integer, intent(in) :: n_mole, tot_n_mole, i_mole
    integer, dimension(:), intent(in) :: n_atom
    integer, dimension(na_nslist,nb_nslist,nc_nslist,2), intent(in) :: headlist
    integer, dimension(MAX_N_MOLE,MAX_N_ATOM,2), intent(in) :: nslist
!   local variables
    integer :: j_mole, j_atom, tmp_mole, tmp_atom, i_atom
    integer :: ia,ib,ic, ia0,ib0,ic0, dia,dib,dic, igrida, igridb, igridc
    real*8 :: ra, rb, rc, rka, rkb, rkc
    real*8, dimension(3) :: r_com_i, r_com_j, shift, dr_com, rij, rtmp

    r_com_i(:) = r_com(i_mole,:)

    rka = dsqrt(dot_product(inv_box(:,1),inv_box(:,1)))
    rkb = dsqrt(dot_product(inv_box(:,2),inv_box(:,2)))
    rkc = dsqrt(dot_product(inv_box(:,3),inv_box(:,3)))

    dia = floor(too_close*rka*dble(na_nslist)) + 1
    dib = floor(too_close*rkb*dble(nb_nslist)) + 1
    dic = floor(too_close*rkc*dble(nc_nslist)) + 1    
    
    overlap_frame = 0
    do i_atom = 1, n_atom(i_mole)
      rtmp(:) = matmul(xyz(i_mole,i_atom,:),inv_box(:,:))
      ra = rtmp(1) * na_nslist
      rb = rtmp(2) * nb_nslist
      rc = rtmp(3) * nc_nslist
!     shift into the box
      ra = ra - floor(ra/na_nslist)*na_nslist
      rb = rb - floor(rb/nb_nslist)*nb_nslist
      rc = rc - floor(rc/nc_nslist)*nc_nslist
      ia0 = floor(ra)
      ib0 = floor(rb)
      ic0 = floor(rc)
      do ia = -dia, dia
        igrida = ia0 + ia
        igrida = igrida - floor(dble(igrida)/dble(na_nslist))*na_nslist
        igrida = igrida + 1
        do ib = -dib, dib
          igridb = ib0 + ib
          igridb = igridb - floor(dble(igridb)/dble(nb_nslist))*nb_nslist
          igridb = igridb + 1
          do ic = -dic, dic
            igridc = ic0 + ic
            igridc = igridc - floor(dble(igridc)/dble(nc_nslist))*nc_nslist
            igridc = igridc + 1
            j_mole = headlist(igrida, igridb, igridc, 1)
            j_atom = headlist(igrida, igridb, igridc, 2)
            do while ( j_mole /= 0 )
              r_com_j(:) = r_com(j_mole,:)
              shift(:) = pbc_shift( r_com_i, r_com_j, box )
              rij = pbc_dr( xyz(i_mole,i_atom,:), xyz(j_mole,j_atom,:), shift )
              if ( abs(rij(1))<too_close .and. abs(rij(2))<too_close .and. abs(rij(3))<too_close .and. &
             &     dot_product(rij,rij) < too_close**2 ) then
                overlap_frame = 1
                goto 999
              endif
              tmp_mole = nslist(j_mole, j_atom, 1)
              tmp_atom = nslist(j_mole, j_atom, 2)
              j_mole = tmp_mole
              j_atom = tmp_atom
            enddo
          enddo
        enddo
      enddo
    enddo
999 continue

    return
    end function overlap_frame

  !**************************************************************************
  ! function searches g function grid, which is used in pme dispersion calc
  ! this assumes all arguments are positive
  !**************************************************************************
  function gfun_g(x,p)
    use global_variables
    real*8 :: gfun_g
    real*8, intent(in) :: x
    integer, intent(in) :: p
    
    Select case (grid_erfc)
    case("yes") 
       if ( x > gfun_max) then
          gfun_g = 0d0
       else
          select case(p)
          case(6)
            gfun_g = g6_table(ceiling(x/gfun_max*dble(gfun_grid)))
          case(8)
            gfun_g = g8_table(ceiling(x/gfun_max*dble(gfun_grid)))
          case(10) 
            gfun_g = g10_table(ceiling(x/gfun_max*dble(gfun_grid)))
          case(12) 
            gfun_g = g12_table(ceiling(x/gfun_max*dble(gfun_grid)))
          case(14)
            gfun_g = g14_table(ceiling(x/gfun_max*dble(gfun_grid)))
          case default
            stop "gfun_g option not recognized"
          end select
       endif
     case("no")
        select case(p)
        case(6)
           gfun_g = (x**4/2d0+x**2+1d0)*dexp(-x**2)
        case(8)
           gfun_g = (x**6+3d0*x**4+6d0*x**2+6d0)/6d0*dexp(-x**2)
        case(10)
           gfun_g = (x**8+4d0*x**6+12d0*x**4+24d0*x**2+24d0)/24d0*dexp(-x**2)
        case(12)
           gfun_g = (x**10+5d0*x**8+20d0*x**6+60d0*x**4+120d0*x**2+120d0)/120d0*dexp(-x**2)
        case(14)
           gfun_g = (x**12+6d0*x**10+30d0*x**8+120d0*x**6+360d0*x**4+720*x**2+720d0)/720d0*dexp(-x**2)
        case default
           stop "gfun_g option not recognized"
        end select
     case default
        stop "gfun_g option not recognized"
     end select

  end function gfun_g

  function wfun(u)
  use global_variables
  implicit none
    real*8 :: wfun
    real*8, intent(in) :: u
    integer :: j, k, p

    p = lgrg_order
    if ( u <= -p .or. u >= p ) then
      wfun = 0.0d0
      return
    endif
    k = floor(u)
    j = -p
    wfun = 1.0d0
    do while ( j <= p-1 )
      if ( j /= k ) then
        wfun = wfun * dble(u+j-k)/dble(j-k)
      endif
      j = j + 1
    enddo

    return
  end function wfun


 !**************************************************************************
  ! Derivative of the Lagrangian weighing function
  ! To save time, it also gives the weighing function value
  !
  ! this was modified by Jesse 6/13/2014.  Before, if u was an integer, there were
  ! rare cases when the old code had a 0/0 expression.
  !**************************************************************************
  subroutine dwfun(u,w,dw)
  use global_variables
  implicit none
    real*8, intent(out) :: w, dw
    real*8, intent(in) :: u
    integer :: i , j, k, p, index
    real*8,dimension(2*lgrg_order) :: deriv_term

    deriv_term=1d0

    p = lgrg_order
    dw = 0.0d0
    if ( u <= -p .or. u >= p ) then
      w = 0.0d0
      dw = 0.0d0
      return
    endif

    k = floor(u)
    j = -p
    w = wfun(u)

    ! old code.  Commented 6/13/2014
    !do while ( j<= p-1 )
    !  if ( j /= k ) then
    !    dw = dw + w / dble(u+j-k)
    !  endif
    !  j = j + 1
    !enddo

    index=1
    do while ( j <= p-1 )
      if ( j /= k ) then
         do i=1, size(deriv_term)
            if ( i /= index ) then
               deriv_term(i) = deriv_term(i) * dble(u+j-k)/dble(j-k)
            else
               deriv_term(i) = deriv_term(i) / dble(j-k)
            endif
         enddo
         index = index + 1
      endif
      j = j + 1
    enddo

    do i=1, index-1
       dw = dw + deriv_term(i)
    enddo

    return
  end subroutine dwfun


  !******************************************************
  !        Here are some string manipulation routines 
  !        written by Dr. George Benthien and taken from
  !        http://www.gbenthien.net/strings/index.html
  !******************************************************

  subroutine parse(str,delims,args,nargs)

    ! Parses the string 'str' into arguments args(1), ..., args(nargs) based on
    ! the delimiters contained in the string 'delims'. Preceding a delimiter in
    ! 'str' by a backslash (\) makes this particular instance not a delimiter.
    ! The integer output variable nargs contains the number of arguments found.

    character(len=*) :: str,delims
    character(len=len_trim(str)) :: strsav
    character(len=*),dimension(:) :: args
    integer,intent(out) ::nargs
    integer :: i,k,na,lenstr


    strsav=str
    call compact(str)
    na=size(args)
    do i=1,na
       args(i)=' '
    end do
    nargs=0
    lenstr=len_trim(str)
    if(lenstr==0) return
    k=0

    do
       if(len_trim(str) == 0) exit
       nargs=nargs+1
       call split(str,delims,args(nargs))
       call removebksl(args(nargs))
    end do
    str=strsav

  end subroutine parse



  subroutine compact(str)

    ! Converts multiple spaces and tabs to single spaces; deletes control characters;
    ! removes initial spaces.

    character(len=*):: str
    character(len=1):: ch
    character(len=len_trim(str)):: outstr
    integer :: i,k,ich,isp,lenstr

    str=adjustl(str)
    lenstr=len_trim(str)
    outstr=' '
    isp=0
    k=0

    do i=1,lenstr
       ch=str(i:i)
       ich=iachar(ch)

       select case(ich)

       case(9,32)     ! space or tab character
          if(isp==0) then
             k=k+1
             outstr(k:k)=' '
          end if
          isp=1

       case(33:)      ! not a space, quote, or control character
          k=k+1
          outstr(k:k)=ch
          isp=0

       end select

    end do

    str=adjustl(outstr)

  end subroutine compact


  subroutine split(str,delims,before,sep)

    ! Routine finds the first instance of a character from 'delims' in the
    ! the string 'str'. The characters before the found delimiter are
    ! output in 'before'. The characters after the found delimiter are
    ! output in 'str'. The optional output character 'sep' contains the 
    ! found delimiter. A delimiter in 'str' is treated like an ordinary 
    ! character if it is preceded by a backslash (\). If the backslash 
    ! character is desired in 'str', then precede it with another backslash.

    character(len=*) :: str,delims,before
    character,optional :: sep
    logical :: pres
    character :: ch,cha
    integer:: i,k,lenstr,ibsl,ipos,iposa

    pres=present(sep)
    str=adjustl(str)
    call compact(str)
    lenstr=len_trim(str)
    if(lenstr == 0) return        ! string str is empty
    k=0
    ibsl=0                        ! backslash initially inactive
    before=' '
    do i=1,lenstr
       ch=str(i:i)
       if(ibsl == 1) then          ! backslash active
          k=k+1
          before(k:k)=ch
          ibsl=0
          cycle
       end if
       if(ch == '\') then          ! backslash with backslash inactive
          k=k+1
          before(k:k)=ch
          ibsl=1
          cycle
       end if
       ipos=index(delims,ch)
       if(ipos == 0) then          ! character is not a delimiter
          k=k+1
          before(k:k)=ch
          cycle
       end if
       if(ch /= ' ') then          ! character is a delimiter that is not a space
          str=str(i+1:)
          if(pres) sep=ch
          exit
       end if
       cha=str(i+1:i+1)            ! character is a space delimiter
       iposa=index(delims,cha)
       if(iposa > 0) then          ! next character is a delimiter
          str=str(i+2:)
          if(pres) sep=cha
          exit
       else
          str=str(i+1:)
          if(pres) sep=ch
          exit
       end if
    end do
    if(i >= lenstr) str=''
    str=adjustl(str)              ! remove initial spaces
    return

  end subroutine split

  !**********************************************************************

  subroutine removebksl(str)

    ! Removes backslash (\) characters. Double backslashes (\\) are replaced
    ! by a single backslash.

    character(len=*):: str
    character(len=1):: ch
    character(len=len_trim(str))::outstr
    integer :: i,k,ibsl,lenstr

    str=adjustl(str)
    lenstr=len_trim(str)
    outstr=' '
    k=0
    ibsl=0                        ! backslash initially inactive

    do i=1,lenstr
       ch=str(i:i)
       if(ibsl == 1) then          ! backslash active
          k=k+1
          outstr(k:k)=ch
          ibsl=0
          cycle
       end if
  if(ch == '\') then          ! backslash with backslash inactive
   ibsl=1
   cycle
  end if
  k=k+1
  outstr(k:k)=ch              ! non-backslash with backslash inactive
end do

str=adjustl(outstr)

end subroutine removebksl

!**********************************************************************

end module routines
