!************************************************************
! This module contains global variables that are used in many subroutines
!
! With a few exceptions, the data in this module doesn't change once initialized.
!
!
! The initialization of this global data occurs in multiple ways
!
! 1) some of this data is hard-coded in.  Any parameters that are hard coded in that the user might want to change are
!    placed at the top of the module, as labeled
!
! 2) some of the technical simulation parameter data is read in from the simulation_parameters.pmt input file
!
! 3) The force field parameters are stored here as global data, and are read in from the force field input file
!
! 4) Finally, some of the mapping arrays are constructed logically by the code.
!
!*************************************************************

module global_variables
  implicit none

  !***************************** NOTE THIS SECTION ***********************************************
  ! here we organize all global variable parameters that are hard-coded in, that the user
  ! may want to adjust in certain cases

  ! these parameters control the size of data arrays.  Make sure MAX_N_MOLE > maximum number of molecules in the simulations
  ! and MAX_N_ATOM > maximum number of atoms per molecule
  integer, parameter :: MAX_FN=100, MAX_ANAME=5, MAX_MNAME=5, MAX_N_MOLE=10000, MAX_N_ATOM=35


  ! these parameters control the Drude oscillator convergence:  Maximum number of iterations and convergence threshold
  integer, parameter :: iteration_limit= 12     ! if scf iterations exceed this, probably polarization catastrophe, so reject move
  real*8, parameter :: force_threshold = 4.0477D-6  !4.0477D-6    this is for rms force in e^2/A^2, equal to .1 KJ/mol/nm


  ! dhf combination rule:  As we used sort of a weird convention for the sign of the dhf cross terms in thebeginning, but changed to an completely attractive dhf cross term convention when fitting virials, we need to implement a variety of combination rules
  ! here "1" is as in JPC C, 2012, 116, 1892-1903
  ! "2" is Aij = positive iff Aii and Ajj positive (difference between 1 and 2 is Aij negative in 2 if Aii and Ajj both negative
  ! "3" is always attractive
  integer, parameter :: dhf_combination_rule=2

  !***********************************************************************************************



  real*8, parameter :: pi=3.14159265
  real*8, parameter :: pi_sqrt=1.7724538509055159d0

  integer, parameter :: N_input_f = 6   ! number of input files
  integer            :: framework_simulation      ! tells subroutines whether or not this simulation includes fixed framework atoms

  ! code version
  integer  :: code_version = 21



  ! input file type is determined by whether or not energy decomposition is requested
  character(3) :: energy_decomposition ! "no" for Lennard Jones ff, "yes" for SAPT based force fields
  character(3) :: solute_cross_parameter_set

  ! penetration force field for short distances , this will only be used with energy decomposition
  character(3) :: penetration_force_field = "no"
  character(3) :: switch_at_minimum = "yes"
  real*8 :: penetration_force_field_threshold

  ! if, energy decomposition, have option of using C8,C10, C12  parameters for dispersion.  This also determines input file structure
  character(3) :: C8_10_dispersion_terms ! if yes, damping for dispersion will be turned on, and buckingham coefficients for dispersion will be read in
  character(3) :: C12_dispersion
  ! explicitly read solute-framework exponents
  character(3) :: explicit_cross_term_exponents="no"
  ! this shouldn't be used, but we'll keep it in so that damping functions are consistent with MC code
  character(3) :: damp_solute_solute_dispersion="yes"



  !force field parameter data storage
  integer:: n_atom_type, n_molecule_type
  character(MAX_ANAME), dimension(MAX_N_ATOM) :: atype_name
  character(MAX_MNAME), dimension(MAX_N_MOLE) :: molecule_name
  real*8, dimension(MAX_N_ATOM) :: atype_chg,atype_pol
  real*8, dimension(MAX_N_ATOM) :: atype_mass    
  ! this stores "A" type coefficients for Bkingham, index 1-5 are exch,elec,induc,dhf,dispersion, respectively
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM,5) :: atype_bkghm_decomp
  ! for buckingham, store A,B,C6 (possibly C8,C10), for LJ, store epsilon,sigma
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM,6) :: atype_lj_parameter
  ! penetration force field, first index is bkghm coeff, 2 index is distance cutoff, 3 index is energy shift
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM,3) :: atype_penetration_ff      
  ! keeps track of whether this atom type is framework(0) or solute(1)
  integer, dimension(MAX_N_ATOM) ::  atype_solute 
  ! this array contains indices for all types of solute molecules (not framework) end is marked with MAX_N_ATOM+1
  integer, dimension(MAX_N_MOLE,MAX_N_ATOM) :: molecule_type 
  character(MAX_MNAME), dimension(MAX_N_MOLE) :: molecule_type_name
  real*8, dimension(MAX_N_MOLE,MAX_N_ATOM,3) :: r_mole_com
  integer, dimension(MAX_N_MOLE) :: molecule_shape      ! flag "0" for nonlinear, polyatomic, "1" for linear molecules, "2" for atom
  ! integer specifies where to look up atom parameters in lj_parameter array
  integer, dimension(MAX_N_MOLE,MAX_N_ATOM):: atom_index  
  integer, dimension(MAX_N_MOLE):: molecule_index  
  integer, dimension(MAX_N_MOLE,MAX_N_ATOM):: drude_atom_map            ! this array maps which atom each Drude oscillator resides on
  integer                 :: lj_bkghm      ! =1 for bkghm, =2 for lj, this is determined by parameter input file
  ! for lj, set to "opls" or "standard"== Lorentz-Berthelot, for bkghm, set to "standard" or "ZIFFF"
  character(10)            :: lj_comb_rule 


  ! simulation parameters
  real*8 :: temp ! temp in K
  real*8 :: too_close         ! if any atoms on different molecules closer than this, reject move right away
  real*8  :: lj_cutoff

  ! parameters for Electrostatic (Ewald/Cutoff/pme)
  real*8 :: alpha_sqrt
  real*8 :: lj_asqrt
  real*8 :: ewald_cutoff
  integer :: screen_type            ! screening for coulomb potential 0 = none, 1 = Tang-Toennies type
  character(10):: electrostatic_type ! choose from pme, cutoff, or none. ewald is only implemented for testing
  integer :: pme_grid
  integer, parameter :: spline_order=6
  integer, parameter :: lgrg_order=2 ! if this equals 1, then the grid is interpolated linearly
  !  real*8, dimension(:,:,:), allocatable :: CB,Q_grid,theta_conv_Q
  character(3) :: grid_erfc = "yes"        ! no or yes
  integer, parameter :: spline_grid=1000000,erfc_grid=1000000, gfun_grid=500000
  real*8, parameter :: erfc_max=10d0              ! this is max value up to which erfc is grid
  real*8, parameter :: gfun_max=5.0d0
  real*8, dimension(spline_grid) :: B6_spline,B5_spline,B4_spline,B3_spline
  real*8, dimension(erfc_grid) :: erfc_table
  real*8, dimension(gfun_grid) :: g6_table, g8_table, g10_table, g12_table, g14_table
  character(3) :: Qeq_charges  ! this determines whether we compute framework charges using QEq scheme

  ! parameters for drude oscillator
  real*8 :: springcon 
  real*8 :: thole 
  integer :: drude_simulation       ! this flag will let sampling routines know whether to call scf drude (=1 for drudes)
  ! for parallel openmp
  integer :: n_threads



  ! these parameters are for the free energy grid 
  ! in KJ/mol, this is threshold for low energy treatment, which increases number of orientations, and uses polarization
  integer :: orientation_try
  integer :: cav_grid_a, cav_grid_b, cav_grid_c
  ! these variables are used iff a partial free energy grid is being computed
  integer :: start_grid_a, finish_grid_a, start_grid_b, finish_grid_b, start_grid_c, finish_grid_c
  integer :: partial_grid_computation   ! 0 if full grid, 1 if partial grid
  real*8           :: cubelet_vol   ! calculated from box size and grid size
  ! in KJ/mol, this is threshold for low energy treatment, which increases number of orientations, and uses polarization
  real*8, parameter :: Epol_thresh=-50.  
  real*8, parameter :: fine_grid_thresh=-50. ! threshold for fine grid
  integer, parameter :: fine_grid_points=4 ! per dimension

  ! link list parameters
  integer :: na_nslist, nb_nslist, nc_nslist

  ! fancy sample method
  character(3),parameter :: vegas="no"
  real*8 :: REL_THRSH, ABS_THRSH, BZ_CUTOFF
  ! real*8, parameter :: REL_THRSH=0.10d0, ABS_THRSH=10.0d0, BZ_CUTOFF=100d0
  ! real*8, parameter :: REL_THRSH=0.05d0, ABS_THRSH=3.0d0, BZ_CUTOFF=100d0
  ! real*8, parameter :: REL_THRSH=0.01d0, ABS_THRSH=1.0d0, BZ_CUTOFF=100d0 ! High accuracy

  ! box shape information
  real*8, dimension(3,3) :: inv_box

  ! quantum correction
  integer :: quantum_correct=0

end module global_variables
