module initialization_routines
  use routines
  use sapt_ff_routines
  implicit none

contains

  !*************************************************************************
  ! this routine reads the input files for the solute, and if present, the framework
  ! and initializes the simulation
  !*************************************************************************
  subroutine initialize_simulation(box, alist, n_mole, tot_n_mole, n_atom, n_atom_drude, xyz, mass, &
          & mass_mol, r_com, chg, chg_drude, tot_chg, ifile_conf, ifile_fconf, ifile_pmt)

    use global_variables 
    use pme_routines
    use penetration_ff_routines
    character(MAX_FN),intent(in) :: ifile_conf,ifile_fconf,ifile_pmt
    character(MAX_ANAME), dimension(MAX_N_MOLE,MAX_N_ATOM),intent(out) :: alist
    integer,intent(out)::n_mole,tot_n_mole
    integer,dimension(:),intent(out)::n_atom, n_atom_drude
    real*8,dimension(:,:,:),intent(out)::xyz
    real*8,dimension(:),intent(out)::mass_mol
    real*8,dimension(:,:),intent(out)::mass,chg,chg_drude,tot_chg
    real*8,dimension(:,:),intent(out)::box,r_com
    integer, dimension(MAX_N_ATOM,MAX_N_ATOM) :: gen_cross_terms
    real*8,dimension(MAX_N_ATOM,MAX_N_ATOM,9) :: temp_lj_parameter
    real*8,dimension(MAX_N_MOLE,MAX_N_ATOM) :: temp_chg
    integer:: i,i_mole,i_atom,n_type_s, flag_charge
    real*8 :: unit_cell(3),sum_chg
    real*8,parameter :: mass_frame = 100.,framework_chg_thresh = 1D-6

    ! initialize random seed, this will be overwritten if seed is read in in read_conf routine
!    call initialize_random_seed

    !************************************** get parameters****************************************************!
    ! get parameters for all atom types, assuming there are no more than MAX_N_ATOM types of atoms total
    ! remember no two atom names can conflict, not even those of framework and solute
    !*****************************************************************************************************!

    ! if energy decomposition need to read in more complicated parameter file.  
    Select Case(energy_decomposition)       ! this determines input file type
    Case("no")
       call read_param( ifile_pmt,gen_cross_terms)
    Case("yes")
       ! temp_lj_parameter array is used to temporarily store solute-solute parameters
       ! if these are different than solute-framework
       call read_param_decomp( ifile_pmt , temp_lj_parameter, n_type_s )
    end Select
    ! make sure we don't have too many atom types
    if ( n_atom_type > MAX_N_ATOM ) then
       stop "number of atom types g.t. MAX_N_ATOM.  Increase this value"
    endif

    ! **************************** coordinates and box size************************************************!
    ! get solute coordinates
    n_atom = 0
    n_atom_drude=0

    call read_conf( ifile_conf, n_mole, n_atom, xyz, alist, box )
    ! if framework simulation,get framework coordinates
    if (framework_simulation .eq. 1) then
       ! Variable box is directly overwriten without furthur checking, by Kuang
       call read_fconf( ifile_fconf, n_mole, tot_n_mole, n_atom, xyz, temp_chg, flag_charge, alist, box )
       ! if flag_charge is set to 1, charges have been read in from this file, and overwrite
       ! those read in from the .pmt file. 
    else
       tot_n_mole = n_mole
    endif

    ! ********************* generate cross terms and fill in parameter arrays****************************!
    Select Case(energy_decomposition)
    Case("no")
       ! fill in total parameter arrays that are stored in global_variables
       call gen_param(alist,tot_n_mole,n_atom, chg, chg_drude,gen_cross_terms)
    Case("yes")
       call gen_param_decomp(alist,tot_n_mole,n_atom, chg, chg_drude , temp_lj_parameter , n_type_s)
    end Select


    !******************** fill in framework charges with charges from the .fconf file if those have been read in there
       Select Case(flag_charge)
          Case(1)
             write(*,*) ""
             write(*,*) "*******************************"
             write(*,*) "Found charges in .fconf file.  These charges will be"
             write(*,*) "used in simulation and corresponding charges in     "
             write(*,*) "*.pmt file will be ignored"
             write(*,*) "*******************************"
             write(*,*) ""
          
         do i_mole=n_mole+1, tot_n_mole
            chg(i_mole,:) = temp_chg(i_mole,:)
         enddo
       End Select


    !************************ generate masses from element type *****************************************!
    mass=0.
    ! generate masses for  atoms
    do i_mole=1,tot_n_mole
       call gen_mass( alist(i_mole,:), n_atom(i_mole), mass(i_mole,:) )
    enddo
    ! generate atom type masses
    call gen_mass( atype_name, n_atom_type, atype_mass )

    !  careful using massage if more than co2
    !  call massage( xyz, n_mole, n_atom, r_mole_com, mass ) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    call update_r_com( xyz, n_mole, n_atom, mass, r_com )
    ! if framework, update r_com for framework
    if (framework_simulation .eq. 1) then
       do i_mole = n_mole+1, tot_n_mole
          r_com(i_mole,:) = xyz(i_mole,1,:)
       end do
    endif

    mass_mol=0.
    do i_mole = 1, n_mole
       do i_atom = 1, n_atom( i_mole )
          mass_mol(i_mole)=mass_mol(i_mole)+mass(i_mole,i_atom)
       end do
    end do

    call gen_molecule_type_data( xyz, n_mole, n_atom, mass, r_com )

    do i_mole = 1, n_mole
       do i_atom = 1, n_atom( i_mole )
          call trim_end( alist( i_mole, i_atom ) )
       end do
    end do
    do i_atom = 1, n_atom_type
       call trim_end( atype_name( i_atom ) )
    end do
    if (framework_simulation .eq. 1) then
       do i_mole = n_mole+1, tot_n_mole
          call trim_end( alist( i_mole, 1 ) )
       end do
    endif

    ! check to make sure there is net zero charge in the unit cell, if we are using electrostatics
    if (electrostatic_type .ne. "none") then
       if (framework_simulation .eq. 1) then
          sum_chg=0d0
          do i_mole = n_mole+1, tot_n_mole
             sum_chg= sum_chg + chg(i_mole,1)
          end do
         write(*,*) "Total charge on the crystal unit cell"
         write(*,*) sum_chg
          if ( abs(sum_chg) > framework_chg_thresh ) then
             stop " unit cell is not neutral! "
          endif
       endif
    endif


    ! make sure to store framework charges, and framework atom numbers
    tot_chg = chg
    n_atom_drude = n_atom
    ! if simulation has drude oscillators, correct atom charges, and intialize mapping between
    ! drude oscillators and their corresponding atoms
    ! even if this is not a drude oscillator simulation, call this subroutine as it will
    ! fill drude_atom_map trivially for the atoms, which will be used in screening functions
       call initialize_drude_oscillators(tot_n_mole, n_mole, n_atom, n_atom_drude, tot_chg , chg_drude)

    ! if we are using a penetration force field, set that up
    Select Case(penetration_force_field)
    Case("yes")
       call initialize_penetration_ff_routines
    Case("no")
       ! we are not using a penetration force field, set all distance thresholds to zero
       ! so that this will never be called
       atype_penetration_ff(:,:,2)=0d0
    End Select

  end subroutine initialize_simulation



  !**************************************************!
  ! this subroutine initializes energy calculations
  !**************************************************!
  subroutine initialize_energy_routines(box,dfti_desc,dfti_desc_inv)
    use global_variables
    use pme_routines
    use MKL_DFTI
    real*8,dimension(:,:),intent(in)::box
    TYPE(DFTI_DESCRIPTOR), pointer,intent(out):: dfti_desc,dfti_desc_inv

    integer:: i,i_mole,i_atom,length(3),status
    real*8 :: a(3), b(3), c(3), ka(3), kb(3), kc(3),kk(3,3),vol
!   New variables by Kuang
    real*8 :: x, exp_x2


    !************************************* initialize ewald/pme ************************************************!


       ! set up fourier transform descriptors
       length=pme_grid
       status=DftiCreateDescriptor(dfti_desc, DFTI_DOUBLE, DFTI_COMPLEX, 3, length)
       status=DftiCommitDescriptor(dfti_desc)
       ! don't scale back transform because pick up factor of K^3 from convolution
       status=DftiCreateDescriptor(dfti_desc_inv, DFTI_DOUBLE, DFTI_COMPLEX, 3, length)
       status = DftiCommitDescriptor(dfti_desc_inv)

!      no longer need to compute CB array here
       a(:) = box(1,:);b(:) = box(2,:);c(:) = box(3,:)
       vol = volume( a, b, c )
       call crossproduct( a, b, kc ); kc = kc /vol 
       call crossproduct( b, c, ka ); ka = ka /vol
       call crossproduct( c, a, kb ); kb = kb /vol
       kk(1,:)=ka(:);kk(2,:)=kb(:);kk(3,:)=kc(:)
!       call CB_array(CB,alpha_sqrt,vol,pme_grid,kk,spline_order)

!      grid B_splines
       if (spline_order .eq. 6) then
          do i=1,spline_grid
             B6_spline(i)=B_spline(6./dble(spline_grid)*dble(i),6)
             B5_spline(i)=B_spline(5./dble(spline_grid)*dble(i),5)
          enddo
       else if (spline_order .eq. 4) then
          do i=1,spline_grid
             B4_spline(i)=B_spline(4./dble(spline_grid)*dble(i),4)
             B3_spline(i)=B_spline(3./dble(spline_grid)*dble(i),3)
          enddo
       else
          stop "requested spline order not implemented"
       endif

       Select case(grid_erfc)
       case("yes")
!         grid complementary error function
          do i=1,erfc_grid
             erfc_table(i)= erfc(erfc_max/dble(erfc_grid)*dble(i))
          enddo
!         initialize g functions used in dispersion PME, by Kuang
          do i = 1, gfun_grid
             x = gfun_max/dble(gfun_grid)*dble(i)
             exp_x2 = dexp(-x**2)
             g6_table(i) = (x**4/2d0+x**2+1d0)*exp_x2
             g8_table(i) = (x**6+3d0*x**4+6d0*x**2+6d0)/6d0*exp_x2
             g10_table(i) = (x**8+4d0*x**6+12d0*x**4+24d0*x**2+24d0)/24d0*exp_x2
             g12_table(i) = (x**10+5d0*x**8+20d0*x**6+60d0*x**4+120d0*x**2+120d0)/120d0*exp_x2
             g14_table(i) = (x**12+6d0*x**10+30d0*x**8+120d0*x**6+360d0*x**4+720*x**2+720d0)/720d0*exp_x2
          enddo
       end select



    !******************************** check for conflicting run parameters*************************************!
    ! let's not confuse anyone by turning energy decomposition on and drude simulation off
    Select Case(energy_decomposition)  
    Case("yes")
!       if (drude_simulation .ne. 1 ) then
!          stop "Do you really want energy decomposition without drude oscillators??? Gonna have to change the source code"
!       endif
       ! no energy decomposition without electrostatics
       select case(electrostatic_type)
       case("none")
          stop "can't run energy decomposition without electrostatics"
       end select
    end select
    ! cannot run drude simulation with no electrostatics!
    Select Case(electrostatic_type)
    Case("none")
       if (drude_simulation .eq.1 ) then
          stop "can't have drude_simulation without electrostatics, gonna have to change the laws of physics!"
       endif
    end select


  end subroutine initialize_energy_routines


  !**************************************************!
  ! this subroutine reads framework coordinates, it will also read
  ! charges if those are given after the coordinates.  This is detected
  ! by the number of arguments on the atom line
  !**************************************************!
  subroutine read_fconf( ifile_fconf, n_mole, tot_n_mole, n_atom, xyz, chg, flag_charge, alist, box )
    use mkl95_lapack
    use routines
    use global_variables
    implicit none
    character(*),intent(in) :: ifile_fconf
    character(*), dimension(:,:),intent(inout) :: alist
    integer,intent(in)::n_mole
    integer,intent(out)::tot_n_mole
    integer,dimension(:),intent(inout)::n_atom
    real*8,dimension(:,:,:),intent(inout)::xyz
    real*8,dimension(:,:),intent(out) :: chg
    integer  , intent(out)         :: flag_charge
    real*8,dimension(:,:),intent(out)::box
    real*8 :: tmp1, tmp2, tmp3
    character(100) :: line
    ! give args enough dimensions to accomodate all the arguments ( > 5) 
    character(20),dimension(10)  :: args
    character(MAX_ANAME) :: aname
    integer:: file_h=15, i_mole, n_frwk_atoms, ios, ipiv(3), nargs

    open(unit=file_h,file=ifile_fconf,status="old")
    read(file_h,*) n_frwk_atoms
    ! see how many arguments are given on the atom line
    read(file_h,'(A)') line
    call parse( line," ", args, nargs )
    backspace(file_h)

    Select Case( nargs )
    Case(4)
       flag_charge=0
    Case(5)
       flag_charge=1
    case default
       write(*,*) "unrecognized file format in file ", ifile_fconf
       stop
    End Select
    chg=0d0

    do i_mole=1,n_frwk_atoms
       Select Case( flag_charge )
       Case(0)
          ! no charges
          read(file_h,*) aname,xyz(n_mole+i_mole,1,1),xyz(n_mole+i_mole,1,2),xyz(n_mole+i_mole,1,3)
       Case(1)
          !charges
          read(file_h,*) aname,xyz(n_mole+i_mole,1,1),xyz(n_mole+i_mole,1,2),xyz(n_mole+i_mole,1,3), chg(n_mole+i_mole,1)
       end Select

       call trim_head( aname )
       alist( n_mole+i_mole,1) = aname
       n_atom( n_mole+i_mole) = 1
    enddo
    read(file_h,*) tmp1, tmp2, tmp3
    read( file_h, *, iostat=ios ) box(2,1), box(2,2), box(2,3)
    if ( ios < 0 ) then ! end of the file
       box = 0.0d0
       box(1,1) = tmp1
       box(2,2) = tmp2
       box(3,3) = tmp3
    else
       box(1,1) = tmp1
       box(1,2) = tmp2
       box(1,3) = tmp3
       read( file_h, * ) box(3,1), box(3,2), box(3,3)
    endif

    inv_box = box
    call DGETRF_MKL95(inv_box,ipiv)
    call DGETRI_MKL95(inv_box,ipiv)

    close(file_h)
    tot_n_mole = n_mole + n_frwk_atoms

  end subroutine read_fconf

  !**************************************************!
  ! this subroutine reads atom_type parameters from input file without energy decomposition
  !*************************************************
  subroutine read_param( ifile_pmt, gen_cross_terms)
    use global_variables
    character(*),intent(in)::ifile_pmt
    integer,dimension(:,:),intent(out) :: gen_cross_terms

    integer::file_h=15, i_mole,j_mole,n_type_s, n_type_f, n_cross, inputstatus,i_search,j_search,ind,ind1,ind2
    character(MAX_ANAME):: atom1, atom2
    real*8,dimension(6)::store
    character(25)::junk
    character(50)::line

    gen_cross_terms=0

    ! store has dimension 6, since atype_lj_parameter has dimension 6 ( changed to accomodate C12).  
    ! Zero these components that will not be used
    store=0d0
    open(unit=file_h,file=ifile_pmt,status="old")
    ! get solute atom types
    do
       Read(file_h,'(A)',Iostat=inputstatus) line
       If(inputstatus < 0) Exit
       ind=INDEX(line,'solute_species')
       IF(ind .NE. 0) Exit
    enddo
    Read(file_h,'(A)',Iostat=inputstatus) line
    read(file_h,*) n_type_s

    do i_mole=1, n_type_s
       read(file_h,*) atype_name(i_mole),atype_chg(i_mole),atype_lj_parameter(i_mole,i_mole,1), &
            &  atype_lj_parameter(i_mole,i_mole,2),atype_lj_parameter(i_mole,i_mole,3),atype_pol(i_mole)
    enddo

    ! get framework atom types
    do
       Read(file_h,'(A)',Iostat=inputstatus) line
       If(inputstatus < 0) Exit
       ind=INDEX(line,'framework_species')
       IF(ind .NE. 0) Exit
    enddo
    Read(file_h,'(A)',Iostat=inputstatus) line
    read(file_h,*) n_type_f
    do j_mole=1, n_type_f
       i_mole= n_type_s + j_mole
       read(file_h,*) atype_name(i_mole),atype_chg(i_mole),atype_lj_parameter(i_mole,i_mole,1), &
            &  atype_lj_parameter(i_mole,i_mole,2),atype_lj_parameter(i_mole,i_mole,3),atype_pol(i_mole)
    enddo

    n_atom_type =n_type_f + n_type_s

    ! fill in atype_solute array
    do i_mole=1,n_atom_type
       if ( i_mole .le. n_type_s ) then
          atype_solute(i_mole) = 1
       else
          atype_solute(i_mole) = 0
       endif
    enddo

    ind=0
    ! find out whether to read cross terms or generate them
    do
       Read(file_h,'(A)',Iostat=inputstatus) line
       If(inputstatus < 0) Exit
       ind=INDEX(line,'cross_terms')
       IF(ind .NE. 0) Exit
    enddo

    if ( ind .ne. 0 ) then
       read(file_h,*) n_cross
       if(n_cross > 0) then
          do i_mole=1, n_cross
             read(file_h,*) ind1,ind2, store(1),store(2),store(3)
             atype_lj_parameter(ind1,ind2,:)=store(:)
             atype_lj_parameter(ind2,ind1,:)=store(:)
             gen_cross_terms(ind1,ind2)=1
             gen_cross_terms(ind2,ind1)=1
          enddo
       endif
    endif

    ! finally make sure we don't have two of the same atom type defined

    do i_mole=1, n_atom_type - 1
       do j_mole = i_mole + 1 , n_atom_type
          if ( atype_name(i_mole) .eq. atype_name(j_mole) ) then
             write(*,*) "atomic parameters defined more than once for atom type", atype_name(i_mole)
             stop
          endif
       enddo
    enddo

    close(file_h)

  end subroutine read_param

  !******************************************************!
  ! this subroutine fills in parameter arrays in global variables
  ! to be used during the simulation
  !*****************************************************!
  subroutine gen_param(alist,tot_n_mole,n_atom, chg, chg_drude,gen_cross_terms)
  use global_variables
    character(*), dimension(:,:),intent(in) :: alist
    real*8, dimension(:,:), intent(out) :: chg, chg_drude
    integer, dimension(:,:), intent(in) :: gen_cross_terms
    integer, dimension(:), intent(in) :: n_atom
    integer, intent(in) :: tot_n_mole
    integer :: i_param, j_param, i_mole , i_atom, ind, i_type,flag

   ! in the disp_usegrid function used to calculate the non-coulomb part of the solute-framework interaction,
   ! implementation is only for either a SAPT-based force field with lj_bkghm=1, or a Lennard-jones force field with lj_bkghm=2
   ! also , warn about combination rules, since Lorentz-Berthelot combination rules are used here

   ! note that this subroutine won't be called for a non-SAPT force field, so it's okay to stop code if lj_bkghm=1
   Select Case(lj_bkghm)
	Case(1)
	write(*,*) ""
	write(*,*) "buckingham force field isn't implemented in PME-dispersion for a non SAPT-based force field"
	write(*,*) "please choose either a Lennard-Jones or SAPT-based force field"
	write(*,*) ""
	stop
   End Select

   Select Case(lj_comb_rule)
	Case("standard")
	! this is fine
	Case default
	write(*,*) ""
	write(*,*) "must use a Lorentz-Bethelot combination rule for Lennard-Jones force field"
	write(*,*) "please set lj_comb_rule='standard' for a Lennard-Jones potential"
	write(*,*) ""
	stop
   End Select 

    ! create cross terms for parameters
    ! if bkghm potential, use buckingham combination rules (all geometric)
    if(lj_bkghm .eq. 1) then
       do i_param=1, n_atom_type
          do j_param=1,n_atom_type
             if (gen_cross_terms(i_param,j_param) .eq. 0) then
               ! if these cross terms weren't given in input file
                atype_lj_parameter(i_param,j_param,1) = sqrt(atype_lj_parameter(i_param,i_param,1)* &
                                                      & atype_lj_parameter(j_param,j_param,1))    ! A coefficient
                atype_lj_parameter(i_param,j_param,2) = sqrt(atype_lj_parameter(i_param,i_param,2)* &
                                                      & atype_lj_parameter(j_param,j_param,2))    ! B coefficient
                atype_lj_parameter(i_param,j_param,3) = sqrt(atype_lj_parameter(i_param,i_param,3)* &
                                                      & atype_lj_parameter(j_param,j_param,3))    ! C coefficient
             else
                ! use given terms
                atype_lj_parameter(i_param,j_param,:) = atype_lj_parameter(i_param,j_param,:)
             endif
          enddo
       enddo
    else if(lj_bkghm .eq. 2) then
       ! if lennard jones potential, use geometric for epsilon, arithmetic combination rule for sigma
       do i_param=1, n_atom_type
          do j_param=1,n_atom_type
             if (gen_cross_terms(i_param,j_param) .eq. 0) then
                ! if these cross terms weren't given in input file
                atype_lj_parameter(i_param,j_param,1) = sqrt(atype_lj_parameter(i_param,i_param,1)* &
                                                      & atype_lj_parameter(j_param,j_param,1))    ! epsilon
                atype_lj_parameter(i_param,j_param,2) = (atype_lj_parameter(i_param,i_param,2)+ &
                                                      & atype_lj_parameter(j_param,j_param,2)) / dble(2)   ! sigma
             else
                ! use given terms
                atype_lj_parameter(i_param,j_param,:) = atype_lj_parameter(i_param,j_param,:)
             endif
          enddo
       enddo
    endif

    ! now create atom index array that links atoms to parameters
    do i_mole = 1, tot_n_mole
       do i_atom =1, n_atom(i_mole)
          flag=0
          do  i_type=1, n_atom_type
             ind=index(atype_name(i_type),alist(i_mole,i_atom))
             IF(ind .NE. 0) then
                flag=1
                i_param = i_type
                exit 
             endif
          enddo
          ! make sure all atom types have parameters
          if (flag .eq. 0 ) then
             write(*,*) "atom type    ", alist(i_mole,i_atom), "doesn't have force field parameters!"
             stop
          endif

          ! set index, chg, and polarizability
          atom_index(i_mole,i_atom) = i_param
          chg(i_mole,i_atom) = atype_chg(i_param)
          chg_drude(i_mole,i_atom) = -sqrt(atype_pol(i_param)*springcon)
       enddo
    enddo

  end subroutine gen_param

  !******************************************************!
  ! this subroutine generates molecule_type arrays to be used for particle insertion
  ! notice this only includes solute molecules, as this is all we need
  !
  ! this subroutine sets global variables:: n_molecule_type, molecule_type
  !*****************************************************!
  subroutine gen_molecule_type_data( xyz, n_mole, n_atom, mass, r_com )
  use global_variables
    integer, intent(in) :: n_mole
    integer, intent(in), dimension(:) :: n_atom
    real*8, intent(in), dimension(:,:) :: mass
    real*8, intent(in), dimension(:,:) :: r_com
    real*8, intent(in), dimension(:,:,:) :: xyz
    integer:: i_mole,j_mole,i_atom, old_type,count

    ! first molecule type is first molecule in input file
    n_molecule_type = 1
    do i_atom=1,n_atom(1)
       molecule_type(1,i_atom) = atom_index(1,i_atom)
    enddo
    molecule_type(1,n_atom(1)+1) = MAX_N_ATOM + 1
    ! update r_mole_com array in global variables
    call update_relative_atom_pos( xyz, 1, 1, n_atom, mass,r_com )

    end subroutine gen_molecule_type_data



  !*********************************************************
  ! This subroutine generates an array which gives atom positions relative to cofm
  ! this changes array r_mole_com in global variables
  !
  !*********************************************************
  subroutine update_relative_atom_pos( xyz, i_mole, i_index, n_atom, mass,r_com )
    use global_variables
    implicit none
    real*8, intent(in), dimension(:,:,:) :: xyz
    integer, intent(in) :: i_mole,i_index
    integer, intent(in), dimension(:) :: n_atom
    real*8, intent(in), dimension(:,:) :: mass
    real*8, intent(in), dimension(:,:) :: r_com

    integer :: i,j, i_atom, flag
    real*8,dimension(3)  :: delta_x,u_axis
    real*8::r_mag,r_magp
    real*8,parameter:: small = .1 , linear_criteria=.99


    ! if more than one atom, find distances to cofm
    if ( n_atom(i_mole) > 1 ) then

!!! check to see if the molecule is linear and create r_mole_com array
       do i_atom=1,n_atom(i_mole)
          delta_x(:)=xyz(i_mole,i_atom,:)-r_com(i_mole,:)
          if(dot_product(delta_x,delta_x) > small) exit
       enddo

       u_axis=delta_x/sqrt(dot_product(delta_x,delta_x))

       flag = 1
       do i_atom=1,n_atom(i_mole)
          delta_x(:)=xyz(i_mole,i_atom,:)-r_com(i_mole,:)

          r_mag = sqrt(dot_product(delta_x,delta_x))
          r_magp = dot_product(delta_x,u_axis)
          ! don't divide by zero
          if (r_mag > small ) then
             if ( ( abs(r_magp)/r_mag ) < linear_criteria ) then
                flag = 0
             endif
          endif
       enddo

       if ( flag == 1 ) then
          ! this is a linear molecule, generate moments of inertia, don't need quaternions,
          ! and don't need to store r_mole_com vector in any particular coordinate frame,
          ! so just use global axis
          molecule_shape(i_index) = 1

          do i_atom=1,n_atom(i_mole)
             delta_x(:)=xyz(i_mole,i_atom,:)-r_com(i_mole,:)
             r_mole_com(i_index,i_atom,:) = delta_x(:)
          enddo

       else

          molecule_shape(i_index) = 0

          ! store r_mole_com 
          do i_atom=1,n_atom(i_mole)
             delta_x(:)=xyz(i_mole,i_atom,:)-r_com(i_mole,:)
             r_mole_com(i_index,i_atom,:) = delta_x(:)
          enddo

       endif


       ! don't want to divide by zero here, if there is only one atom
    else
       molecule_shape(i_index) = 2
       r_mole_com(i_index,1,:) = 0d0
    endif

  end subroutine update_relative_atom_pos




end module initialization_routines
