!***********************************************************
! Here is the main subroutine that controls the construction of
! the free energy grid lattice model for a solute in a crystal
! framework
!**************************************************************

module lattice_sampling
  use routines
contains

  !*******************************************************
  ! the variables input to this subroutine are as follows:
  !
  !  box:          This array contains the three cartesian components (2nd index) of each of the three
  !                box vectors (1st index).
  !
  !
  !  tot_n_mole:   This is the total number of molecules in the system, including fixed, framework atoms.
  !                Note that every fixed, framework atom is treated as its own molecule
  !
  !  n_mole:       This is the number of solute molecules in the simulation.  Should be 1 here, since we
  !                are constructing a free energy grid
  !
  !  n_atom:       This array gives the number of atoms for each molecule, not including drude oscillators
  !
  !  n_atom_drude: This array gives the number of atoms plus drude oscillators for each molecule
  !
  !  xyz:          This array gives the cartesian coordinates of every atom in the system
  !
  !  mass:         This array contains the mass of every atom in the system
  !
  !  r_com:        This array contains the center of mass for each molecule in the system
  !
  !  tot_chg:      This array gives the total charges of every atom in the system. IMPORTANT::  These
  !                total charges include the contribution from any Drude oscillator attached to the atom. 
  !                therefore if an atom has a Drude oscillator attached to it with a charge -q(drude),
  !                this array will hold a charge of q(tot) = q(fixed) + q(drude) for that atom.
  !                Also, the Drude oscillator charges are stored in this array, after all the actual
  !                atoms.  So for example if molecule 5 has 3 atoms, then the charges of the 3 atoms
  !                are stored in chg(5,1),chg(5,2), and chg(5,3) and the charges of the corresponding 
  !                Drude oscillators are stored in chg(5,4),chg(5,5), and chg(5,6) respectively
  !
  !  atom_chg:     This array holds the STATIC charges (no Drude oscillator contribution) of all the atoms
  !
  !
  !  dfti_desc, dfti_desc_inv:
  !                These pointers are using in the MKL Fast Fourier Transform routines
  !
  !  log_file:     This is the file handle of the output file
  !
  !*******************************************************
  subroutine energy_grid(box,tot_n_mole,n_mole,n_atom,n_atom_drude,xyz,mass,r_com,tot_chg,atom_chg,dfti_desc,dfti_desc_inv,log_file)
    use global_variables
    use MKL_DFTI
    use eq_drude
    use pme_routines
    use linklist
    use boltzman_factor
    use space_group_info

    integer,intent(in)::log_file
    integer,intent(in)::n_mole,tot_n_mole
    integer,dimension(:),intent(in)::n_atom, n_atom_drude
    real*8,dimension(:,:,:),intent(in)::xyz
    real*8,dimension(:,:),intent(in)::mass,tot_chg,atom_chg
    real*8,dimension(:,:),intent(in)::box
    real*8,dimension(:,:),intent(in)::r_com
    TYPE(DFTI_DESCRIPTOR), pointer,intent(in):: dfti_desc,dfti_desc_inv
    integer, dimension(3)  :: cav_grid
    real*8, dimension(3,3) :: cubelet_vec
    real*8, dimension(:,:,:,:), allocatable :: cavity_bias_grid_center
    real*8, dimension(:,:,:), allocatable :: boltzmann_config, min_energy_config, stat_err
    integer, dimension(:,:,:), allocatable :: ns_size

    real*8, dimension(MAX_N_MOLE,MAX_N_ATOM,3) :: xyz_config,force_junk,xyz_drude_junk
    real*8,dimension(MAX_N_MOLE,3)     ::   new_r_com
    real*8, dimension(orientation_try):: Energy_orientation,dE_elec,dE_bh
    real*8, dimension(:),allocatable :: Energy_orientation_low, dE_elec_low,dE_bh_low


    ! other variables
    integer           :: i,j,k,ix,iy,iz,i_mole,j_mole,i_atom,j_atom,count,i_orient,is_overlap,iteration,extra_grid
    integer           :: i_start, i_finish, j_start, j_finish, k_start, k_finish
    real*8            :: E_ins_min,rand, E_junk,E_pol,vol,partition_sum,prob_conf
    real*8,dimension(3) :: disp,u_axis,unit,fine_grid_pos
    real*8,dimension(5) :: energy_components
    real*8,parameter :: small = 1D-4, max_rep=100.
    integer :: count_avg
    ! ns list
    integer, dimension(:,:,:,:), allocatable :: headlist
    integer, dimension(MAX_N_MOLE,MAX_N_ATOM,2) :: nslist
    ! eletrostatic PME
    real*8,dimension(:,:,:),allocatable :: recip_pot
    real*8 :: dE_nopol
    ! dispersion PME
    integer :: atom_id1, atom_id2, ntype_solute, split_do
    real*8 :: t1, t2
    real*8, dimension(:,:,:,:), allocatable :: lrdisp_pot
    real*8, dimension(3) :: randvec
    ! Vegas sampling
    real*8 :: sum_yy, miser_err, tmp, switch
    real*8, dimension(5) :: posit_p, posit_pb, posit_pe
    integer(fgsl_int) :: status
    integer(fgsl_size_t) :: dim, nsample
    type(fgsl_monte_function) :: gbz_fac
    type(c_ptr) :: ptr
    type(fgsl_monte_vegas_state) :: s_vegas
    type(fgsl_monte_miser_state) :: s_miser
    type(fgsl_rng_type) :: t_vegas
    type(fgsl_rng) :: r_vegas
    integer(c_size_t) :: n_vegas
    type(c_ptr) :: params
    integer(c_size_t):: n
    ! symmetry related
    integer :: iop, di, dj, dk, icurr,jcurr,kcurr, isym,jsym,ksym
    real*8, dimension(3) :: rb, re
    integer, dimension(2,2,2,3) :: new_ind


    !   allocate array
    allocate( cavity_bias_grid_center(cav_grid_a,cav_grid_b,cav_grid_c,3) )
    allocate( boltzmann_config(cav_grid_a,cav_grid_b,cav_grid_c) )
    allocate( min_energy_config(cav_grid_a,cav_grid_b,cav_grid_c) )
    allocate( stat_err(cav_grid_a,cav_grid_b,cav_grid_c) )
    allocate( ns_size(cav_grid_a,cav_grid_b,cav_grid_c) )
    allocate( headlist(na_nslist,nb_nslist,nc_nslist,2) )
    allocate( recip_pot(pme_grid,pme_grid,pme_grid) )


    !   construct neighbour list, added by Kuang
    call construct_linklist(box,na_nslist,nb_nslist,nc_nslist,tot_n_mole,n_mole,n_atom,xyz,headlist,nslist)

    !   the grid for long range eletrostatic interaction
    recip_pot = 0d0
    ! tot_chg and atom_chg should be the same without drude oscillators, but lets just be explicit
    ! also only framework charges should be used here, so it shouldn't really matter which charge array we use anyway
    Select Case( drude_simulation )
    Case(1)
       call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,tot_chg,dfti_desc,dfti_desc_inv,recip_pot,1)
    Case(0)
       call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,atom_chg,dfti_desc,dfti_desc_inv,recip_pot,1)
    End Select


    !   the grid for long range dispersion interaction
    !   Count number of atom types in solute molecule
    do ntype_solute = 1, n_atom_type
       if ( atype_solute(ntype_solute) == 0 ) then
          goto 999
       endif
    enddo
999 continue
    ntype_solute = ntype_solute - 1
    allocate( lrdisp_pot(ntype_solute,pme_grid,pme_grid,pme_grid) )
    !   construct long-range dispersion grid
    lrdisp_pot = 0d0
    do atom_id1 = 1, ntype_solute
       call construct_recip_disp_grid(atom_id1,n_mole,tot_n_mole,n_atom,lrdisp_pot(atom_id1,:,:,:),box,xyz,dfti_desc,dfti_desc_inv)
    enddo


    vol = volume ( box(1,:),box(2,:),box(3,:) )

    ! first collapse cavity grid size parameters from global_variables into a local array for use
    cav_grid(1) = cav_grid_a; cav_grid(2) = cav_grid_b; cav_grid(3) = cav_grid_c; 
    ! find cubelet vectors
    do i = 1,3
       cubelet_vec(i,:) = box(i,:) / dble (cav_grid(i))
    enddo
    ! cubelet  volume
    cubelet_vol = volume ( cubelet_vec(1,:),cubelet_vec(2,:),cubelet_vec(3,:) )


    ! assuming one molecule in the input file for energy grid
    if ( n_mole .ne. 1 ) then
       stop "must have one and only one solute molecule for energy grid"
    endif
    ! make sure threshold for drude oscillators is less than that for fine grid
    if ( Epol_thresh < fine_grid_thresh ) then
       stop "fine_grid_thresh must be less than Epol_thresh"
    endif
    Select case ( drude_simulation )
    Case(0)
       write(*,*) "fine grid will not be used as there are no drude oscillators"
    End Select
    ! pme will always be used for electrostatics
    write(*,*) "pme is always used for electrostatics in energy grid (unless 'none' option is used)"

    ! ************************ write output******************************************
    write( log_file, *) "*******run parameters***************************"
    write( log_file, *) "this is a free energy lattice grid "
    write( log_file, *) "lj_bkghm ", lj_bkghm
    write( log_file, *) "lj cutoff", lj_cutoff
    write( log_file, *) "electrostatics","   ", electrostatic_type
    write( log_file, *) "electrostatic (Ewald) cutoff", ewald_cutoff
    write( log_file, *) "grid error function","   ", grid_erfc
    write( log_file, *) "electrostatic screening", screen_type
    write( log_file, *) "drude simulation", drude_simulation
    write( log_file, *) "distance for assumed repulsive wall (A)", too_close
    write( log_file, *) "temperature", temp
    write( log_file, *) "volume", vol
    write( log_file, *) "lattice grid (x,y,z) " , cav_grid_a , cav_grid_b , cav_grid_c
    write( log_file, *) "number of orientations per lattice cell ",orientation_try
    write( log_file, *) "cubelet_volume" , cubelet_vol

    i_mole = 1

    ! if molecule is linear, get axis for vegas
    Select Case(molecule_shape(i_mole))
    Case(1)
       do i_atom=1,n_atom(i_mole)
          disp(:)=xyz(i_mole,i_atom,:)-r_com(i_mole,:)
          if(dot_product(disp,disp) > small) exit
       enddo
       u_axis=disp/sqrt(dot_product(disp,disp))
    case default
       ! cannot use vegas for non-linear molecules
       if ( vegas == "yes" ) then
          stop "cannot use vegas for non-linear molecules"
       endif
    end select


    ! now loop over all cubelets, calculating their centers, and calculating energies if  there is no overlap
    ! center vectors are in cartesian basis, not unit cell basis, but these are usually the same

    partition_sum=0d0
    ! added by Kuang
    if ( vegas == "yes" ) then
       ! set running environment for vegas algorithm
       tot_n_mole_bz = tot_n_mole
       n_mole_bz = n_mole
       n_atom_bz = n_atom
       cubelet_vec_bz = cubelet_vec
       cubelet_vol_bz = cubelet_vol
       u_axis_bz = u_axis
       box_bz = box
       xyz_bz = xyz
       xyz_config_bz = xyz
       new_r_com_bz = r_com
       r_com_bz = r_com
       atom_chg_bz = atom_chg
       mass_bz = mass
       allocate( headlist_bz(na_nslist,nb_nslist,nc_nslist,2) )
       headlist_bz = headlist
       nslist_bz = nslist
       allocate( recip_pot_bz(pme_grid,pme_grid,pme_grid) )
       recip_pot_bz = recip_pot
       allocate( lrdisp_pot_bz(ntype_solute,cav_grid_a,cav_grid_b,cav_grid_c) )
       lrdisp_pot_bz = lrdisp_pot
       nsample = orientation_try
       dim = 5
       posit_pb = (/ -0.5d0, -0.5d0, -0.5d0, 0d0, 0d0 /)
       posit_pe = (/  0.5d0,  0.5d0,  0.5d0,  pi, pi /)
       t_vegas = fgsl_rng_default
       r_vegas = fgsl_rng_alloc(t_vegas)
       gbz_fac = fgsl_monte_function_init(bz_fac, dim, ptr)
       s_vegas = fgsl_monte_vegas_alloc(dim)
       s_miser = fgsl_monte_miser_alloc(dim)
    else
       xyz_config = xyz
       new_r_com = r_com
    endif
    count=0

    boltzmann_config = 0.0d0

    if ( n_threads > 1 ) then
       split_do = cav_grid(1)/n_threads/2
    endif

    ! figure out whether we are computing the entire grid, or just part of it
    Select Case(partial_grid_computation)
    Case(0)
       ! do the full grid
       i_start=1;i_finish=cav_grid(1)
       j_start=1;j_finish=cav_grid(2)
       k_start=1;k_finish=cav_grid(3)
    Case(1)
       ! do the requested portion of the grid
       i_start=start_grid_a;i_finish=finish_grid_a
       j_start=start_grid_b;j_finish=finish_grid_b
       k_start=start_grid_c;k_finish=finish_grid_c

       ! warn against using symmetry with a partial grid computation
       if ( n_ops > 0 ) then
          write(*,*) "IMPORTANT NOTE:  Because this is a partial grid computation, symmetry "
          write(*,*) "will not be utilized "
       endif
    End Select


    !******************************* loop over lattice grid points

    do i = i_start, i_finish
       xyz_config = xyz
       new_r_com = r_com
       do j= j_start,j_finish
          do k= k_start,k_finish
             ! if this point is already sampled
             if ( abs(boltzmann_config(i,j,k) - 0.0d0) > 1d-9 ) then
                goto 777
             endif


             if ( vegas == "no" ) then
                cavity_bias_grid_center(i,j,k,:) = (dble(i)-.5d0)*cubelet_vec(1,:) + (dble(j)-.5d0)*cubelet_vec(2,:) &
                     & + (dble(k)-.5d0)*cubelet_vec(3,:)   
                boltzmann_config(i,j,k) = 0d0
                sum_yy = 0d0
                ! call cpu_time(t1)
                E_ins_min = 10000d0

                ! sample over orientation_try number of random configurations within each lattice site
                do i_orient=1, orientation_try

                   call random_number(randvec)
                   new_r_com(i_mole,:) = cavity_bias_grid_center(i,j,k,:) + (randvec(1)-0.5d0)*cubelet_vec(1,:) &
                        & + (randvec(2)-0.5d0)*cubelet_vec(2,:) + (randvec(3)-0.5d0)*cubelet_vec(3,:)

                   ! using the first molecule
                   do i_atom=1,n_atom(i_mole)
                      xyz_config(i_mole,i_atom,:) = xyz(i_mole,i_atom,:) - r_com(i_mole,:) + new_r_com(i_mole,:)
                   enddo

                   call random_rotation( xyz_config, new_r_com,n_atom, i_mole )



                   ! check distance from nearest atom, modified by Kuang
                   is_overlap=overlap_frame( xyz_config, tot_n_mole, n_mole, n_atom, i_mole, box, new_r_com, headlist, nslist )
                   if(is_overlap .eq. 1 ) then
                      ! if too close, set as max repulsive value
                      Energy_orientation(i_orient) = max_rep
                   else
                      Select Case(electrostatic_type)
                      Case("none")
                         dE_elec(i_orient) = 0d0
                      case default  ! modified by Kuang
                         if ( drude_simulation == 0 ) then
                            dE_elec(i_orient) = pme_usegrid( box, i_mole, tot_n_mole, n_mole, n_atom, new_r_com &
                                 & , xyz_config, atom_chg, recip_pot, headlist, nslist)
                         else
                            call scf_drude_usegrid( dE_elec(i_orient), dE_nopol, xyz_config, tot_chg, new_r_com, box, tot_n_mole &
                                 & ,n_mole, n_atom, n_atom_drude, iteration, xyz_drude_junk, dfti_desc, dfti_desc_inv, log_file &
                                 & ,recip_pot, headlist, nslist)
                         endif
                      End Select
                      dE_bh(i_orient) = disp_usegrid( box, i_mole, tot_n_mole, n_mole, n_atom, new_r_com, xyz_config, &
                           &               lrdisp_pot, headlist, nslist, mass)
                      ! calculate total energy
                      Energy_orientation(i_orient) = dE_elec(i_orient) + dE_bh(i_orient)
                      if (Energy_orientation(i_orient) < E_ins_min ) then
                         E_ins_min = Energy_orientation(i_orient)
                      endif
                   endif


                   boltzmann_config(i,j,k) = boltzmann_config(i,j,k) + dexp(-Energy_orientation(i_orient)*1000/8.314d0/temp)
                   sum_yy = sum_yy + dexp(-Energy_orientation(i_orient)*1000/8.314d0/temp)**2
                   ! check convergence
                   if ( mod(i_orient,10) == 0 ) then
                      stat_err(i,j,k) = dsqrt((sum_yy - boltzmann_config(i,j,k)**2/i_orient))/i_orient
                      tmp = boltzmann_config(i,j,k) / i_orient
                      switch = 1d0/(dexp(0.1*(tmp-BZ_CUTOFF))+1d0)
                      if ( abs(stat_err(i,j,k)/tmp) < switch*ABS_THRSH/tmp+(1d0-switch)*REL_THRSH ) then
                         goto 888
                      endif
                   endif
                enddo
                ! call cpu_time(t2)
                ! print *, t2 - t1

888             continue
                ns_size(i,j,k) = i_orient
                boltzmann_config(i,j,k) = boltzmann_config(i,j,k) / dble(ns_size(i,j,k))
                min_energy_config(i,j,k)= E_ins_min

                ! change Ewald self back to no polarization value
             else if ( vegas=="yes" ) then
                center_bz(:) = (dble(i)-.5d0)*cubelet_vec(1,:) + (dble(j)-.5d0)*cubelet_vec(2,:) + (dble(k)-.5d0)*cubelet_vec(3,:)
                boltzmann_config(i,j,k) = 0d0
                !              Vegas algorithm, by Kuang
                status = fgsl_monte_vegas_integrate( gbz_fac, posit_pb, posit_pe, dim, nsample/5, r_vegas &
                     &, s_vegas, boltzmann_config(i,j,k), stat_err(i,j,k))
                !              Miser algorithm, by Kuang
                !               status = fgsl_monte_miser_integrate( gbz_fac, posit_pb, posit_pe, dim, nsample, r_vegas &
                !                                                 &, s_miser, boltzmann_config(i,j,k), stat_err)
                !              scan
                !               posit_p(1) = 0.0d0
                !               posit_p(2) = 0.0d0
                !               posit_p(3) = 0.0d0
                !               do ix = 1, 50
                !                 posit_p(4) = pi*dble(ix-1)/50d0
                !                 do iy = 1, 50
                !                   posit_p(5) = pi*dble(iy-1)/50d0
                !                   write(*,'(2F12.6,E15.6)') posit_p(4), posit_p(5), bz_fac( posit_p, n, params )
                !                 enddo
                !                 write(*,*)
                !               enddo
                !              Naive Monte Carlo
                !               do i_orient = 1, orientation_try
                !                 call random_number( posit_p )
                !                 posit_p(1) = posit_p(1) - 0.5d0
                !                 posit_p(2) = posit_p(2) - 0.5d0
                !                 posit_p(3) = posit_p(3) - 0.5d0
                !                 posit_p(4) = posit_p(4) * pi
                !                 posit_p(5) = posit_p(5) * 2d0 * pi
                !                 boltzmann_config(i,j,k) = boltzmann_config(i,j,k) + 2d0*pi**2*bz_fac(posit_p,n_vegas,params)
                !               enddo
                !               boltzmann_config(i,j,k) = boltzmann_config(i,j,k) / orientation_try
             endif

             !symmetry tests
             do iop = 1, n_ops
                isym = 1000
                jsym = 1000
                ksym = 1000
                do di = 0, 1
                   rb(1) = dble(i+di-1) / cav_grid(1)
                   do dj = 0, 1
                      rb(2) = dble(j+dj-1) / cav_grid(2)
                      do dk = 0, 1
                         rb(3) = dble(k+dk-1) / cav_grid(3)
                         ! symmetry operation in direct coordinate
                         call sym_operate( rb, iop, re )
                         ! convert to grid coordinate
                         re(1) = re(1)*cav_grid(1)
                         re(2) = re(2)*cav_grid(2)
                         re(3) = re(3)*cav_grid(3)
                         ! re is not on the grid
                         if ( (re(1)-nint(re(1)))>1d-5 .or. (re(2)-nint(re(2)))>1d-5 .or. (re(3)-nint(re(3)))>1d-5 ) then
                            goto 777
                         endif
                         ! find the corresponding indices for new cubelet after symmetry operation
                         if ( nint(re(1))+1 < isym ) then
                            isym = nint(re(1))+1
                         endif
                         if ( nint(re(2))+1 < jsym ) then
                            jsym = nint(re(2))+1
                         endif
                         if ( nint(re(3))+1 < ksym ) then
                            ksym = nint(re(3))+1
                         endif
                      enddo
                   enddo
                enddo
                if ( abs(boltzmann_config(isym,jsym,ksym)) < 1d-9 ) then
                   boltzmann_config(isym,jsym,ksym) = boltzmann_config(i,j,k)
                   cavity_bias_grid_center(isym,jsym,ksym,:) = (dble(isym)-.5d0)*cubelet_vec(1,:) &
                        &+(dble(jsym)-.5d0)*cubelet_vec(2,:)+(dble(ksym)-.5d0)*cubelet_vec(3,:)
                   min_energy_config(isym,jsym,ksym) = min_energy_config(i,j,k)
                   stat_err(isym,jsym,ksym) = stat_err(i,j,k)
                   ns_size(isym,jsym,ksym) = ns_size(i,j,k)

                endif
             enddo

777          continue

          enddo
       enddo
    enddo

    ! output session
    do i = i_start,i_finish
       do j= j_start,j_finish
          do k= k_start,k_finish
             write( log_file, *) " cavity index " , i,j,k
             write( log_file, *) cavity_bias_grid_center(i,j,k,:)
             write( log_file, *) "min energy (over orientations) ", min_energy_config(i,j,k)
             write( log_file, *) "orientationally averaged boltzmann factor for this cubelet ", boltzmann_config(i,j,k)
!!$    if ( vegas == "yes" ) then
!!$       write( log_file, *) "Statistical Error: ", stat_err(i,j,k)/boltzmann_config(i,j,k) &
!!$            &   , fgsl_monte_vegas_chisq(s_vegas)
!!$    else
!!$       write( log_file, *) "Statistical Error: ", stat_err(i,j,k)/boltzmann_config(i,j,k)
!!$    endif
!!$    write( log_file, *) "Sampling size: ", ns_size(i,j,k)
          enddo
       enddo
    enddo

    if ( vegas == "yes" ) then
       call fgsl_monte_vegas_free(s_vegas)
       call fgsl_monte_miser_free(s_miser)
       call fgsl_rng_free(r_vegas)
    endif

  end subroutine energy_grid

end module lattice_sampling
