!**************************************************************************************
! This is the main program involved in constructing a lattice-model free energy grid
! for use in a lattice Monte Carlo simulation to calculate gas adsorption isotherms
! in porous materials
!
! This code was written by Kuang Yu, using code taken from the Schmidt Group atomistic Monte Carlo program
! This code has been modified by Jesse McDaniel
!
! the algorithms used in this code have been described in 
! Yu, K.; McDaniel, J. G.; Schmidt, J. R. An Efficient Multi-Scale Lattice Model Approach to Screening Nano-Porous Adsorbents, The Journal of Chemical Physics 2012, 137, 244102.
!
! The manual for this code, examples of how to run it, and parameters for our
! SAPT-based force fields can be downloaded from the group website
!
!              schmidt.chem.wisc.edu
!
!**************************************************************************************
program lattice_grid
  use routines
  use global_variables
  use MKL_DFTI
  use pme_routines
  use eq_drude
  use simulation_parameters
  use initialization_routines
  use lattice_sampling
  use space_group_info
  use Qeq

  implicit none
  character(MAX_FN) :: ifile_conf, ifile_fconf, ifile_pmt,ifile_simpmt, ofile_log,  symfn
  character(MAX_ANAME), dimension(MAX_N_MOLE,MAX_N_ATOM) :: alist
  integer :: n_mole, tot_n_mole, i_mole, i_atom
  integer, dimension(MAX_N_MOLE) :: n_atom, n_atom_drude
  real*8, dimension(MAX_N_MOLE) :: mass_mol
  real*8, dimension(MAX_N_MOLE,3) :: r_com
  real*8, dimension(MAX_N_MOLE,MAX_N_ATOM,3) :: xyz
  real*8, dimension(MAX_N_MOLE,MAX_N_ATOM) :: chg, chg_drude,tot_chg,mass
  TYPE(DFTI_DESCRIPTOR), POINTER :: dfti_desc,dfti_desc_inv
  real*8, dimension(3,3) :: box
  integer :: log_file=97, i, j, status,n_files

  n_files = IARGC ()

  if(n_files == N_input_f) then
     framework_simulation=1
  else
     stop "number of input files is not recognized"
  endif

     call getarg( 1, ifile_conf)
     call getarg( 2, ifile_fconf)
     call getarg( 3, ifile_pmt )
     call getarg( 4, ifile_simpmt )
     call getarg( 5, ofile_log )
     call getarg( 6, symfn )



  open( log_file,  file=ofile_log, status='new' )

  call read_simulation_parameters( ifile_simpmt )

  call initialize_simulation(box, alist, n_mole, tot_n_mole, n_atom, n_atom_drude,xyz, mass, mass_mol, &
       & r_com, chg, chg_drude, tot_chg, ifile_conf, ifile_fconf,ifile_pmt )

    ! read space group information
   call read_oplist( symfn, n_ops )

    ! check that symmetry is consistent with unit cell
   call check_symmetry_fconf_consistency( alist, n_mole, tot_n_mole, n_atom, xyz, box )

  call initialize_energy_routines(box,dfti_desc, dfti_desc_inv)

  ! if we are using Qeq charges, calculate those charges
  Select Case(Qeq_charges)
     Case("yes")
        write(*,*) ""
        write(*,*) "*******************IMPORTANT**********************"
        write(*,*) " the charges read in from the parameter file will not be used"
        write(*,*) " for the crystal framework, as the Qeq method will be used to "
        write(*,*) " determine the charges on the framework"
        write(*,*) "**************************************************"
        write(*,*) ""
        call Periodic_Qeq_solver(chg, alist, n_mole, tot_n_mole, n_atom, xyz, box, dfti_desc, dfti_desc_inv )
        ! fill in tot_charge array for framework
        do i_mole=n_mole+1, tot_n_mole
           tot_chg(i_mole,:) = chg(i_mole,:)
        enddo
  End Select





  !**********************************************************************
  ! this is the main subroutine that is in charge of constructing the free energy lattice model
  !**********************************************************************
  call energy_grid(box, tot_n_mole, n_mole, n_atom, n_atom_drude, xyz, mass, r_com, tot_chg,chg, dfti_desc, dfti_desc_inv, log_file)


  stop


end program lattice_grid
