module pme_routines
  use routines
  implicit none

!************************************************
! This module contains pme subroutines for energy and force
! for both electrostatics and dispersion interactions
! pme routines use discrete fourier transforms in MKL library
!
! For this lattice model code, these are the main energy routines
! used, since all energy evaluations are based on PME
!************************************************


contains

 !*************************************************************
  !  this subroutine adds intra molecular polarization (drude oscillator) electrostatic interactions
  !  it is meant to be used with Electrostatic cutoff routine, as no reciprocal space terms are subtracted
  !
  !*************************************************************
  subroutine intra_elec_energy(E_intra,xyz,chg,i_mole,i_atom,j_atom,n_atom)
    use global_variables
    implicit none
    real*8, intent(out):: E_intra
    integer, intent(in), dimension(:) :: n_atom
    real*8, intent(in), dimension(:,:) :: chg
    real*8, intent(in), dimension(:,:,:) :: xyz
    integer,intent(in)::i_mole,i_atom,j_atom

    real*8,dimension(3)::rij
    integer::i_drude,j_drude,sign_chg_i,sign_chg_j,flag_drudei,flag_drudej, flag_same_atom
    real*8::norm_dr,pol1,pol2
    real*8,parameter::small=1D-8


!!!!!!!!!if oscillators present, get rid of intra-molecular fixed chg-chg interactions, but allow screened induced-dipole interactions
    call drude_atom_indices(i_mole,n_atom,i_atom,j_atom,i_drude,j_drude,sign_chg_i,sign_chg_j,flag_drudei,flag_drudej, flag_same_atom)


    ! we only consider intra molecular interactions between drude oscillators, or the corresponding drude oscillator charge on an atom
    ! also, we do not consider intra molecular interactions between a drude oscillator and it's corresponding atom

    ! flag_drudei tells us whether i_atom is either a drude oscillator, or has a drude_oscillator attached to it
    ! flag_drudej tells us whether j_atom is either a drude oscillator, or has a drude_oscillator attached to it
    ! flag_same_atom tells us whether j_atom is the drude oscillator on i_atom



    if ( (flag_drudei == 1 ) .and. (flag_drudej == 1 ) .and. (flag_same_atom == 0 ) ) then
       rij(:) = xyz(i_mole,i_atom,:) - xyz(i_mole, j_atom,:)
       norm_dr = sqrt( dot_product( rij, rij ) )
!!!!!!!!!!!!!!! add screened atom-atom, get atom chg from corresponding oscillator
       pol1=chg(i_mole,i_drude)**2/springcon; pol2=chg(i_mole,j_drude)**2/springcon
       E_intra = dble(sign_chg_i*sign_chg_j)*chg(i_mole,i_drude) * chg(i_mole,j_drude) * thole_screen(pol1,pol2,norm_dr,thole)/ norm_dr
    else
       E_intra = 0D0
    endif


  end subroutine intra_elec_energy



  !*************************************************************
  !  this subroutine adds intra molecular polarization (drude oscillator) electrostatic forces
  !  it is meant to be used with Electrostatic cutoff routine, as no reciprocal space terms are subtracted
  !
  !*************************************************************
  subroutine intra_elec_force(f_ij,xyz,chg,i_mole,n_atom,i_atom,j_atom)
    use global_variables
    implicit none
    real*8, dimension(:), intent(out):: f_ij
    integer, intent(in), dimension(:) :: n_atom
    real*8, intent(in), dimension(:,:) :: chg
    real*8, intent(in), dimension(:,:,:) :: xyz
    integer,intent(in)::i_mole,i_atom,j_atom

    real*8,dimension(3)::rij
    integer::i_drude,j_drude,n_pairs,sign_chg_i,sign_chg_j,flag_drudei,flag_drudej, flag_same_atom
    real*8::norm_dr,pol1,pol2
    real*8,parameter::small=1D-8


!!!!!!!!!if oscillators present, get rid of intra-molecular fixed chg-chg interactions, but allow screened induced-dipole interactions
    call drude_atom_indices(i_mole,n_atom,i_atom,j_atom,i_drude,j_drude,sign_chg_i,sign_chg_j,flag_drudei,flag_drudej, flag_same_atom)

    ! we only consider intra molecular interactions between drude oscillators, or the corresponding drude oscillator charge on an atom
    ! also, we do not consider intra molecular interactions between a drude oscillator and it's corresponding atom


    ! flag_drudei tells us whether i_atom is either a drude oscillator, or has a drude_oscillator attached to it
    ! flag_drudej tells us whether j_atom is either a drude oscillator, or has a drude_oscillator attached to it
    ! flag_same_atom tells us whether j_atom is the drude oscillator on i_atom, or vice-versa  



    if ( (flag_drudei == 1 ) .and. (flag_drudej == 1 ) .and. (flag_same_atom == 0 ) ) then
       rij(:) = xyz(i_mole,i_atom,:) - xyz(i_mole, j_atom,:)
       norm_dr = sqrt( dot_product( rij, rij ) )
!!!!!!!!!!!!add screened atom-drude, get atom chg from corresponding oscillator
       pol1=chg(i_mole,i_drude)**2/springcon; pol2=chg(i_mole,j_drude)**2/springcon
       f_ij= dble(sign_chg_i*sign_chg_j)*chg(i_mole,i_drude) * chg(i_mole,j_drude) * (rij * thole_screen(pol1,pol2,norm_dr,thole)/ norm_dr**3 - d_thole_screen(pol1,pol2,rij,thole) / norm_dr)
    else
       f_ij =0D0
    endif



  end subroutine intra_elec_force





!*****************************************************************
! This subroutine interpolates charges onto Q grid to be used in pme reciprocal space
! routines
!*****************************************************************
  subroutine grid_Q(Q,chg,xyz,n_atom,n_mole,K,n)
    use global_variables
    use omp_lib
    integer, intent(in) :: n_mole
    integer, intent(in), dimension(:) :: n_atom
    real*8, intent(in), dimension(:,:) :: chg
    real*8, intent(in), dimension(:,:,:) :: xyz
    integer,intent(in)::K,n
    real*8,dimension(:,:,:),intent(out)::Q
    integer::i,j,k1,k2,k3,n1,n2,n3,nn1,nn2,nn3,nearpt(3)
    real*8::sum,chg_i
    real*8,dimension(3)::u,arg
    integer :: split_do

    Q = 0d0
    ! decide how to split the parallel section
    if (n_threads .eq. 1 ) then
       split_do = 1
    else
       split_do = n_mole/n_threads+1
    endif
! parameters spline_grid, spline_order undeclared, but ok
! call OMP_SET_NUM_THREADS(n_threads)
!!$OMP PARALLEL DEFAULT(PRIVATE) SHARED(n_threads,n_mole,n_atom,xyz,chg,n,B6_spline,B4_spline,K,split_do) REDUCTION(+:Q)
!!$OMP DO SCHEDULE(dynamic, split_do)
    do i=1,n_mole
       do j=1,n_atom(i)
          chg_i=chg(i,j)
          u=xyz(i,j,:)
          nearpt=floor(u)
          do k1=0,n
             n1=nearpt(1)-k1
             arg(1)=u(1)-dble(n1)
             do k2=0,n
                n2=nearpt(2)-k2
                arg(2)=u(2)-dble(n2)
                do k3=0,n
                   n3=nearpt(3)-k3
                   arg(3)=u(3)-dble(n3)
                   if( arg(1)<dble(n) .and. 0.<arg(1) .and. arg(2)<dble(n) .and. 0.<arg(2) .and. arg(3)<dble(n) .and. &
                      & 0.<arg(3) ) then 
                      if(spline_order .eq.6) then
                        sum = chg_i*B6_spline(ceiling(arg(1)/6.D0*dble(spline_grid))) &
                            & * B6_spline(ceiling(arg(2)/6.D0*dble(spline_grid)))*B6_spline(ceiling(arg(3)/6.D0*dble(spline_grid)))
                      else
                        sum = chg_i*B4_spline(ceiling(arg(1)/4.D0*dble(spline_grid))) &
                            & * B4_spline(ceiling(arg(2)/4.D0*dble(spline_grid)))*B4_spline(ceiling(arg(3)/4.D0*dble(spline_grid)))
                      endif
                   else
                      sum=0.D0
                   endif
                   if(n1<0) then
                      nn1=n1+K
                   else
                      nn1=n1
                   endif
                   if(n2<0) then
                      nn2=n2+K
                   else
                      nn2=n2
                   endif
                   if(n3<0) then
                      nn3=n3+K
                   else
                      nn3=n3
                   endif

                   Q(nn1+1,nn2+1,nn3+1)=Q(nn1+1,nn2+1,nn3+1)+sum
                enddo
             enddo
          enddo
       enddo
    enddo
!!$OMP END DO NOWAIT
!!$OMP END PARALLEL

  end subroutine grid_Q



!!!!!!!!!!!!!!!!!!!!!!!!
    ! this subroutine computes derivatives of grid Q with respect 
    ! to changes in position of i_atom in i_mole
    ! output is the force, which is computed by taking the product of
    ! the derivative Q array with FQ, which is the convoluted array
    !
    ! currently, transformation from scaled coordinate derivative to 
    ! box coordinate derivative is for rectangular boxes only
    !
    ! notice that this is being called in a parallel section of code
!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine derivative_grid_Q(force,FQ,chg,xyz,i_atom,i_mole,K,box,n)
    use global_variables
    integer,intent(in)::K,n
    real*8,dimension(3),intent(out)::force
    real*8,dimension(:,:,:),intent(in)::FQ
    integer, intent(in) :: i_mole,i_atom
    real*8, intent(in), dimension(:,:) :: chg
    real*8,intent(in),dimension(3,3)::box
    real*8, intent(in), dimension(:,:,:) :: xyz
    integer::k1,k2,k3,n1,n2,n3,nn1,nn2,nn3,nearpt(3)
    integer::g1(3),g2(3)
    real*8::fac(3),chg_i
    real*8,dimension(3)::u,arg1,arg2

    force=0.D0
    chg_i=chg(i_mole,i_atom)
    u=xyz(i_mole,i_atom,:)
    nearpt=floor(u)

    do k1=0,n
       n1=nearpt(1)-k1
       arg1(1)=u(1)-dble(n1)
       arg2(1)=u(1)-1.D0-dble(n1)
       do k2=0,n
          n2=nearpt(2)-k2
          arg1(2)=u(2)-dble(n2)
          arg2(2)=u(2)-1.D0-dble(n2)
          do k3=0,n
             n3=nearpt(3)-k3
             arg1(3)=u(3)-dble(n3)
             arg2(3)=u(3)-1.-dble(n3)
             ! get bspline values from grid, therefore must check whether to see if function is being evaluated in non-zero domain
             ! otherwise, can't call array
             fac=0.D0
             ! force in x direction
             g2=ceiling(arg2/real(n-1)*real(spline_grid))
             g1=ceiling(arg1/real(n)*real(spline_grid));g1(1)=ceiling(arg1(1)/real(n-1)*real(spline_grid))
             if( arg1(1)<real(n-1) .and. 0.<arg1(1) .and. arg1(2)<real(n) .and. 0.<arg1(2) .and. arg1(3)<real(n) .and. &
                & 0.<arg1(3) ) then 
                if(spline_order.eq.6) then
                   fac(1)=chg_i*(B5_spline(g1(1))*B6_spline(g1(2))*B6_spline(g1(3)))
                else
                   fac(1)=chg_i*(B3_spline(g1(1))*B4_spline(g1(2))*B4_spline(g1(3)))  
                endif
             endif
             if( arg2(1)<real(n-1) .and. 0.<arg2(1) .and. arg1(2)<real(n) .and. 0.<arg1(2) .and. arg1(3)<real(n) &
                & .and. 0.<arg1(3) ) then 
                if(spline_order.eq.6) then
                   fac(1)=fac(1)+chg_i*(-B5_spline(g2(1))*B6_spline(g1(2))*B6_spline(g1(3)))
                else
                   fac(1)=fac(1)+chg_i*(-B3_spline(g2(1))*B4_spline(g1(2))*B4_spline(g1(3)))  
                endif
             endif
             ! force in y direction
             g1=ceiling(arg1/real(n)*real(spline_grid));g1(2)=ceiling(arg1(2)/real(n-1)*real(spline_grid))
             if( arg1(2)<real(n-1) .and. 0.<arg1(2) .and. arg1(1)<real(n) .and. 0.<arg1(1) .and. arg1(3)<real(n) &
                & .and. 0.<arg1(3) ) then 
                if(spline_order.eq.6) then
                   fac(2)=chg_i*(B5_spline(g1(2))*B6_spline(g1(1))*B6_spline(g1(3)))
                else
                   fac(2)=chg_i*(B3_spline(g1(2))*B4_spline(g1(1))*B4_spline(g1(3)))  
                endif
             endif
             if( arg2(2)<real(n-1) .and. 0.<arg2(2) .and. arg1(1)<real(n) .and. 0.<arg1(1) .and. arg1(3)<real(n) &
                & .and. 0.<arg1(3) ) then 
                if(spline_order.eq.6) then
                   fac(2)=fac(2)+chg_i*(-B5_spline(g2(2))*B6_spline(g1(1))*B6_spline(g1(3)))
                else
                   fac(2)=fac(2)+chg_i*(-B3_spline(g2(2))*B4_spline(g1(1))*B4_spline(g1(3)))  
                endif
             endif
             ! force in z direction
             g1=ceiling(arg1/real(n)*real(spline_grid));g1(3)=ceiling(arg1(3)/real(n-1)*real(spline_grid))
             if( arg1(3)<real(n-1) .and. 0.<arg1(3) .and. arg1(1)<real(n) .and. 0.<arg1(1) .and. arg1(2)<real(n) &
                & .and. 0.<arg1(2) ) then 
                if(spline_order.eq.6) then
                   fac(3)=chg_i*(B5_spline(g1(3))*B6_spline(g1(1))*B6_spline(g1(2)))
                else
                   fac(3)=chg_i*(B3_spline(g1(3))*B4_spline(g1(1))*B4_spline(g1(2)))  
                endif
             endif
             if( arg2(3)<real(n-1) .and. 0.<arg2(3) .and. arg1(1)<real(n) .and. 0.<arg1(1) .and. arg1(2)<real(n) &
                & .and. 0.<arg1(2) ) then 
                if(spline_order.eq.6) then
                   fac(3)=fac(3)+chg_i*(-B5_spline(g2(3))*B6_spline(g1(1))*B6_spline(g1(2)))
                else
                   fac(3)=fac(3)+chg_i*(-B3_spline(g2(3))*B4_spline(g1(1))*B4_spline(g1(2)))  
                endif
             endif
             if(n1<0) then
                nn1=n1+K
             else
                nn1=n1
             endif
             if(n2<0) then
                nn2=n2+K
             else
                nn2=n2
             endif
             if(n3<0) then
                nn3=n3+K
             else
                nn3=n3
             endif
             force=force+fac*FQ(nn1+1,nn2+1,nn3+1)
          enddo
       enddo
    enddo

!!! change to box coordinates, for rectangular box
    force(1)=-force(1)*dble(K)/box(1,1)
    force(2)=-force(2)*dble(K)/box(2,2)
    force(3)=-force(3)*dble(K)/box(3,3)

  end subroutine derivative_grid_Q



!***********************************************
! this function calculates B_splines which are used in pme as interpolating
! functions.  B_splines are calculated recursively, and therefore it's a good idea
! to grid them
!************************************************
  real*8 function B_spline(u,n)
    real*8,intent(in)::u
    integer,intent(in)::n
    integer::i,j
    real,dimension(n-1,n-1)::mn
    real*8::ui

    ! define m2 for n-1 values
    do i=1,n-1
       ui=u-dble(i-1)
       if((ui<0.).or.(ui>2.)) then
          mn(1,i)=0.D0
       else
          mn(1,i)=1.D0-abs(ui-1.D0)
       endif
    enddo

   ! define mj recursively for n-1-(j-1) values
    do j=2,n-1
       do i=1,n-j
          ui=u-dble(i-1)
          mn(j,i)=(ui/dble(j))*mn(j-1,i)+((dble(j+1)-ui)/dble(j))*mn(j-1,i+1)
       enddo
    enddo

    B_spline=mn(n-1,1)

  end function B_spline


!***********************************************************
! This is a routine for reciprocal space pme calculation
!***********************************************************
  subroutine CB_array(CB,alpha_sqrt,vol,K,kk,n)
    real*8,dimension(:,:,:),intent(out)::CB
    real*8,intent(in)::alpha_sqrt,vol
    real*8,dimension(3,3),intent(in)::kk
    integer,intent(in)::K,n
    real*8, parameter :: pi=3.14159265
    real*8,dimension(3)::mm
    real*8::mag
    integer::i,j,l,m1,m2,m3
    do i=0,K-1
       if(i>K/2) then
          m1=i-K
       else
          m1=i
       endif
       do j=0,K-1
          if(j>K/2) then
             m2=j-K
          else
             m2=j
          endif
          do l=0,K-1
             if(l>K/2) then
                m3=l-K
             else
                m3=l
             endif
             mm(:)=m1*kk(1,:)+m2*kk(2,:)+m3*kk(3,:)
             mag=dot_product(mm,mm)
             CB(i+1,j+1,l+1)=1./(vol*pi)*exp(-pi**2*mag/alpha_sqrt**2)/mag*bm_sq(i,n,K)*bm_sq(j,n,K)*bm_sq(l,n,K)
          enddo
       enddo
    enddo

    CB(1,1,1)=0.D0

  end subroutine CB_array


!******************************************************
! this is needed in reciprocal space pme calculation
!******************************************************
  function bm_sq(m,n,K)
    use global_variables
    real*8::bm_sq
    integer,intent(in)::m,n,K
    integer::i
    complex*16::bm,sum
    real*8::tmp

    sum=0.D0
    do i=0,n-2
       tmp=2.D0*pi*dble(m*i)/dble(K)
       sum=sum+B_spline(dble(i+1),n)*cmplx(cos(tmp),sin(tmp))
    enddo

    bm=1.D0/sum
    bm_sq=dble(bm)**2+aimag(bm)**2

  end function bm_sq

!****************************************************************************
! This subroutine construct the reciprocal PME potenatial grids
! Major output:
! real*8, dimension(:,:,:) :: recip_pot ! the reciprocal PME potential grid
! It can be generalized to the dispersion situation, where, U ~ 1/r^p
! In that case, chg should be Cp(ij), the dispersion coefficients
!****************************************************************************
  subroutine recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,chg,dfti_desc,dfti_desc_inv,recip_pot,p)
  use global_variables
  use routines
  use MKL_DFTI
  implicit none 
    real*8, dimension(:,:), intent(in) :: box
    integer, intent(in) :: tot_n_mole, n_mole, p
    integer, dimension(:), intent(in) :: n_atom
    real*8, dimension(:,:,:), intent(in) :: xyz
    real*8, dimension(:,:), intent(in) :: chg
    TYPE(DFTI_DESCRIPTOR), pointer,intent(in):: dfti_desc,dfti_desc_inv
    real*8, dimension(:,:,:), intent(out) :: recip_pot
!   local variables
    real*8, dimension(:,:,:), allocatable :: Q
    complex*16,dimension(:,:,:), allocatable :: FQ, CBp
    integer :: n,K, i_mole, i_atom, idim, status, m1,m2,m3
    real*8 :: vol,a(3),b(3),c(3),ka(3),kb(3),kc(3),kk(3,3),mag,mm(3)
    real*8, dimension(:,:,:), allocatable :: xyz_scale
    real*8, dimension(:), allocatable ::q_1r
    complex*16,dimension(:), allocatable ::q_1d
   

!   allocate local variables
    allocate( Q(pme_grid,pme_grid,pme_grid) )
    allocate( FQ(pme_grid,pme_grid,pme_grid), CBp(pme_grid,pme_grid,pme_grid) )
    allocate( q_1r(pme_grid**3) )
    allocate( q_1d(pme_grid**3) ) 
    n=spline_order
!   This pme calculation is controled by pme_grid
    K=pme_grid
    
    a(:) = box(1,:)
    b(:) = box(2,:)
    c(:) = box(3,:)

    ! calculate the volume and the reciprocal vectors (notice no 2pi)
    vol = volume( a, b, c )
    kk(1,:)=inv_box(:,1)
    kk(2,:)=inv_box(:,2)
    kk(3,:)=inv_box(:,3)


    ! create scaled coordinates
    allocate(xyz_scale(tot_n_mole-n_mole,maxval(n_atom(n_mole+1:)),3))

!   only loops over frame atoms
    do i_mole = n_mole+1, tot_n_mole
      do i_atom = 1, n_atom(i_mole)
        do idim = 1, 3
          xyz_scale(i_mole-n_mole,i_atom,idim) = dble(K) * dot_product(kk(idim,:),xyz(i_mole,i_atom,:))
          ! if atoms are out of grid, shift them back in
          if ( xyz_scale(i_mole-n_mole,i_atom,idim) < 0. ) then
            xyz_scale(i_mole-n_mole,i_atom,idim) = xyz_scale(i_mole-n_mole,i_atom,idim) + dble(K)
        !  else if( xyz_scale(i_mole-n_mole,i_atom,idim) > dble(K) ) then   ! changed by Jesse, 6/10/2014 see below line.  In rare cases, this could lead to out of bounds array calls when constructing Q
          else if( xyz_scale(i_mole-n_mole,i_atom,idim) >= dble(K) ) then
            xyz_scale(i_mole-n_mole,i_atom,idim) = xyz_scale(i_mole-n_mole,i_atom,idim) - dble(K)
          endif
        enddo
      enddo
    enddo

!   Fourier transform of Q grid
    call grid_Q(Q,chg(n_mole+1:,:),xyz_scale,n_atom(n_mole+1:),tot_n_mole-n_mole,K,n)


    q_1r = RESHAPE(Q, (/K**3/) )
    q_1d = cmplx(q_1r,0.,16)
    status=DftiComputeForward(dfti_desc, q_1d)
    FQ = RESHAPE(q_1d, (/K,K,K/) )
!   CB' matrixm, debug
    if ( p == 1 ) then
      call CB_prime(CBp,alpha_sqrt,vol,K,kk,n,p)
    else
      call CB_prime(CBp,lj_asqrt,vol,K,kk,n,p)
    endif

    FQ = FQ * CBp

!   inverse transform
    q_1d = RESHAPE(FQ, (/K**3/))
    status = DftiComputeBackward(dfti_desc_inv, q_1d)
    FQ=RESHAPE(q_1d, (/K,K,K/) )

    recip_pot = recip_pot + dble(FQ)

    deallocate( Q, FQ, CBp, q_1r , q_1d )

    return
  end subroutine recip_pme_grid




!***************************************************************
! Modified from bm_sq, this is used to calculate potential grid
!***************************************************************
  function bm(m,n,K)
      use global_variables
      integer,intent(in)::m,n,K
      integer::i
      complex*16::bm,sum
      real*8::tmp
      sum=0.D0
      do i=0,n-2
         tmp=2.D0*pi*dble(m*i)/dble(K)
         sum=sum+cmplx(B_spline(dble(i+1),n),0.0d0,16)*cmplx(cos(tmp),sin(tmp),16)
!!$     sum=sum+B6_spline(dble(i+1)/6.*dble(spline_grid))*cmplx(cos(tmp),sin(tmp))
      enddo
      tmp = 2.0d0*pi*dble((n-1)*m)/dble(K)
      bm = cmplx(cos(tmp),sin(tmp),16)/sum
  end function bm

!**********************************************************************
! Here, the complex CB' matrix is:
! CB' = 1/pi/V*exp(-pi**2*m**2/alpha)/m**2 * b1(m1)*b2(m2)*b3(m3)
! compared to normal CB array, the B_spline functions are not squared
!**********************************************************************
  subroutine CB_prime(CBp,a_sqrt,vol,K,kk,n,p)
  use omp_lib
  use global_variables
    complex*16,dimension(:,:,:),intent(out)::CBp
    real*8,intent(in)::a_sqrt,vol
    real*8,dimension(3,3),intent(in)::kk
    integer,intent(in)::K,n,p
    real*8,dimension(3)::mm
    real*8::mag, x, f12
    integer::i,j,l,m1,m2,m3,split_do

 ! decide how to split the parallel section
    if (n_threads .eq. 1 ) then
       split_do = 1
    else
       split_do = K/n_threads+1
    endif

    CBp = 0.d0
! call OMP_SET_NUM_THREADS(n_threads)
!!$OMP PARALLEL DEFAULT(PRIVATE) SHARED(vol,K,kk,n,p) REDUCTION(+:CBp)
!!$OMP DO SCHEDULE(DYNAMIC,split_do)
    do i=0,K-1
       if(i>K/2) then
          m1=i-K
       else
          m1=i
       endif
       do j=0,K-1
          if(j>K/2) then
             m2=j-K
          else
             m2=j
          endif
          do l=0,K-1
             if(l>K/2) then
                m3=l-K
             else
                m3=l
             endif
             mm(:)=m1*kk(1,:)+m2*kk(2,:)+m3*kk(3,:)
             mag=dot_product(mm,mm)
             x = pi*dsqrt(mag)/a_sqrt
!            Note, due to the different definition of FFT, we need the complex conjugate here.
             select case(p)
             case(1)
               CBp(i+1,j+1,l+1) = pi_sqrt*pi/vol/a_sqrt**2*exp(-x**2)/pi_sqrt/x**2*conjg( bm(i,n,K)*bm(j,n,K)*bm(l,n,K) )
!               CBp(i+1,j+1,l+1) = 1.d0/(vol*pi)*exp(-pi**2*mag/a_sqrt**2)/mag*conjg( bm(i,n,K)*bm(j,n,K)*bm(l,n,K) )
             case(6)
               CBp(i+1,j+1,l+1) = pi_sqrt*pi/vol*a_sqrt**3* &
                                & ( (1d0-2d0*x**2)*exp(-x**2) + 2d0*pi_sqrt*x**3*erfc_g(x) ) / 3d0 &
                                & *conjg( bm(i,n,K)*bm(j,n,K)*bm(l,n,K) )
             case(8)
               CBp(i+1,j+1,l+1) = pi_sqrt*pi/vol*a_sqrt**5* &
                                & ( (3d0-2d0*x**2+4d0*x**4)*exp(-x**2) - 4d0*pi_sqrt*x**5*erfc_g(x) ) / 45d0 &
                                & *conjg( bm(i,n,K)*bm(j,n,K)*bm(l,n,K) )
             case(10)
               CBp(i+1,j+1,l+1) = pi_sqrt*pi/vol*a_sqrt**7* &
                                & ( (15d0-6d0*x**2+4d0*x**4-8d0*x**6)*exp(-x**2) + 8d0*pi_sqrt*x**7*erfc_g(x) ) / 1260d0 &
                                & *conjg( bm(i,n,K)*bm(j,n,K)*bm(l,n,K) )
             case(12)
               CBp(i+1,j+1,l+1) = pi_sqrt*pi/vol*a_sqrt**9* &
                                & ( (105d0-30d0*x**2+12d0*x**4-8d0*x**6+16d0*x**8)*exp(-x**2) - 16d0*pi_sqrt*x**9*erfc_g(x) ) &
                                & / 56700d0 * conjg( bm(i,n,K)*bm(j,n,K)*bm(l,n,K) )
             case(14)
               ! f12
               f12 = ( (105d0-30d0*x**2+12d0*x**4-8d0*x**6+16d0*x**8)*exp(-x**2) - 16d0*pi_sqrt*x**9*erfc_g(x) ) / 56700d0
               CBp(i+1,j+1,l+1) = pi_sqrt*pi/vol*a_sqrt**11* &
                                & ( 1d0/3960d0*exp(-x**2) - x**2/33d0*f12 ) &
                                & * conjg( bm(i,n,K)*bm(j,n,K)*bm(l,n,K) )
             case default
               write(*,*) 'Error: Interactions other than r^-1, -6, -8, -10, -12, -14 are not supported!'
               stop
             end select
!             CBp(i+1,j+1,l+1) = conjg( bm(i,n,K)*bm(j,n,K)*bm(l,n,K) )
          enddo
       enddo
    enddo
!!$OMP END DO NOWAIT
!!$OMP END PARALLEL

    if ( p == 1 ) then
      CBp(1,1,1) = cmplx(0.0d0,0.0d0,16)
    endif

  end subroutine CB_prime

!**********************************************************************************************************
! This subroutine calculates pme energy of a single molecule (specified by i_mole) in the framework field
! Modified 2012-11-08, general box shape supported
!**********************************************************************************************************
  function pme_usegrid( box, i_mole, tot_n_mole, n_mole, n_atom, r_com, xyz_config, atom_chg, recip_pot, headlist, nslist)
  use global_variables
  use routines
  implicit none
    integer, intent(in) :: tot_n_mole, n_mole, i_mole
    integer, dimension(:), intent(in) :: n_atom
    integer, dimension(:,:,:,:), intent(in) :: headlist
    integer, dimension(MAX_N_MOLE,MAX_N_ATOM,2), intent(in) :: nslist
    real*8, dimension(:,:,:), intent(in) :: xyz_config, recip_pot
    real*8, dimension(:,:), intent(in) :: atom_chg, box, r_com
    real*8 :: pme_usegrid
!   local variables
    integer :: i_atom, j_mole, j_atom, tmp_mole, tmp_atom
    real*8 :: ra,rb,rc, dra,drb,drc,dr, ewald_cutoff_use, pme_real, E_intra, rka, rkb, rkc
    integer :: ia,ib,ic, ia0,ib0,ic0, igrida,igridb,igridc, dia,dib,dic, split_do
    real*8, dimension(-lgrg_order+1:lgrg_order,-lgrg_order+1:lgrg_order,-lgrg_order+1:lgrg_order) :: weight
!    real*8, dimension(2,2,2) :: weight
    real*8, dimension(3) :: r_com_i, r_com_j, shift, dr_com,rij, rtmp

!   effective cutoff for real space calculations
    ewald_cutoff_use = min(box(1,1)/2.,box(2,2)/2.,box(3,3)/2.,ewald_cutoff)
!   real space cutoff for cell link list
    rka = dsqrt(dot_product(inv_box(:,1),inv_box(:,1)))
    rkb = dsqrt(dot_product(inv_box(:,2),inv_box(:,2)))
    rkc = dsqrt(dot_product(inv_box(:,3),inv_box(:,3)))
    dia = floor(ewald_cutoff*rka*dble(na_nslist)) + 1
    dib = floor(ewald_cutoff*rkb*dble(nb_nslist)) + 1
    dic = floor(ewald_cutoff*rkc*dble(nc_nslist)) + 1

    if (n_threads .eq. 1 ) then
       split_do = 1
    else
       split_do = (dia*2+1)/n_threads
    endif

    pme_usegrid = 0.0d0
    do i_atom = 1, n_atom(i_mole)

      rtmp(:) = matmul(xyz_config(i_mole,i_atom,:),inv_box(:,:))
      ra = rtmp(1) * pme_grid
      rb = rtmp(2) * pme_grid
      rc = rtmp(3) * pme_grid
!     shift into the box
      ra = ra - floor(ra/pme_grid)*pme_grid
      rb = rb - floor(rb/pme_grid)*pme_grid
      rc = rc - floor(rc/pme_grid)*pme_grid
      weight = 1.0d0
      ia0 = floor(ra)
      ib0 = floor(rb)
      ic0 = floor(rc)
      do ia = -lgrg_order+1, lgrg_order
        weight( ia,:,: ) = weight( ia,:,: ) * wfun( ia - (ra-ia0) )
      enddo
      do ib = -lgrg_order+1, lgrg_order
        weight( :,ib,: ) = weight( :,ib,: ) * wfun( ib - (rb-ib0) )
      enddo
      do ic = -lgrg_order+1, lgrg_order
        weight( :,:,ic ) = weight( :,:,ic ) * wfun( ic - (rc-ic0) )
      enddo
!     interpolate the reciprocal potential grid
      do ia = -lgrg_order+1, lgrg_order
        igrida = ia0 + ia
        igrida = igrida - floor(dble(igrida)/dble(pme_grid))*pme_grid
        igrida = igrida + 1
        do ib = -lgrg_order+1, lgrg_order
          igridb = ib0 + ib
          igridb = igridb - floor(dble(igridb)/dble(pme_grid))*pme_grid
          igridb = igridb + 1
          do ic = -lgrg_order+1, lgrg_order
            igridc = ic0 + ic
            igridc = igridc - floor(dble(igridc)/dble(pme_grid))*pme_grid
            igridc = igridc + 1
            pme_usegrid = pme_usegrid + weight(ia,ib,ic)*recip_pot(igrida,igridb,igridc)*atom_chg(i_mole,i_atom)
          enddo
        enddo
      enddo
!     real space interaction, based on COM coordinate
      r_com_i = r_com(i_mole, :)
      rtmp(:) = matmul(r_com(i_mole,:),inv_box(:,:))
      ra = rtmp(1) * na_nslist
      rb = rtmp(2) * nb_nslist
      rc = rtmp(3) * nc_nslist
      ra = ra - floor(ra/na_nslist)*na_nslist
      rb = rb - floor(rb/nb_nslist)*nb_nslist
      rc = rc - floor(rc/nc_nslist)*nc_nslist
      ia0 = floor(ra)
      ib0 = floor(rb)
      ic0 = floor(rc)
      pme_real = 0.0d0
! NOTE if this OMP section is uncommented, must add "use omp_lib" to the use section of the subroutine
! call OMP_SET_NUM_THREADS(n_threads)
!!$OMP PARALLEL DEFAULT(PRIVATE) SHARED(dia,dib,dic,ia0,ib0,ic0,headlist,nslist,r_com,r_com_i,box,ewald_cutoff_use,xyz_config,atom_chg,n_atom,i_mole,i_atom) REDUCTION(+:pme_real)
!!$OMP DO SCHEDULE(DYNAMIC,split_do)
      do ia = -dia, dia
        igrida = ia0 + ia
        igrida = igrida - floor(dble(igrida)/dble(na_nslist))*na_nslist
        igrida = igrida + 1
        do ib = -dib, dib
          igridb = ib0 + ib
          igridb = igridb - floor(dble(igridb)/dble(nb_nslist))*nb_nslist
          igridb = igridb + 1
          do ic = -dic, dic
            igridc = ic0 + ic
            igridc = igridc - floor(dble(igridc)/dble(nc_nslist))*nc_nslist
            igridc = igridc + 1
            j_mole = headlist(igrida, igridb, igridc, 1)
            j_atom = headlist(igrida, igridb, igridc, 2)
            do while ( j_mole > 0 )
              r_com_j = r_com(j_mole, :)
              shift = pbc_shift( r_com_i, r_com_j, box )
              dr_com = pbc_dr( r_com_i, r_com_j, shift )
              if ( abs(dr_com(1)) < ewald_cutoff_use .and. abs(dr_com(2)) < ewald_cutoff_use &
             &   .and. abs(dr_com(3)) < ewald_cutoff_use ) then
                if ( dot_product( dr_com, dr_com ) < ewald_cutoff_use**2 ) then
                  rij = pbc_dr( xyz_config(i_mole,i_atom,:), xyz_config(j_mole,j_atom,:), shift )
                  dr = dsqrt( dot_product( rij, rij ) )
!                 only inter-molecular interaction is possible here.
                  pme_real = pme_real + atom_chg(i_mole,i_atom)*atom_chg(j_mole,j_atom)*(erfc_g(dr*alpha_sqrt)-1.D0) / dr
                  pme_real = pme_real + atom_chg(i_mole,i_atom)*atom_chg(j_mole,j_atom) &
                 &              *screen(i_mole,j_mole,i_atom,j_atom,n_atom,dr,atom_chg) / dr
                endif
              endif
              tmp_mole = nslist(j_mole, j_atom, 1)
              tmp_atom = nslist(j_mole, j_atom, 2)
              j_mole = tmp_mole
              j_atom = tmp_atom
            enddo
          enddo
        enddo
      enddo
!!$OMP END DO NOWAIT
!!$OMP END PARALLEL
      pme_usegrid = pme_usegrid + pme_real
    enddo


!   The intramolecular part, if there are drudes oscillators
    if ( drude_simulation == 1 ) then
      do i_atom = 1, n_atom(i_mole)-1
        do j_atom = i_atom+1, n_atom(i_mole)

          call intra_elec_energy(E_intra,xyz_config,atom_chg,i_mole,i_atom,j_atom,n_atom)
          pme_usegrid = pme_usegrid + E_intra
        enddo
      enddo
    endif


    pme_usegrid = pme_usegrid *  0.52914D0 * 627.51D0 * 4.184D0 ! convert from e^2/A to kJ/mol

    return
  end function pme_usegrid

!************************************************************************************
! This subroutine calculates PME electric force on drude oscillators,  using the 
! potential grid.
! Lagrange interpolation of the grid is assumed
! Major output:
! real*8, dimension(3) :: force  ! the electrostatic force 
! Note, the unit of the potential grid should be in e^2/A
! the unit of the output force should be in e^2/A^2
! Currently, this code only works for orthorhombic box
! Note, need r_com for the real space cutoff calculations.
!************************************************************************************
  subroutine drude_force_usegrid( force, drude_atoms, tot_n_mole, n_mole, tot_n_atom, tot_xyz &
       &, chg, box, r_com, recip_pot, headlist, nslist )
    use global_variables
    implicit none
    integer,dimension(MAX_N_MOLE,MAX_N_ATOM)::drude_atoms
    integer, intent(in) :: tot_n_mole, n_mole
    integer, dimension(MAX_N_MOLE), intent(in) :: tot_n_atom
    real*8, dimension(MAX_N_MOLE,MAX_N_ATOM,3), intent(in) ::tot_xyz
    real*8, dimension(MAX_N_MOLE,MAX_N_ATOM), intent(in) :: chg
    real*8, dimension(:,:), intent(in) :: r_com
    real*8, dimension(:,:), intent(in) :: box
    real*8, dimension(:,:,:), intent(in) :: recip_pot
    integer, dimension(:,:,:,:), intent(in) :: headlist
    integer, dimension(MAX_N_MOLE,MAX_N_ATOM,2), intent(in) :: nslist
    real*8, dimension(MAX_N_MOLE,MAX_N_ATOM,3), intent(out) ::force
    !   local variables
    integer :: i_mole, i_atom, t_atom, j_mole, j_atom, tmp_mole, tmp_atom
    real*8, dimension(3) :: r_com_i, r_com_j, shift, dr_com, rij, xyz0, fij, rtmp, ftmp
    real*8, dimension(-lgrg_order+1:lgrg_order) :: wa, wb, wc, dwa, dwb, dwc
    integer :: ia0,ib0,ic0, ia,ib,ic, igrida,igridb,igridc, dia,dib,dic
    real*8 :: ra, rb, rc, ewald_cutoff_use, dr, rka, rkb, rkc

    !   effective cutoff for real space calculations
    ewald_cutoff_use = min(box(1,1)/2.,box(2,2)/2.,box(3,3)/2.,ewald_cutoff)
    !   real space cutoff for cell link list
    rka = dsqrt(dot_product(inv_box(:,1),inv_box(:,1)))
    rkb = dsqrt(dot_product(inv_box(:,2),inv_box(:,2)))
    rkc = dsqrt(dot_product(inv_box(:,3),inv_box(:,3)))
    dia = floor(ewald_cutoff_use*rka*dble(na_nslist)) + 1
    dib = floor(ewald_cutoff_use*rkb*dble(nb_nslist)) + 1
    dic = floor(ewald_cutoff_use*rkc*dble(nc_nslist)) + 1

    do i_mole = 1, n_mole
       do i_atom = 1, tot_n_atom(i_mole)
          if(drude_atoms(i_mole,i_atom).eq.0) then
             goto 100
!!$!!!!!!!!!!!!!move to next molecule
          endif
          t_atom = drude_atoms(i_mole, i_atom)
          xyz0 = tot_xyz(i_mole, t_atom, :)
          rtmp(:) = matmul(xyz0,inv_box(:,:))
          ra = rtmp(1) * pme_grid
          rb = rtmp(2) * pme_grid
          rc = rtmp(3) * pme_grid
          !       shift into box
          ra = ra - floor(ra/pme_grid)*pme_grid
          rb = rb - floor(rb/pme_grid)*pme_grid
          rc = rc - floor(rc/pme_grid)*pme_grid
          ia0 = floor(ra) ; ib0 = floor(rb) ; ic0 = floor(rc)
          !       Interpolate reciprocal potential
          do ia = -lgrg_order+1, lgrg_order
             call dwfun( ia-(ra-ia0), wa(ia), dwa(ia) )
             dwa(ia) = -1.0d0 * dwa(ia)  ! dw/dra
          enddo
          do ib = -lgrg_order+1, lgrg_order
             call dwfun( ib-(rb-ib0), wb(ib), dwb(ib) )
             dwb(ib) = -1.0d0 * dwb(ib)  ! dw/drb
          enddo
          do ic = -lgrg_order+1, lgrg_order
             call dwfun( ic-(rc-ic0), wc(ic), dwc(ic) )
             dwc(ic) = -1.0d0 * dwc(ic)  ! dw/drc
          enddo
          force(i_mole, i_atom, :) = 0.d0
          do ia = -lgrg_order+1, lgrg_order 
             igrida = ia0 + ia
             igrida = igrida - floor(dble(igrida)/dble(pme_grid))*pme_grid
             igrida = igrida + 1
             do ib = -lgrg_order+1, lgrg_order
                igridb = ib0 + ib
                igridb = igridb - floor(dble(igridb)/dble(pme_grid))*pme_grid
                igridb = igridb + 1
                do ic = -lgrg_order+1, lgrg_order
                   igridc = ic0 + ic
                   igridc = igridc - floor(dble(igridc)/dble(pme_grid))*pme_grid
                   igridc = igridc + 1
                   ftmp(1) = - chg(i_mole,t_atom)*recip_pot(igrida,igridb,igridc)*dwa(ia)*wb(ib)*wc(ic)
                   ftmp(2) = - chg(i_mole,t_atom)*recip_pot(igrida,igridb,igridc)*wa(ia)*dwb(ib)*wc(ic)
                   ftmp(3) = - chg(i_mole,t_atom)*recip_pot(igrida,igridb,igridc)*wa(ia)*wb(ib)*dwc(ic)
                    ! force(i_mole,i_atom,:) = force(i_mole,i_atom,:) + matmul(ftmp(:),inv_box(:,:))*dble(pme_grid)  ! changed by Jesse 6/9/2014.  This is a bug.  Next line should be correct, since inv_box is diagonal for orthorhombic unit cells, this is only an issue for non-orthorhombic unit cells
                   force(i_mole,i_atom,:) = force(i_mole,i_atom,:) + matmul(inv_box(:,:),ftmp(:))*dble(pme_grid)
                   !              force(i_mole,i_atom,1) = force(i_mole,i_atom,1) - dble(pme_grid)/box(1,1)* &
                   !                                     & chg(i_mole,t_atom)*recip_pot(igrida,igridb,igridc)*dwa(ia)*wb(ib)*wc(ic)
                   !              force(i_mole,i_atom,2) = force(i_mole,i_atom,2) - dble(pme_grid)/box(2,2)* &
                   !                                     & chg(i_mole,t_atom)*recip_pot(igrida,igridb,igridc)*wa(ia)*dwb(ib)*wc(ic)
                   !              force(i_mole,i_atom,3) = force(i_mole,i_atom,3) - dble(pme_grid)/box(3,3)* &
                   !                                     & chg(i_mole,t_atom)*recip_pot(igrida,igridb,igridc)*wa(ia)*wb(ib)*dwc(ic)
                enddo
             enddo
          enddo
          !       framework real space contribution
          r_com_i = r_com(i_mole, :)
          rtmp(:) = matmul( r_com_i(:), inv_box(:,:) )
          ra = rtmp(1) * na_nslist
          rb = rtmp(2) * nb_nslist
          rc = rtmp(3) * nc_nslist
          !       shift back to the box
          ra = ra - floor(ra/na_nslist)*na_nslist
          rb = rb - floor(rb/nb_nslist)*nb_nslist
          rc = rc - floor(rc/nc_nslist)*nc_nslist
          ia0 = floor(ra)
          ib0 = floor(rb)
          ic0 = floor(rc)
          !       searching neighbour list
          do ia = -dia, dia
             igrida = ia0 + ia
             igrida = igrida - floor(dble(igrida)/dble(na_nslist))*na_nslist
             igrida = igrida + 1
             do ib = -dib, dib
                igridb = ib0 + ib
                igridb = igridb - floor(dble(igridb)/dble(nb_nslist))*nb_nslist
                igridb = igridb + 1
                do ic = -dic, dic
                   igridc = ic0 + ic
                   igridc = igridc - floor(dble(igridc)/dble(nc_nslist))*nc_nslist
                   igridc = igridc + 1
                   j_mole = headlist(igrida, igridb, igridc, 1)
                   j_atom = headlist(igrida, igridb, igridc, 2)
                   do while ( j_mole > 0 )
                      r_com_j = r_com(j_mole, :)
                      shift = pbc_shift( r_com_i, r_com_j, box )
                      dr_com = pbc_dr( r_com_i, r_com_j, shift )
                      if ( abs(dr_com(1)) < ewald_cutoff_use .and. abs(dr_com(2)) < ewald_cutoff_use &
                           & .and. abs(dr_com(3)) < ewald_cutoff_use ) then
                         if ( dot_product( dr_com, dr_com ) < ewald_cutoff_use**2 ) then
                            rij = pbc_dr( xyz0, tot_xyz(j_mole,j_atom,:), shift )
                            dr = dsqrt( dot_product( rij, rij ) )
                            !                   only inter-molecular interaction is possible here.
                            force(i_mole,i_atom,:) = force(i_mole,i_atom,:) - &
                                 & rij(:)/dr * chg(i_mole,t_atom)*chg(j_mole,j_atom)*(erfc_g(dr*alpha_sqrt)-1.D0)/dr**2
                            force(i_mole,i_atom,:) = force(i_mole,i_atom,:) - &
                                 & rij(:)/dr * chg(i_mole,t_atom)*chg(j_mole,j_atom) &
                                 & *screen(i_mole,j_mole,t_atom,j_atom,tot_n_atom,dr,chg)/dr**2
                            force(i_mole,i_atom,:) = force(i_mole,i_atom,:) - &
                                 & rij(:)/dr * chg(i_mole,t_atom)*chg(j_mole,j_atom) &
                                 & *2.D0*alpha_sqrt/pi_sqrt*dexp(-(alpha_sqrt*dr)**2)/dr
                            force(i_mole,i_atom,:) = force(i_mole,i_atom,:) + &
                                 & chg(i_mole,t_atom)*chg(j_mole,j_atom) &
                                 & *d_screen(i_mole,j_mole,t_atom,j_atom,tot_n_atom,rij,dr,chg)/dr
                         endif
                      endif
                      tmp_mole = nslist(j_mole, j_atom, 1)
                      tmp_atom = nslist(j_mole, j_atom, 2)
                      j_mole = tmp_mole
                      j_atom = tmp_atom
                   enddo
                enddo
             enddo
          enddo

          !   The intramolecular force, if there are drudes oscillators
          if ( drude_simulation == 1 ) then
             do j_atom = 1, tot_n_atom(i_mole)
                if ( j_atom /= t_atom ) then
                   call intra_elec_force(fij,tot_xyz,chg,i_mole,tot_n_atom,t_atom,j_atom)
                   force(i_mole,i_atom,:) = force(i_mole,i_atom,:) + fij
                endif
             enddo
          endif

       enddo
100    continue
    enddo

    return
  end subroutine drude_force_usegrid

!**********************************************************************************************************************
! This subroutine construct the long-range dispersion interaction grid for a certain type of atom, which is specified
! by the variable atom_id1. It constructs the pseudo "charge" array based on dispersion coefficients, and feed it to 
! the reciprocal PME grid subroutine to generate the dispersion interaction grid.
! Important output:
! real*8, dimension(:,:,:) :: lrdisp_pot   ! the long-range dispersion grid
!**********************************************************************************************************************
  subroutine construct_recip_disp_grid(atom_id1,n_mole,tot_n_mole,n_atom,lrdisp_pot,box,xyz,dfti_desc,dfti_desc_inv)
  use global_variables
  use MKL_DFTI
    implicit none
    integer, intent(in) :: atom_id1, n_mole, tot_n_mole
    integer, dimension(:), intent(in) :: n_atom
    real*8, dimension(:,:), intent(in) :: box
    real*8, dimension(:,:,:), intent(in) :: xyz
    TYPE(DFTI_DESCRIPTOR),pointer,intent(in):: dfti_desc,dfti_desc_inv
    real*8, dimension(:,:,:), intent(out) :: lrdisp_pot
!   locale variables
    integer :: i_mole, atom_id2, i_atom
    real*8, dimension(MAX_N_MOLE,MAX_N_ATOM) :: cpij, c6, c8, c10, c12, c14
    real*8 :: hb, beta, mr, tmp

    lrdisp_pot = 0.0d0
    do i_mole = 1, tot_n_mole
      if ( i_mole <= n_mole ) then ! solute molecules
        do i_atom = 1, n_atom(i_mole)
          cpij(i_mole,i_atom) = 1.0d0
        enddo
      else
        do i_atom = 1, n_atom(i_mole)
          atom_id2 = atom_index(i_mole,i_atom)
          if ( lj_bkghm == 1 ) then
            cpij(i_mole,i_atom) = -1d0*atype_lj_parameter(atom_id1,atom_id2,3)
          else if ( lj_bkghm == 2 ) then
            cpij(i_mole,i_atom) = -4d0*atype_lj_parameter(atom_id1,atom_id2,1)*atype_lj_parameter(atom_id1,atom_id2,2)**6
          endif
        enddo
      endif
    enddo
    c6 = cpij
    c8 = 0.0d0
    c10 = 0.0d0
    c12 = 0.0d0
    c14 = 0.0d0
    do i_mole = 1, n_mole
      do i_atom = 1, n_atom(i_mole)
        c8(i_mole,i_atom) = 1.0d0
        c10(i_mole,i_atom) = 1.0d0
        c12(i_mole,i_atom) = 1.0d0
        c14(i_mole,i_atom) = 1.0d0
      enddo
    enddo
    if ( C8_10_dispersion_terms == 'yes' ) then 
!     C8 C10 terms
      do i_mole = n_mole+1, tot_n_mole
        do i_atom = 1, n_atom(i_mole)
          atom_id2 = atom_index(i_mole,i_atom)
          c8(i_mole,i_atom) = -1d0*atype_lj_parameter(atom_id1,atom_id2,4)
          c10(i_mole,i_atom) = -1d0*atype_lj_parameter(atom_id1,atom_id2,5)
          Select Case(C12_dispersion)
            Case("yes")
              c12(i_mole,i_atom) = -1d0*atype_lj_parameter(atom_id1,atom_id2,6)
          end Select
        enddo
      enddo
!      call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,cpij,dfti_desc,dfti_desc_inv,lrdisp_pot,8)
!      call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,cpij,dfti_desc,dfti_desc_inv,lrdisp_pot,10)
    endif
    if ( quantum_correct == 1 ) then ! quantum correction on
      do i_mole = n_mole + 1, tot_n_mole
        do i_atom = 1, n_atom(i_mole)
          atom_id2 = atom_index(i_mole,i_atom)
          beta = 1.0d0/8.314d0/temp*6.022e23 ! J-1
          hb = 6.626d-34/2d0/pi ! J*s
          mr = reduced_mass(atom_id1,atom_id2,n_atom)/1000d0/6.022e23 ! kg
          tmp = beta*hb**2/24d0/mr*1d20 ! in A^2
          c14(i_mole,i_atom) = c14(i_mole,i_atom) + tmp*12*11*c12(i_mole,i_atom)
          c12(i_mole,i_atom) = c12(i_mole,i_atom) + tmp*10*9*c10(i_mole,i_atom)
          c10(i_mole,i_atom) = c10(i_mole,i_atom) + tmp*8*7*c8(i_mole,i_atom)
          c8(i_mole,i_atom)  = c8(i_mole,i_atom)  + tmp*6*5*c6(i_mole,i_atom)
        enddo
      enddo
    endif

    call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,c6,dfti_desc,dfti_desc_inv,lrdisp_pot,6)   
    if ( C8_10_dispersion_terms == 'yes' .or. quantum_correct == 1 ) then
      call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,c8,dfti_desc,dfti_desc_inv,lrdisp_pot,8)
    endif
    if ( C8_10_dispersion_terms == 'yes' ) then
      call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,c10,dfti_desc,dfti_desc_inv,lrdisp_pot,10)
      call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,c12,dfti_desc,dfti_desc_inv,lrdisp_pot,12)
      if ( quantum_correct == 1 ) then
        call recip_pme_grid( box,tot_n_mole,n_mole,n_atom,xyz,c14,dfti_desc,dfti_desc_inv,lrdisp_pot,14) 
      endif
    endif

  end subroutine construct_recip_disp_grid

!*************************************************************************************************************
! This function calculates the dispersion energy between the guest molecules and the framework, using the 
! dispersion potential grid lrdisp_pot.
! Important input:
! real*8, dimension(:,:,:,:), intent(in) :: lrdisp_pot ! variable calculated using construct_recip_disp_grid
!*************************************************************************************************************
  function disp_usegrid( box,i_mole,tot_n_mole,n_mole,n_atom,r_com,xyz_config,lrdisp_pot,headlist,nslist,mass)
  use global_variables
  use routines
  implicit none
    integer, intent(in) :: tot_n_mole, n_mole, i_mole
    integer, dimension(:), intent(in) :: n_atom
    integer, dimension(:,:,:,:), intent(in) :: headlist
    integer, dimension(MAX_N_MOLE,MAX_N_ATOM,2), intent(in) :: nslist
    real*8, dimension(:,:,:), intent(in) :: xyz_config
    real*8, dimension(:,:,:,:), intent(in) :: lrdisp_pot
    real*8, dimension(:,:), intent(in) :: box, r_com, mass
    real*8 :: disp_usegrid
!   local variables
    integer :: i_atom, j_mole, j_atom, tmp_mole, tmp_atom, atom_id1, atom_id2
    real*8 :: ra,rb,rc, dra,drb,drc,dr, ewald_cutoff_use, pme_real, E_short, E_long, tmp, rka, rkb, rkc
    real*8, dimension(5) :: energy_decomp
    integer :: ia,ib,ic, ia0,ib0,ic0, igrida,igridb,igridc, dia,dib,dic, split_do
    real*8, dimension(-lgrg_order+1:lgrg_order,-lgrg_order+1:lgrg_order,-lgrg_order+1:lgrg_order) :: weight
!    real*8, dimension(2,2,2) :: weight
    real*8, dimension(3) :: r_com_i, r_com_j, shift, dr_com,rij, rtmp
    real*8 :: c6, c8, c10, c12, beta, hb, mr, lambda

!   effective cutoff for real space calculations
    ewald_cutoff_use = min(box(1,1)/2.,box(2,2)/2.,box(3,3)/2.,lj_cutoff)
!   real space cutoff for cell link list
    rka = dsqrt(dot_product(inv_box(:,1),inv_box(:,1)))
    rkb = dsqrt(dot_product(inv_box(:,2),inv_box(:,2)))
    rkc = dsqrt(dot_product(inv_box(:,3),inv_box(:,3)))
    dia = floor(ewald_cutoff_use*rka*dble(na_nslist)) + 1
    dib = floor(ewald_cutoff_use*rkb*dble(nb_nslist)) + 1
    dic = floor(ewald_cutoff_use*rkc*dble(nc_nslist)) + 1

    if (n_threads .eq. 1 ) then
       split_do = 1
    else
       split_do = (dia*2+1)/n_threads
    endif

    disp_usegrid = 0.0d0
    do i_atom = 1, n_atom(i_mole)
      atom_id1 = atom_index(i_mole, i_atom)
      rtmp(:) = matmul(xyz_config(i_mole,i_atom,:),inv_box(:,:))
      ra = rtmp(1) * pme_grid
      rb = rtmp(2) * pme_grid
      rc = rtmp(3) * pme_grid
!     shift back to the box
      ra = ra - floor(ra/pme_grid)*pme_grid
      rb = rb - floor(rb/pme_grid)*pme_grid
      rc = rc - floor(rc/pme_grid)*pme_grid
      weight = 1.0d0
      ia0 = floor(ra)
      ib0 = floor(rb)
      ic0 = floor(rc)
      do ia = -lgrg_order+1, lgrg_order
        weight( ia,:,: ) = weight( ia,:,: ) * wfun( ia - (ra-ia0) )
      enddo
      do ib = -lgrg_order+1, lgrg_order
        weight( :,ib,: ) = weight( :,ib,: ) * wfun( ib - (rb-ib0) )
      enddo
      do ic = -lgrg_order+1, lgrg_order
        weight( :,:,ic ) = weight( :,:,ic ) * wfun( ic - (rc-ic0) )
      enddo
!     interpolate the reciprocal potential grid
      do ia = -lgrg_order+1, lgrg_order
        igrida = ia0 + ia
        igrida = igrida - floor(dble(igrida)/dble(pme_grid))*pme_grid
        igrida = igrida + 1
        do ib = -lgrg_order+1, lgrg_order
          igridb = ib0 + ib
          igridb = igridb - floor(dble(igridb)/dble(pme_grid))*pme_grid
          igridb = igridb + 1
          do ic = -lgrg_order+1, lgrg_order
            igridc = ic0 + ic
            igridc = igridc - floor(dble(igridc)/dble(pme_grid))*pme_grid
            igridc = igridc + 1 
            disp_usegrid = disp_usegrid + weight(ia,ib,ic)*lrdisp_pot(atom_id1,igrida,igridb,igridc)
          enddo
        enddo
      enddo
!     real space interaction, based on COM coordinate and neighbour list cell
      r_com_i = r_com(i_mole, :)
      rtmp(:) = matmul(r_com(i_mole,:),inv_box(:,:))
      ra = rtmp(1) * na_nslist
      rb = rtmp(2) * nb_nslist
      rc = rtmp(3) * nc_nslist
      ra = ra - floor(ra/na_nslist)*na_nslist
      rb = rb - floor(rb/nb_nslist)*nb_nslist
      rc = rc - floor(rc/nc_nslist)*nc_nslist
      ia0 = floor(ra)
      ib0 = floor(rb)
      ic0 = floor(rc)
      pme_real = 0.0d0
! NOTE if this OMP section is uncommented, must add "use omp_lib" to the use section of the subroutine
! call OMP_SET_NUM_THREADS(n_threads)
!!$OMP PARALLEL DEFAULT(PRIVATE) SHARED(dia,dib,dic,ia0,ib0,ic0,headlist,nslist,r_com,r_com_i,box,ewald_cutoff_use,xyz_config,mass,n_atom,i_mole,i_atom,atom_index,lj_bkghm,atype_lj_parameter,atom_id1) REDUCTION(+:pme_real)
!!$OMP DO SCHEDULE(DYNAMIC,split_do)
      do ia = -dia, dia
        igrida = ia0 + ia
        igrida = igrida - floor(dble(igrida)/dble(na_nslist))*na_nslist
        igrida = igrida + 1
        do ib = -dib, dib
          igridb = ib0 + ib
          igridb = igridb - floor(dble(igridb)/dble(nb_nslist))*nb_nslist
          igridb = igridb + 1
          do ic = -dic, dic
            igridc = ic0 + ic
            igridc = igridc - floor(dble(igridc)/dble(nc_nslist))*nc_nslist
            igridc = igridc + 1
            j_mole = headlist(igrida, igridb, igridc, 1)
            j_atom = headlist(igrida, igridb, igridc, 2)
            do while ( j_mole > 0 )
              r_com_j = r_com(j_mole, :)
              shift(:) = pbc_shift( r_com_i, r_com_j, box )
              dr_com = pbc_dr( r_com_i, r_com_j, shift )
              if ( abs(dr_com(1)) < ewald_cutoff_use .and. abs(dr_com(2)) < ewald_cutoff_use &
             &   .and. abs(dr_com(3)) < ewald_cutoff_use ) then
                if ( dot_product( dr_com, dr_com ) < ewald_cutoff_use**2 ) then
                  rij = pbc_dr( xyz_config(i_mole,i_atom,:), xyz_config(j_mole,j_atom,:), shift )
                  dr = dsqrt( dot_product( rij, rij ) )
!                 only inter-molecular interaction is possible here.
                  atom_id2 = atom_index(j_mole,j_atom)
                  if ( lj_bkghm == 1 ) then ! Buckingham potential
                    c6 = atype_lj_parameter(atom_id1,atom_id2,3)
                    c8 = atype_lj_parameter(atom_id1,atom_id2,4)
                    c10 = atype_lj_parameter(atom_id1,atom_id2,5)
                    select case(C12_dispersion)
                    case("yes")
                      c12 = atype_lj_parameter(atom_id1,atom_id2,6)
                    case("no")
                      c12 = 0.0d0
                    end select
                    ! short range interactions
                    call eval_short_range_bkghm(E_short, energy_decomp, atom_id1,atom_id2, dr)
                    pme_real = pme_real + E_short
                    ! C6 term
                    pme_real = pme_real - ( gfun_g(lj_asqrt*dr,6) - 1d0 + C6_C10_damp(atom_id1,atom_id2,dr,6) ) * c6/dr**6
                    ! C8 term
                    pme_real = pme_real - ( gfun_g(lj_asqrt*dr,8) - 1d0 + C6_C10_damp(atom_id1,atom_id2,dr,8) ) * c8/dr**8
                    ! C10 term
                    pme_real = pme_real - ( gfun_g(lj_asqrt*dr,10) - 1d0 + C6_C10_damp(atom_id1,atom_id2,dr,10) ) * c10/dr**10
                    ! if using C12 terms
                    if ( C12_dispersion == "yes" ) then
                      pme_real = pme_real - ( gfun_g(lj_asqrt*dr,12) - 1d0 + C6_C10_damp(atom_id1,atom_id2,dr,12) ) * c12/dr**12
                    endif
                    ! add up quantum correction
                    if ( quantum_correct == 1 ) then
                      beta = 1.0d0/8.314d0/temp*6.022e23 ! J-1
                      hb = 6.626d-34/2d0/pi ! J*s
                      mr = reduced_mass(atom_id1,atom_id2,n_atom)/1000d0/6.022e23 ! kg
                      tmp = beta*hb**2/24d0/mr*1d20 ! A^2
                      ! exponential part
                      lambda = atype_lj_parameter(atom_id1,atom_id2,2)
                      pme_real = pme_real + tmp*atype_lj_parameter(atom_id1,atom_id2,1) &
                               & * exp(-lambda*dr) * ( lambda**2 - 2d0*lambda/dr )
                      ! C6 terms
                      pme_real= pme_real - tmp*c6/dr**6*( ddC6_C10_damp(atom_id1,atom_id2,dr,6) &
                             & - 2d0*(6-1)/dr*dC6_C10_damp_scalar(atom_id1,atom_id2,dr,6) )
                      pme_real= pme_real - tmp*6*5*c6/dr**8*(gfun_g(lj_asqrt*dr,8) - 1d0 + C6_C10_damp(atom_id1,atom_id2,dr,6))
                      ! C8 terms
                      pme_real= pme_real - tmp*c8/dr**8*( ddC6_C10_damp(atom_id1,atom_id2,dr,8) &
                             & - 2d0*(8-1)/dr*dC6_C10_damp_scalar(atom_id1,atom_id2,dr,8) )
                      pme_real= pme_real - tmp*8*7*c8/dr**10*(gfun_g(lj_asqrt*dr,10) - 1d0 + C6_C10_damp(atom_id1,atom_id2,dr,8))
                      ! C10 terms
                      pme_real= pme_real - tmp*c10/dr**10*( ddC6_C10_damp(atom_id1,atom_id2,dr,10) &
                             & - 2d0*(10-1)/dr*dC6_C10_damp_scalar(atom_id1,atom_id2,dr,10) )
                      pme_real= pme_real - tmp*10*9*c10/dr**12*(gfun_g(lj_asqrt*dr,12) - 1d0 + C6_C10_damp(atom_id1,atom_id2,dr,10))
                      ! C12 terms
                      if ( C12_dispersion == "yes" ) then
                        pme_real= pme_real - tmp*c12/dr**12*( ddC6_C10_damp(atom_id1,atom_id2,dr,12) &
                               & - 2d0*(12-1)/dr*dC6_C10_damp_scalar(atom_id1,atom_id2,dr,12) )
                        pme_real= pme_real - tmp*12*11*c12/dr**14*(gfun_g(lj_asqrt*dr,14) - 1d0 &
                                                                 & + C6_C10_damp(atom_id1,atom_id2,dr,12))
                      endif
                    endif 
                  else if ( lj_bkghm == 2 ) then ! Lennard-Jones potential
                    c12 = 4d0 * atype_lj_parameter(atom_id1,atom_id2,1)*atype_lj_parameter(atom_id1,atom_id2,2)**12
                    c6  = 4d0 * atype_lj_parameter(atom_id1,atom_id2,1)*atype_lj_parameter(atom_id1,atom_id2,2)**6
                    E_short = c12/dr**12
                    pme_real = pme_real + E_short
                    pme_real = pme_real - gfun_g(alpha_sqrt*dr,6) * c6/dr**6
                    if ( quantum_correct == 1 ) then
                      beta = 1.0d0/8.314d0/temp*6.022e23 ! J-1
                      hb = 6.626d-34/2d0/pi ! J*s
                      mr = reduced_mass(atom_id1,atom_id2,n_atom)/1000d0/6.022e23 ! kg
                      tmp = beta*hb**2/24d0/mr*1d20 ! A^2
                      ! C12 terms
                      pme_real = pme_real + tmp*12*11*c12/dr**14
                      ! C6 terms
                      pme_real = pme_real - tmp*6*5*c6/dr**8*gfun_g(lj_asqrt*dr,8)
                    endif
                  else
                    stop 'lj_bkghm in disp_usegrid is unknown'
                  endif
                endif
              endif
              tmp_mole = nslist(j_mole, j_atom, 1)
              tmp_atom = nslist(j_mole, j_atom, 2)
              j_mole = tmp_mole
              j_atom = tmp_atom
            enddo
          enddo
        enddo
      enddo
!!$OMP END DO NOWAIT
!!$OMP END PARALLEL
      disp_usegrid = disp_usegrid + pme_real
    enddo

    return
  end function disp_usegrid


  !*************************************************
  ! this subroutine evaluates the short range part of a buckingham pair wise potential
  ! the main reason for a separate routine here is if a separate penetration force field is 
  ! being used.  Note there is no select case for penetration_force_field.  Rather,
  ! if a penetration force field is not used, atype_penetration(:,:,2) will be set to zero
  ! for all atoms,which is sufficient
  !*************************************************
  subroutine eval_short_range_bkghm(E_short, energy_decomp,atom_id1,atom_id2,r_ij)
    use global_variables
    integer,intent(in) :: atom_id1, atom_id2
    real*8,intent(in) :: r_ij
    real*8,intent(out) :: E_short
    real*8,dimension(:),intent(inout) :: energy_decomp

    real*8 :: Exponential,sum, energy_test=1d-8
    integer :: i_type

    ! determine whether pairwise distance is beyond cutoff
    if ( r_ij > atype_penetration_ff(atom_id1,atom_id2,2) ) then
       ! decomposed force field, check energy decomposition
       Select Case(energy_decomposition)
       Case("no")
          E_short =  atype_lj_parameter(atom_id1,atom_id2,1)*exp(-atype_lj_parameter(atom_id1,atom_id2,2)*r_ij)
       Case("yes")
          Exponential = exp(-atype_lj_parameter(atom_id1,atom_id2,2)*r_ij)
          sum=0d0
          do i_type =1, 5
             energy_decomp(i_type) = energy_decomp(i_type) + atype_bkghm_decomp(atom_id1,atom_id2,i_type) * Exponential
             sum = sum + atype_bkghm_decomp(atom_id1,atom_id2,i_type) * Exponential
          enddo
          E_short = atype_lj_parameter(atom_id1,atom_id2,1) * Exponential
          ! check consistency of energy decomp
          if ( abs(sum - E_short) > energy_test ) then
             write(*,*) "short range energy decomposition doesn't equal E_short for atom types ", atom_id1, atom_id2
             stop
          endif
       End Select
    else
       ! penetration force field, no contribution to energy decomposition
       E_short = atype_penetration_ff(atom_id1,atom_id2,1)*exp(-atype_lj_parameter(atom_id1,atom_id2,2)*r_ij) &
              &  - atype_penetration_ff(atom_id1,atom_id2,3)
    endif

  end subroutine eval_short_range_bkghm



end module pme_routines
