module space_group_info
  use global_variables
  implicit none
  !****************************************************************************************
  ! This module contains all variables related to space group symmetry operations
  ! Variables in this module:
  ! integer :: n_ops   ! Number of useful symmetry operations in the current space group
  ! real*8, dimension(n_ops,3,3) :: rotat_list  ! Rotation operation list 
  ! real*8, dimension(n_ops,3) :: trans_list    ! Translation operation list
  !****************************************************************************************
  integer :: n_ops
  real*8, dimension(:,:,:), allocatable :: rotat_list
  real*8, dimension(:,:), allocatable :: trans_list

contains

  !******************************************************************************
  ! This subroutine reads the symmetry operations from the input file
  !******************************************************************************
  subroutine read_oplist( symfn, n_ops )
    implicit none
    character(MAX_FN), intent(in) :: symfn
    integer, intent(out) :: n_ops
    integer :: i_ops
    integer :: ifile=99

    open( ifile, file=symfn, status='old' )
    read( ifile, '(I)' ) n_ops

    allocate( rotat_list(n_ops,3,3), trans_list(n_ops,3) )

    do i_ops = 1, n_ops
       rotat_list(i_ops,:,:) = 0.0d0
       trans_list(i_ops,:) = 0.0d0
       read( ifile, '(12F10.6)' ) rotat_list(i_ops,1,1), rotat_list(i_ops,1,2), rotat_list(i_ops,1,3), trans_list(i_ops,1) &
            & ,rotat_list(i_ops,2,1), rotat_list(i_ops,2,2), rotat_list(i_ops,2,3), trans_list(i_ops,2) &
            & ,rotat_list(i_ops,3,1), rotat_list(i_ops,3,2), rotat_list(i_ops,3,3), trans_list(i_ops,3)
    enddo
    close(ifile)

    return
  end subroutine read_oplist

  !********************************************************************************
  ! Given the certain symmetry operation (specified by iop), this subroutine gives
  ! the corresponding result position.
  ! Everything should be in direct coordinate
  !********************************************************************************
  subroutine sym_operate( rb, iop, re )
    implicit none
    integer, intent(in) :: iop
    real*8, dimension(3), intent(in) :: rb
    real*8, dimension(3), intent(out) :: re

    re = matmul( rotat_list(iop,:,:), rb ) + trans_list(iop,:)
    ! shift back to box
    re(1) = re(1) - floor(re(1))
    re(2) = re(2) - floor(re(2))
    re(3) = re(3) - floor(re(3))

    return
  end subroutine sym_operate


  !*******************************************************************************
  ! This subroutine checks that all the symmetry operations are compatible with
  ! The .fconf file.  
  ! for example even though every translated unit cell can be used without symmetry,
  ! translations will be incompatible with the original symmetry operations for 
  ! a specific unit cell
  !
  ! Also there may be some extra symmetry imposed on the system.  for example,
  ! The hydrogens of a methyl group might break the symmetry, but the symmetry operations
  ! in the .cif file may ignore this.
  !
  ! So lets ignore Hydrogen atoms when searching symmetry operations
  !*******************************************************************************
  subroutine check_symmetry_fconf_consistency( alist, n_mole, tot_n_mole, n_atom, xyz, box )
    use routines
    character(MAX_ANAME), dimension(:,:), intent(in) :: alist    
    integer, intent(in) :: n_mole, tot_n_mole
    integer, dimension(:), intent(in) :: n_atom
    real*8, dimension(:,:,:), intent(in) :: xyz
    real*8, dimension(:,:) , intent(in) :: box

    integer :: i_atom, i_mole,j_mole,j_atom, iop, i, flag, flagH
    character(MAX_ANAME) :: i_type, j_type, testtype
    real*8,dimension(3) :: ri,rj, rjxyz, dr,shift
    real*8 :: small = 1D-3  ! careful, dont make this too small, lattice positions probably aren't accurate to within < .001 Angstroms



    ! only check if there are symmetry operations
    if ( n_ops > 0 ) then


       do i_mole = n_mole + 1, tot_n_mole
          do i_atom = 1, n_atom(i_mole)

             i_type = alist(i_mole,i_atom)

             ! rm spaces for index
             testtype = i_type
             call trim_head( testtype )

             ! IGNORE Hydrogen atoms
             flagH=0
             if((index( testtype,'h').eq.1).or.(index(testtype,'H').eq.1)) then
                flagH=1
             endif

             if ( flagH /= 1 ) then

                ri(:) = xyz(i_mole,i_atom,:)
                ! direct coordinates
                ri = matmul(ri, inv_box )

                ! loop over all symmetry operations, verifying that there is a symmetric atom there
                do iop = 1, n_ops
                   flag=0

                   call sym_operate( ri , iop, rj )

                   ! back to xyz coords
                   rjxyz=0d0
                   do i=1,3
                      rjxyz(:) = rjxyz(:) + rj(i) * box(i,:)
                   enddo


                   ! loop over every framework atom, try to find symmetric atom
                   do j_mole = n_mole + 1 , tot_n_mole
                      do j_atom = 1, n_atom(j_mole)

                         j_type = alist(j_mole,j_atom)

                         if ( i_type .eq. j_type ) then

                            shift = pbc_shift( rjxyz, xyz(j_mole, j_atom, :), box )
                            dr = pbc_dr( rjxyz, xyz(j_mole, j_atom, :), shift )

                            if ( ( abs(dr(1)) < small ) .and. ( abs(dr(2)) < small ) .and. ( abs(dr(3)) < small ) ) then
                               flag=1
                            endif
                         endif
                      enddo
                   enddo

                   ! if we couldn't find a symmetry atom, stop code
                   if ( flag == 0 ) then
                      write(*,*) ""
                      write(*,*) " symmetry operation ", iop, " is inconsistent with unit cell"
                      write(*,*) " please make sure symmetry operations are correct, check that"
                      write(*,*) " unit cell hasn't been translated, not using supercell, etc. "
                      write(*,*) ""
                      stop
                   endif

                enddo ! end loop over symmetry operations

             endif ! end if on Hydrogen atom
          enddo  ! end loop over i_atom
       enddo  ! end loop over i_mole
    endif


  end subroutine check_symmetry_fconf_consistency

end module space_group_info
