!************************************************************
!
! This module contains global variables that are used in many subroutines
! throughout the Monte Carlo program
!
! With a few exceptions (such as the PME, Ewald charge grids), the global data
! in this module doesn't change once initialized.
!
! trajectory data that is changing during the course of the simulations (such as xyz coordinates)
! is normally not stored in this module
!
! The initialization of this global data occurs in multiple ways
!
! 1) some of this data is hard-coded in.  Any parameters that are hard coded in that the user might want to change are
!    placed at the top of the module, as labeled
!
! 2) some of the technical simulation parameter data is read in from the simulation_parameters.pmt input file
!
! 3) The force field parameters are stored here as global data, and are read in from the force field input file
!
! 4) Finally, some of the mapping arrays are constructed logically by the code.
!
!*************************************************************


module global_variables
implicit none

!***************************** NOTE THIS SECTION ***********************************************
! here we organize all global variable parameters that are hard-coded in, that the user
! may want to adjust in certain cases

  ! these parameters control the size of data arrays.  Make sure MAX_N_MOLE > maximum number of molecules in the simulations
  ! and MAX_N_ATOM > maximum number of atoms per molecule
  integer, parameter :: MAX_FN=100, MAX_ANAME=5, MAX_MNAME=5, MAX_N_MOLE=10000, MAX_N_ATOM=40

  ! these parameters control the Drude oscillator convergence:  Maximum number of iterations and convergence threshold
  integer,parameter::iteration_limit = 15    ! if scf iterations exceed this limit, reject move
  real*8,parameter:: force_threshold = 4.0477D-6  ! this is for rms force in e^2/A^2, equal to .1 KJ/mol/nm

  ! this shifts the vdw part of the force field to zero at the cutoff
  integer, parameter :: lj_shift  = 0       ! 0 = no shift, 1 = shift to zero at cutoff

! interpolation for PME dispersion grid.
  integer, parameter :: lgrg_order=2 ! if this equals 1, then the grid is interpolated linearly

! dhf combination rule:  As we used sort of a weird convention for the sign of the dhf cross terms in the beginning, but changed to an completely attractive dhf cross term convention when fitting virials, we need to implement a variety of combination rules
! here "1" is as in JPC C, 2012, 116, 1892-1903
! "2" is Aij = positive iff Aii and Ajj positive (difference between 1 and 2 is Aij negative in 2 if Aii and Ajj both negative
! "3" is always attractive
integer, parameter :: dhf_combination_rule=2

! these variables determine whether to grid expensive functions in memory.  Code is much faster when these are set to 'yes'
character(3)  :: grid_erfc = "yes"        ! grid error function for pme
character(3)  :: grid_Tang_Toennies ! grid damping functions
!***********************************************************************************************



  real*8, parameter :: pi=3.14159265
  real*8, parameter :: pi_sqrt=1.7724538509d0
  real*8,dimension(12) :: factorial    ! this saves the first 12 factorials which are used in Tang-Toennies damping functions


! framework
  integer, parameter :: N_input_f = 6, N_input = 5      ! number of input files with and without use of framework
  integer            :: framework_simulation            ! tells subroutines whether or not this simulation includes fixed framework atoms

! code version
  integer  :: code_version = 30

! if yes, this will not run a simulation, but calculate the interaction energy between a solute molecule and the framework over specified grid
 character(3) :: energy_grid_run = "no"    ! will just call energy_grid subroutine

! for puddle filling
character(3) :: puddle_filling_on = "no"
real*8  :: puddle_min = -15.
real*8  :: puddle_delta_E

! Feynmann Hibbs quantum correction for hydrogen
character(3) :: Feynmann_Hibbs_correction
character(3) :: Feynmann_Hibbs_forces

! this changes the input data structure, as now the code must be able to handle starting with zero molecules, also it reads in a random number seed.
  character(3) :: replica_input = "no"
  integer,dimension(:),allocatable :: seed_replica

! choose ensemble
  character(3) :: select_ensemble        ! nvt, npt, uvt, egc

! input file type is determined by whether or not a sapt_type force field is being used
 character(3) :: sapt_type_ff
 character(3) :: energy_decomposition     ! an explicit energy decomposition will be given if using a (purely) sapt-based force field ( and without pme dispersion )

! penetration force field for short distances , this will only be used with energy decomposition
character(3) :: penetration_force_field
character(3) :: switch_at_minimum = "no"
real*8 :: penetration_force_field_threshold

! if we want to use a different solute parameter set to create solute framework cross terms, we have to read this in separately
 character(3) :: solute_cross_parameter_set 

! if, energy decomposition, have option of using C8,C10 parameters for dispersion.  This also determines input file structure
 character(3) :: C8_10_dispersion_terms  ! if yes, damping for dispersion will be turned on, and buckingham coefficients for dispersion will be read in
 character(3) :: C12_dispersion

 ! three body force field flags
 character(3) :: three_body_dispersion
 character(3) :: three_body_exchange


! if using pme dispersion
 character(3) :: pme_disp
 real*8, dimension(:,:,:,:), allocatable :: lrdisp_pot

! explicitly read solute-framework exponents
 character(3) :: explicit_cross_term_exponents="no"

! flag tells code whether to damp solute-solute dispersion interactions
 character(3) :: damp_solute_solute_dispersion

! hybrid mc/md or regular mc?
character(3)  :: hybrid_md_mc         ! "no" for standard mc translation/rotation moves (very slow for drudes!), "yes" for hybrid_md_mc

!force field parameters
  integer:: n_atom_type_store  ! used for expanded GCMC where we have to introduce additional atom types
  integer:: n_atom_type, n_molecule_type
  character(MAX_ANAME), dimension(MAX_N_ATOM) :: atype_name
  character(MAX_MNAME), dimension(MAX_N_MOLE) :: molecule_name
  real*8, dimension(MAX_N_ATOM) :: atype_chg,atype_pol
  real*8, dimension(MAX_N_ATOM) :: atype_mass    
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM,5) :: atype_bkghm_decomp                   ! this stores "A" type coefficients for Bkingham, index 1-5 are exch,elec,induc,dhf,dispersion, respectively
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM,6) :: atype_lj_parameter        ! for buckingham, store A,B,C6 (possibly C8,C10), for LJ, store epsilon,sigma ; index 6 is for possible C12 terms
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM,MAX_N_ATOM) :: atype_3body_C9            ! this stores C9 coefficients, for 3-body dispersion
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM,MAX_N_ATOM,2) :: atype_3body_exchange      ! this stores parameters, for 3-body exchange
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM,4) :: atype_penetration_ff      ! penetration force field, first index is bkghm coeff, 2 index is distance cutoff, 3 index is energy shift, 4 index is exponent scale factor
  integer, dimension(MAX_N_ATOM) ::  atype_solute                      ! keeps track of whether this atom type is framework(0) or solute(1)
  integer, dimension(MAX_N_ATOM,MAX_N_ATOM) ::  atype_damp_dispersion   ! keeps track of whether we damp the dispersion interaction between these atom types
  integer, dimension(MAX_N_MOLE,MAX_N_ATOM) :: molecule_type      ! this array contains indices for all types of solute molecules (not framework) end is marked with MAX_N_ATOM+1
  character(MAX_MNAME), dimension(MAX_N_MOLE) :: molecule_type_name
!  real*8, dimension(MAX_N_MOLE,3,3) :: molecule_type_inertia 
  real*8, dimension(MAX_N_MOLE,MAX_N_ATOM,3) :: r_mole_com     ! stores configuration for all molecule types. for non-linear molecules, configurations are stored in principle coordinate basis
  integer, dimension(MAX_N_MOLE) :: molecule_shape      ! flag "0" for nonlinear, polyatomic, "1" for linear molecules, "2" for atom
  real*8,dimension(MAX_N_MOLE,3) :: molecule_inertia
  real*8,dimension(MAX_N_MOLE) :: molecule_mass
  integer, dimension(MAX_N_MOLE,MAX_N_ATOM):: atom_index            ! integer specifies where to look up atom parameters in lj_parameter array
  integer, dimension(MAX_N_MOLE):: molecule_index            ! integer indexes molecule type for all solute molecules in simulation
  integer, dimension(MAX_N_MOLE,MAX_N_ATOM):: drude_atom_map            ! this array maps which atom each Drude oscillator resides on
  integer                 :: lj_bkghm                ! =1 for bkghm, =2 for lj
  character(10)            :: lj_comb_rule            ! for lj, set to "opls" or "standard"== Lorentz-Berthelot, for bkghm, set to "standard" or "ZIFFF"
 ! if lj_bkghm is set to 3, meaning we are using hybrid lj/bkghm force field, we need a 2nd combination rule.  The first "lj_comb_rule" will then be used for the bkghm force field for solute-framework interactions, and "lj_comb_rule2" will be used for the lj force field for solute-solute interactions
  character(10)            :: lj_comb_rule2            ! this will only be used if lj_bkghm=3
  integer, dimension(MAX_N_MOLE,MAX_N_ATOM) :: lj_bkghm_index ! this  maps which atom-atom types use a lj interaction, and which use a buckingham interaction, which is really only necessary for lj_bkghm=3, but is used always

  ! this are for gridding the ewald real space atom type interactions
  ! note that these are atom type specific because of the damping functions
  character(3) :: grid_ewald_realspace_interaction="no"
  integer,parameter :: ewald_realspace_interaction_grid_size = 1000000
  real*8, dimension(:,:), allocatable :: ewald_realspace_interaction_grid
  integer,dimension(MAX_N_ATOM, MAX_N_ATOM) :: ewald_realspace_interaction_grid_map
  real*8 :: ewald_realspace_table_pad = 10d0 ! this is padding beyond ewald_cutoff for table construction which is used because cutoff is applied on a molecule by molecule basis, and so atom-atom distances can be farther than this
  real*8 :: max_ewald_table



! mc parameters, specified here temporarilly
  integer   :: n_update_stepsize, n_output, n_step
  real*8 :: temp  
  integer :: cycles_per_vol_step  ! this is the average cycles ( 1 hybrid move, or nmole single particle_moves)/vol_move
  real*8 :: too_close         ! if any atoms on different molecules closer than this, reject move right away
  real*8 :: delta_t1           ! this is time step for integrator in hybridmc (ps)
  real*8 :: vol_step           ! this is (twice the maximum) percentage change in ln(volume) (.05 is reasonable)
  real*8 :: max_delta_t1 , max_vol_step , target_acc
  real*8 :: target_acc_v      ! desired percentage accepted moves
  real*8, parameter:: step_update=.1,step_update_v=.05  ! percentage by which to change step sizes
  real*8 :: pressure   ! pressure in bar
  real*8, parameter :: p_conv=(1D-03/16.6054)  ! bar to kj/mol/A^3
! Buckingham related parameters
  real*8  :: lj_cutoff 
  integer  :: lj_lrc          ! 0 = no lrc, 1 = lrc (lrc will never be computed if there is framework)
  real*8, dimension(MAX_N_ATOM,MAX_N_ATOM) :: lj_shift_store

! transformation matrix to box vectors
  real*8,dimension(3,3) :: xyz_to_box_transform

! Verlet list
  character(3) :: use_verlet_list
  integer,dimension(:),allocatable :: verlet_neighbor_list
  integer,dimension(MAX_N_MOLE) :: verlet_point
  real*8       :: verlet_cutoff
  integer,parameter :: safe_verlet=1.2  ! this is factor that we multiply theoretically needed size of verlet list to be safe

! single molecule moves
  real*8 :: pct_singlemv_hybrid = .9
  integer, parameter :: translation_method= 1 ! 1 for displacement, 2 for lattice hop
  integer, parameter :: rotate_method = 2  ! 1 for linear, 2 for general euler angles
  integer, parameter :: reset_trans_rot = 10 ! see subroutine update_stepsize_trans_rot
  real*8 :: single_mol_step =.2 , max_single_mol_step = 4d0
  real*8 :: single_mol_rotate = .2 ,max_single_mol_rotate = 5d0 ! for type 1 rotation
  real*8 :: phi_step = .1 , psi_step = .1 , theta_step = .05 ! for type 2 rotation
  real*8 :: max_phi_step = 2.6 , max_psi_step = 2.6 , max_theta_step = 1.3 ! for type 2 rotation

! parameters for Electrostatic (Ewald/Cutoff/pme)
  integer, parameter :: kmax=100, maxk=400000, maxn=1000
  real*8, parameter :: ksqmax=10. ! ksqmax is cutoff for k vectors
  real*8            :: ewald_cutoff
  real*8            :: Electro_cutoff
  integer           :: screen_type           ! screening for coulomb potential 0 = none, 1 = Tang-Toennies type
  real*8,parameter  :: screen_distance_max_sq = 50d0 ! this is the maximum distance (squared) for which the screening function will be used in the electrostatic force calculation, since it is expensive, and it is extremely close to 1 at this point
  ! PME parameters.  alpha_sqrt controls the width of the smearing Gaussians in the Ewald sum.  As always, changing this value
  ! determines the relative magnitudes of reciprocal and real space sums, and therefore appropriate real space cutoff and PME grid size
  real*8             ::  alpha_sqrt   ! Gaussian width parameter in A^-1,
  integer            ::  pme_grid     ! this is the PME grid size in each dimension
  integer            ::  spline_order ! this is the order of Beta-splines used in the PME charge interpolation

  character(10)     :: electrostatic_type     ! choose from pme, cutoff, or none. ewald is only implemented for testing
  ! global variables in Ewald calculations
  integer :: totk
  character(3) :: ewald_volume_move ! this flag signals which global variables the ewald function should update
  complex*16, dimension(maxk) :: rou_k, rou_k_try
  real*8, dimension(maxk,3) :: k_vec,new_k_vec
  real*8, dimension(maxk) :: B_factor,new_B_factor
  real*8 :: Ewald_self, Ewald_framework ! These two have different units: e^2/A
  real*8 :: E_disp_lrc,E_disp_lrc_try

  ! parameters for pme
  real*8,dimension(:,:,:),allocatable::CB,Q_grid,theta_conv_Q
  ! real*8,dimension(pme_grid,pme_grid,pme_grid)::Q_grid_try
  !real*8::pme_realE,pme_realE_try
  ! grid Bsplines,for order n, need n spline and n-1 spline grids
  ! B(i) corresponds to B(i*n/gridsize)
  integer,parameter::spline_grid=1000000,erfc_grid=1000000, gfun_grid=500000
  real*8,parameter::erfc_max=10d0              ! this is max value up to which erfc is grid
  real*8,parameter::gfun_max=5.0d0
  real*8,dimension(spline_grid)::B6_spline,B5_spline,B4_spline,B3_spline
  real*8,dimension(erfc_grid):: erfc_table
  real*8,dimension(gfun_grid):: g6_table, g8_table, g10_table, g12_table, g14_table


  ! C6-C12 dispersion damping function tables
  ! the value of the 6th order Tang Toennies damping function at x=30 is 0.99999988268 , 8th order is 0.999997953924096, 10th order is 0.9999776512, 12th order is 0.9998323
  ! This should be a fine max values, since at x=30, the C10/R^10, C12/R^12 terms should be basically zero for the corresponding distance
  real*8,parameter:: Tang_Toennies_max=30d0  
  integer,parameter:: Tang_Toennies_grid=1000000
  real*8,dimension(4,Tang_Toennies_grid) :: Tang_Toennies_table, dTang_Toennies_table ! C6 is in 1st index, C8 in 2nd, C10 in 3rd, C12 in 4th
  real*8,dimension(Tang_Toennies_grid) :: Tang_Toennies_table3  ! this is used for three-body dispersion damping functions

  ! parameters for drude oscillator
  real*8          :: springcon
  real*8          :: thole 
  integer        :: drude_simulation         ! this flag will let sampling routines know whether to call scf drude (=1 for drudes)


  ! parameters for three body dispersion
  real*8 :: three_body_dispersion_cutoff
  integer :: three_body_cutoff_type   ! this should be "0" if three body cutoff acts on molecules as a whole, or "1" if it is an atom by atom basis
  integer :: na_nslist, nb_nslist, nc_nslist   ! link list parameters
  integer, dimension(:,:,:), allocatable :: headlist
  integer, dimension(MAX_N_MOLE) :: nslist

  ! parameters for parallel openmp
  integer:: n_threads

  ! debug: 0= normal, 1= print force calculation time decomposition for pme, 2=general time decomposition, options > 10 are for testing certain parts of code.  See specific test jobs.
  integer :: debug
  character(8)::date
  character(10)::time

  ! Gibbs ensemble, need copies of certain variables for 2nd box (first box uses regular variables)
real*8 :: E_disp_lrc2,E_disp_lrc_try2
real*8,dimension(2) :: Ewald_self_gibbs
!real*8,dimension(pme_grid,pme_grid,pme_grid)::CB2,Q_grid2,theta_conv_Q2
real*8,parameter:: gibbs_vol_pct=0.2,gibbs_swap_pct=0.2        ! fraction of moves that change volume,swap particles
integer  :: orientation_try    ! number of orientation tries for particle insertion bias, =1 means unbiased

  ! grand canonical ensemble
real*8          :: gcmc_ins_rm        ! fraction of moves that attempt to insert or remove a particle, for expanded grand, this is attempts to change coupling
real*8          :: chem_potential
   
 ! expanded grand canonical ensemble
integer :: partial_molecule_index
integer :: partial_molecule_coupling = 0  ! our initially energy calculation will have all molecules at full coupling, so this should have an initial value of 0
integer,parameter :: partial_coupling_steps = 5
real*8  :: coupling_increment = 1d0/dble(partial_coupling_steps)
real*8,dimension(partial_coupling_steps) :: step_weight

! cavity-bias
character(3)      :: cavity_bias = "no"   ! yes or no
integer,parameter :: cav_grid_a = 26, cav_grid_b = 26 , cav_grid_c = 26
integer,parameter :: energy_bias = 0 ! 0= naive cavity bias, 1= lennard jones boltzmann factor
real*8,parameter :: min_frmwk_dist = 2d0                 ! min distance from center of cubelet to framework for cubelet to be considered for insertion
real*8           :: cubelet_vol                            ! calculated from box size and grid size

end module
