#!/usr/bin/perl

# this script calculates the chemical potential of N2 at a given temperature
# and pressure based on the Peng Robinson equation of state

#*************Temperature and Pressure of interest
$temp  = 184;               #K
$press = 758;               # torr
$press = $press / 750.06;

#******** !!!!!!!!!!!!!!! important !!!!!!!!!!!! *****
# set reference molar volume (ideal gas), this depends on temperature of interest
# = .025 m^3/mol for ideal gas at 300K,1bar
$v_ref = .25;    #25 ; # m^3/mol

#***********EOS parameters (NIST) *****************
$T_crit = 126.192;
$P_crit = 33.958;    #bar
$w      = .0372;     # acentric factor

#**************** definitions lets use SI units
$T_red         = $temp / $T_crit;
$R             = 8.314;
$bar_to_pascal = 10**5;
$k             = .37464 + 1.54226 * $w - 0.26992 * $w**2;
$a             = .457235 * $R**2 * $T_crit**2 / ( $P_crit * $bar_to_pascal );
$b             = .077796 * $R * $T_crit / ( $P_crit * $bar_to_pascal );
$alpha         = ( 1 + $k * ( 1 - $T_red**.5 ) )**2;

#*************** reference values (ideal gas)
$pi     = 3.141592654;
$mass   = 28;
$kg_mol = $mass / 1000;

# this converts kg/m3 to molecules/A3
$kgm3_moleA3 = 6.022 * 10**23 / $kg_mol * 10**(-30);
$dens_conv   = ( 10**5 / ( 8.314 * $temp ) ) * $kg_mol * $kgm3_moleA3;
$lambda      = 126.178 / sqrt( 2 * $pi * $mass * $temp * 8.314 );

# reference pressure
$ref_press_pa = pressure( $R, $temp, $v_ref, $alpha, $a, $b );
$ref_press = $ref_press_pa / $bar_to_pascal;

$dens        = $dens_conv * $ref_press;
$chempot_ref = .008314 * $temp * log( $lambda**3 * $dens );

# find target molar volume.  Rather than inverting the pressure formula, we do this iteratively
# start from pv=RT
$thresh = .001;
$v0 = $temp * $R / ( $press * $bar_to_pascal );
print "$v0\n";
$chgv = 1;
do {
    $p0 = pressure( $R, $temp, $v0, $alpha, $a, $b );
    $dp0 = dpdv( $R, $temp, $v0, $alpha, $a, $b );

    print "$p0\n";
    print "$dp0\n";

    $delv = ( $press * $bar_to_pascal - $p0 ) / $dp0;

    # don't go to negative volumes!
    $inc  = 1;
    $vtry = -1;
    do {
        $vtry = $v0 + $inc * $delv;
        $inc  = $inc * .5;
    } until ( $vtry > 0 );

    $chgv = ( $vtry - $v0 ) / $v0;
    $v0   = $vtry;

    print "$v0\n";

} until ( abs($chgv) < $thresh );

$vtarget = $v0;

# now find dmu from gibbs duhem starting at the reference molar volume (constant temp )
$sum    = 0;
$nsteps = 10000000;
$dv     = ( $vtarget - $v_ref ) / $nsteps;
$v_step = $v_ref;

for ( my $i = 0 ; $i < $nsteps ; $i++ ) {
    $dp     = dpdv( $R, $temp, $v_step, $alpha, $a, $b );
    $sum    = $sum + $v_step * $dp * $dv;
    $v_step = $v_step + $dv;
}

# correction is J/mol, chempot is in KJ/mol, so divide by 1000

$chempot = $chempot_ref + $sum / 1000;

print "reference chemical potential at $temp K and $ref_press bar\n";
print "chempot is $chempot_ref kJ/mol \n";
print
  " chemical potential at $temp K and $press bar using Peng Robinson EOS \n";
print " chempot is $chempot KJ/mol \n";

# pressure subroutine, input is R , T , molar volume , alpha, a, b

sub pressure {
    $R     = $_[0];
    $T     = $_[1];
    $v     = $_[2];
    $alpha = $_[3];
    $a     = $_[4];
    $b     = $_[5];
    $p = $R * $T / ( $v - $b ) - $a * $alpha / ( $v**2 + 2 * $b * $v - $b**2 );
    return $p;
}

# dp/dv subroutine where v is molar volume.  input same as above

sub dpdv {
    $R     = $_[0];
    $T     = $_[1];
    $v     = $_[2];
    $alpha = $_[3];
    $a     = $_[4];
    $b     = $_[5];

    $dp =
      -$R * $T / ( $v - $b )**2 +
      ( $a * $alpha ) * ( 2 * $v + 2 * $b ) / ( $v**2 + 2 * $b * $v - $b**2 )
      **2;
    return $dp;
}
